/***********************************************************************\
*
* $Revision: 2164 $
* $Date: 2017-09-14 08:41:11 +0200 (Thu, 14 Sep 2017) $
* $Author: torsten $
* Contents: command to view changes log
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;

// graphics
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Widget;

/****************************** Classes ********************************/

/** view changes log command
 */
class CommandChangesLog
{
  /** changes log comparator
   */
  class ChangeLogDataComparator extends LogDataComparator
  {
    /** create changes log comparator
     * @param table table
     * @param sortColumn column to sort
     */
    ChangeLogDataComparator(Table table, TableColumn sortColumn)
    {
      if      (table.getColumn(0) == sortColumn) setSortMode(LogDataComparator.SortModes.REVISION      );
      else if (table.getColumn(1) == sortColumn) setSortMode(LogDataComparator.SortModes.AUTHOR        );
      else if (table.getColumn(2) == sortColumn) setSortMode(LogDataComparator.SortModes.DATETIME      );
      else if (table.getColumn(3) == sortColumn) setSortMode(LogDataComparator.SortModes.COMMIT_MESSAGE);
      else                                       setSortMode(LogDataComparator.SortModes.REVISION      );
    }
  }

  /** dialog data
   */
  class Data
  {
    java.util.List<LogData> logDataList;
    String                  author;
    LogData                 selectedLogData0,selectedLogData1;
    FileData                fileData;

    Data()
    {
      this.logDataList      = null;
      this.author           = "";
      this.selectedLogData0 = null;
      this.selectedLogData1 = null;
      this.fileData         = null;
    }
  };

  // --------------------------- constants --------------------------------
//  private final Color COLOR_CURRENT   = Onzen.COLOR_YELLOW;
  private final Color COLOR_SELECTED0;
  private final Color COLOR_SELECTED1;

  // --------------------------- variables --------------------------------

  // global variable references
  private final RepositoryTab repositoryTab;
  private final Display       display;
  private final Clipboard     clipboard;

  // dialog
  private final Data          data = new Data();
  private final Shell         dialog;

  // widgets
  private final Text          widgetSelectedLog0;
  private final Text          widgetSelectedLog1;
  private final Table         widgetChangesLog;
  private final TableColumn   widgetChangesLogColumn;
  private final List          widgetFiles;
  private final Text          widgetFind;
  private final Combo         widgetAuthor;
  private final Button        widgetDiff;
  private final Button        widgetPatch;
  private final Text          widgetSelectedLog;
  private final Button        widgetEdit;
  private final Button        widgetView;
  private final Button        widgetSave;
//  private final Button        widgetRevert;
  private final Button        widgetClose;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create revision view
   * @param shell shell
   * @param repositoryTab repository tab
   * @param fileData file data
   * @param revision revision to show or null
   */
  CommandChangesLog(final Shell shell, final RepositoryTab repositoryTab)
  {
    Composite   composite,subComposite;
    Label       label;
    TableColumn tableColumn;
    Menu        menu;
    MenuItem    menuItem;
    Button      button;

    // initialize variables
    this.repositoryTab = repositoryTab;

    // get display, clipboard
    display   = shell.getDisplay();
    clipboard = new Clipboard(display);

    // init colors
    COLOR_SELECTED0 = new Color(display,Settings.colorSelect0.background);
    COLOR_SELECTED1 = new Color(display,Settings.colorSelect1.background);

    // create dialog
    dialog = Dialogs.open(shell,"Changes log",new double[]{1.0,0.0},1.0);

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{0.7,0.0,0.3},1.0));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      // changes log
      widgetChangesLog = Widgets.newTable(composite,SWT.H_SCROLL|SWT.V_SCROLL|SWT.SINGLE);
      widgetChangesLog.setBackground(Onzen.COLOR_WHITE);
      Widgets.layout(widgetChangesLog,0,0,TableLayoutData.NSWE,0,0,4);
      SelectionListener selectionListener = new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          TableColumn             tableColumn             = (TableColumn)selectionEvent.widget;
          ChangeLogDataComparator changeLogDataComparator = new ChangeLogDataComparator(widgetChangesLog,tableColumn);

          synchronized(widgetChangesLog)
          {
            Widgets.sortTableColumn(widgetChangesLog,tableColumn,changeLogDataComparator);
          }
        }
      };
      tableColumn = Widgets.addTableColumn(widgetChangesLog,0,"Revision", SWT.RIGHT);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetChangesLog,1,"Author",   SWT.LEFT);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetChangesLog,2,"Date/Time",SWT.LEFT);
      tableColumn.addSelectionListener(selectionListener);
      widgetChangesLogColumn = Widgets.addTableColumn(widgetChangesLog,3,"Message",SWT.LEFT);
      widgetChangesLogColumn.addSelectionListener(selectionListener);
      Widgets.setTableColumnWidth(widgetChangesLog,Settings.geometryChangesLogColumns.width);
      menu = Widgets.newPopupMenu(dialog);
      {
/*
        menuItemGotoRevision = Widgets.addMenuItem(menu,"Goto revision");
        menuItemGotoRevision.setEnabled(false);
        menuItemGotoRevision.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            show(data.selectedLog);
          }
        });
*/
      }
      widgetChangesLog.setMenu(menu);
      widgetChangesLog.setToolTipText("Changes log. Double-click to view revision info.");

      // find
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(1.0,new double[]{0.0,0.6,0.0,0.4}));
      Widgets.layout(subComposite,1,0,TableLayoutData.WE);
      {
        label = Widgets.newLabel(subComposite,"Find:",SWT.NONE,Settings.keyFind);
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetFind = Widgets.newText(subComposite,SWT.SEARCH|SWT.ICON_CANCEL);
        widgetFind.setMessage("Enter text to find");
        Widgets.layout(widgetFind,0,1,TableLayoutData.WE);

        label = Widgets.newLabel(subComposite,"Author:");
        Widgets.layout(label,0,2,TableLayoutData.W);

        widgetAuthor = Widgets.newCombo(subComposite,SWT.READ_ONLY);
        Widgets.layout(widgetAuthor,0,3,TableLayoutData.WE);
      }

      // files
      widgetFiles = Widgets.newList(composite);
      widgetFiles.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetFiles,2,0,TableLayoutData.NSWE);
      menu = Widgets.newPopupMenu(dialog);
      {
        menuItem = Widgets.addMenuItem(menu,"Revision info"+"\u2026",Settings.keyRevisionInfo);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            CommandRevisionInfo commandRevisionInfo = new CommandRevisionInfo(shell,repositoryTab,data.fileData,data.selectedLogData1.revisionId);
            commandRevisionInfo.run();
          }
        });

        menuItem = Widgets.addMenuItem(menu,"Diff"+"\u2026",Settings.keyDiff);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetDiff);
          }
        });

        menuItem = Widgets.addMenuItem(menu,"Patch"+"\u2026",Settings.keyDiff);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetPatch);
          }
        });

        menuItem = Widgets.addMenuItem(menu,"Save"+"\u2026",Settings.keyDiff);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetSave);
          }
        });
      }
      widgetFiles.setMenu(menu);
      widgetFiles.setToolTipText("Changed files.\nCtrl-A to select all, Ctrl-C to copy filenames.");
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,0.0}));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"For");
      Widgets.layout(label,0,0,TableLayoutData.W);

      widgetSelectedLog0 = Widgets.newStringView(composite);
      widgetSelectedLog0.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetSelectedLog0,0,1,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetSelectedLog0,data)
      {
        public void modified(Control control)
        {
          if (!widgetSelectedLog0.isDisposed()) widgetSelectedLog0.setText((data.selectedLogData0 != null) ? data.selectedLogData0.revisionId : "");
        }
      });

      label = Widgets.newLabel(composite,"->");
      Widgets.layout(label,0,2,TableLayoutData.W);

      widgetSelectedLog1 = Widgets.newStringView(composite);
      widgetSelectedLog1.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetSelectedLog1,0,3,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetSelectedLog1,data)
      {
        public void modified(Control control)
        {
          if (!widgetSelectedLog1.isDisposed()) widgetSelectedLog1.setText((data.selectedLogData1 != null) ? data.selectedLogData1.revisionId : "");
        }
      });

      label = Widgets.newLabel(composite,"do:");
      Widgets.layout(label,0,4,TableLayoutData.W);

      widgetDiff = Widgets.newButton(composite,"Diff"+"\u2026",Settings.keyDiff);
      widgetDiff.setEnabled(false);
      Widgets.layout(widgetDiff,0,5,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetDiff,data)
      {
        public void modified(Control control)
        {
          if (!control.isDisposed()) control.setEnabled((data.selectedLogData0 != null) && (data.selectedLogData1 != null) && (data.fileData != null));
        }
      });
      widgetDiff.addSelectionListener(new SelectionListener()
      {
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          if ((data.selectedLogData1 != null) && (data.fileData != null))
          {
            CommandDiff commandDiff = new CommandDiff(dialog,
                                                      repositoryTab,
                                                      data.fileData,
                                                      (data.selectedLogData0 != null) ? data.selectedLogData0.revisionId : repositoryTab.repository.getFirstRevision(),
                                                      (data.selectedLogData1 != null) ? data.selectedLogData1.revisionId : null
                                                     );
            commandDiff.run();
          }
        }
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
      });

      widgetPatch = Widgets.newButton(composite,"Patch"+"\u2026");
      widgetPatch.setEnabled(false);
      Widgets.layout(widgetPatch,0,6,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetPatch,data)
      {
        public void modified(Control control)
        {
          if (!control.isDisposed()) control.setEnabled((data.selectedLogData0 != null) && (data.selectedLogData1 != null) && (data.fileData != null));
        }
      });
      widgetPatch.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          if ((data.selectedLogData1 != null) && (data.fileData != null))
          {
            CommandCreatePatch commandCreatePatch;
            if (data.selectedLogData0 != null)
            {
              commandCreatePatch = new CommandCreatePatch(dialog,
                                                          repositoryTab,
                                                          data.fileData.toSet(),
                                                          data.selectedLogData0.revisionId,
                                                          data.selectedLogData1.revisionId,
                                                          false
                                                         );
            }
            else
            {
              commandCreatePatch = new CommandCreatePatch(dialog,
                                                          repositoryTab,
                                                          data.fileData.toSet(),
                                                          data.selectedLogData1.revisionId,
                                                          false
                                                         );
            }
            commandCreatePatch.run();
          }
        }
      });

      label = Widgets.newLabel(composite,"For");
      Widgets.layout(label,0,7,TableLayoutData.W);

      widgetSelectedLog = Widgets.newStringView(composite);
      widgetSelectedLog.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetSelectedLog,0,8,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetSelectedLog,data)
      {
        public void modified(Control control)
        {
          if (!widgetSelectedLog.isDisposed()) widgetSelectedLog.setText((data.selectedLogData1 != null) ? data.selectedLogData1.revisionId : "");
        }
      });

      label = Widgets.newLabel(composite,"do:");
      Widgets.layout(label,0,9,TableLayoutData.W);

      widgetEdit = Widgets.newButton(composite,"Edit"+"\u2026");
      widgetEdit.setEnabled(false);
      Widgets.layout(widgetEdit,0,10,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetEdit,data)
      {
        public void modified(Control control)
        {
          if (!control.isDisposed()) control.setEnabled(repositoryTab.repository.supportEditCommit() && (data.selectedLogData1 != null) && (data.fileData != null));
        }
      });
      widgetEdit.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          if ((data.selectedLogData1 != null) && (data.fileData != null))
          {
            editCommitMessage(data.selectedLogData1);
          }
        }
      });

      widgetView = Widgets.newButton(composite,"View"+"\u2026");
      widgetView.setEnabled(false);
      Widgets.layout(widgetView,0,11,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetView,data)
      {
        public void modified(Control control)
        {
          if (!control.isDisposed()) control.setEnabled((data.selectedLogData1 != null) && (data.fileData != null));
        }
      });
      widgetView.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          if ((data.selectedLogData1 != null) && (data.fileData != null))
          {
            CommandView commandView = new CommandView(dialog,repositoryTab,data.fileData,data.selectedLogData1.revisionId);
            commandView.run();
          }
        }
      });

      widgetSave = Widgets.newButton(composite,"Save"+"\u2026");
      widgetSave.setEnabled(false);
      Widgets.layout(widgetSave,0,12,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetSave,data)
      {
        public void modified(Control control)
        {
          if (!control.isDisposed()) control.setEnabled((data.selectedLogData1 != null) && (data.fileData != null));
        }
      });
      widgetSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          if ((data.selectedLogData1 != null) && (data.fileData != null))
          {
            try
            {
              // get file
              byte[] fileDataBytes =  repositoryTab.repository.getFileBytes(data.fileData,data.selectedLogData1.revisionId);

              // save to file
              String fileName = Dialogs.fileSave(dialog,"Save file");
              if (fileName != null)
              {
                try
                {
                  DataOutputStream output = new DataOutputStream(new FileOutputStream(fileName));
                  output.write(fileDataBytes);
                  output.close();
                }
                catch (IOException exception)
                {
                  Dialogs.error(dialog,"Cannot save file '%s' (error: %s)",data.fileData.getName(repositoryTab.repository),exception.getMessage());
                }
              }
            }
            catch (RepositoryException exception)
            {
              Dialogs.error(dialog,"Cannot save file '%s' (error: %s)",data.fileData.getName(repositoryTab.repository),exception.getMessage());
              Onzen.printStacktrace(exception);
            }
          }
        }
      });

/*
      widgetRevert = Widgets.newButton(composite,"Revert"+"\u2026",Settings.keyRevert);
      widgetRevert.setEnabled(false);
      Widgets.layout(widgetRevert,0,13,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetRevert,data)
      {
        public void modified(Control control)
        {
          if (!control.isDisposed()) control.setEnabled((data.selectedLogData1 != null));
        }
      });
      widgetRevert.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          if ((data.selectedLogData1 != null) && (data.fileData != null))
          {
            if (Dialogs.confirm(dialog,String.format("Revert file '%s' to revision %s?",fileData.getName(repositoryTab.repository),data.selectedLogData1.revisionId)))
            {
              try
              {
                // revert file
                repositoryTab.repository.revert(fileData,data.selectedLogData1.revisionId);

                // update state
                repositoryTab.repository.updateStates(fileData);
                display.syncExec(new Runnable()
                {
                  public void run()
                  {
                    if (!repositoryTab.widgetComposite.isDisposed()) repositoryTab.updateTreeItem(fileData);
                  }
                });
              }
              catch (RepositoryException exception)
              {
                Dialogs.error(dialog,"Cannot revert file '%s' (error: %s)",fileData.getName(repositoryTab.repository),exception.getMessage());
                Onzen.printStacktrace(exception);
              }
            }
          }
        }
      });
*/

      widgetClose = Widgets.newButton(composite,"Close");
      Widgets.layout(widgetClose,0,14,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetClose.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          Settings.geometryChangesLog        = dialog.getSize();
          Settings.geometryChangesLogColumns = new Settings.ColumnSizes(Widgets.getTableColumnWidth(widgetChangesLog));

          Dialogs.close(dialog);
        }
      });
    }

    // listeners
    widgetChangesLog.addMouseListener(new MouseListener()
    {
      public void mouseDoubleClick(MouseEvent mouseEvent)
      {
        Table widget = (Table)mouseEvent.widget;

        int index = widget.getSelectionIndex();
        if (index >= 0)
        {
          LogData logData = (LogData)widget.getItem(index).getData();

          CommandRevisionInfo commandRevisionInfo = new CommandRevisionInfo(dialog,
                                                                            repositoryTab,
                                                                            data.fileData,
                                                                            logData.revisionId
                                                                           );
          commandRevisionInfo.run();
        }
      }
      public void mouseDown(MouseEvent mouseEvent)
      {
        Table widget = (Table)mouseEvent.widget;

        int index = widget.getSelectionIndex();
        if (index >= 0)
        {
          // selected revision
          selectRevision((LogData)widget.getItem(index).getData());

          // notify modification
          Widgets.modified(data);
        }
      }
      public void mouseUp(MouseEvent mouseEvent)
      {
      }
    });
    widgetChangesLog.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (Widgets.isAccelerator(keyEvent,Settings.keyFind))
        {
          widgetFind.forceFocus();
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    });
    widgetFiles.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        List     widget    = (List)selectionEvent.widget;
        String[] fileNames = widget.getSelection();

        if ((data.selectedLogData1 != null) && (data.fileData != null))
        {
          CommandDiff commandDiff = new CommandDiff(dialog,
                                                    repositoryTab,
                                                    data.fileData,
                                                    (data.selectedLogData0 != null) ? data.selectedLogData0.revisionId : repositoryTab.repository.getFirstRevision(),
                                                    (data.selectedLogData1 != null) ? data.selectedLogData1.revisionId : null
                                                   );
          commandDiff.run();
        }
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        List     widget    = (List)selectionEvent.widget;
        String[] fileNames = widget.getSelection();

        if (fileNames.length > 0)
        {
          data.fileData = repositoryTab.repository.getFileDataInstance(fileNames[0]);

          Widgets.modified(data);
        }
      }
    });
    widgetFiles.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (Widgets.isAccelerator(keyEvent,Settings.keyFind))
        {
          widgetFind.forceFocus();
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyDiff))
        {
          Widgets.invoke(widgetDiff);
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyCreatePatch))
        {
          Widgets.invoke(widgetPatch);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'a'))
        {
          widgetFiles.selectAll();
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'c'))
        {
          String[] fileNames = widgetFiles.getSelection();

          if (fileNames.length > 0)
          {
            StringBuilder buffer = new StringBuilder();
            for (String fileName : fileNames)
            {
              buffer.append(fileName); buffer.append('\n');
            }
            Widgets.setClipboard(clipboard,buffer.toString());
          }
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    });
    widgetFind.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        refresh();
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        refresh();
      }
    });
    widgetFind.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
      }
      public void keyReleased(KeyEvent keyEvent)
      {
        refresh();
      }
    });
    widgetAuthor.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        data.author = widget.getText().trim();

        refresh();
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        data.author = widget.getText().trim();

        refresh();
      }
    });

    KeyListener keyListener = new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (Widgets.isAccelerator(keyEvent,Settings.keyFind))
        {
          widgetFind.forceFocus();
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'c'))
        {
          TableItem[] tableItems = widgetChangesLog.getSelection();
          Widgets.setClipboard(clipboard,tableItems,4);
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    };
    widgetChangesLog.addKeyListener(keyListener);
    widgetFind.addKeyListener(keyListener);

    // show dialog
    Dialogs.show(dialog,Settings.geometryChangesLog,Settings.setWindowLocation);

    // show
    show();
  }

  /** set scroll value
   * @param scrollX,scrollY scroll values
   */
  public void setScroll(int scrollX, int scrollY)
  {
    widgetChangesLog.getHorizontalBar().setSelection(scrollX);
    widgetChangesLog.getVerticalBar().setSelection(scrollY);
    widgetChangesLog.redraw();
  }

  /** run dialog
   */
  public void run()
  {
    widgetFind.setFocus();
    Dialogs.run(dialog);
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "CommandChangesLog {}";
  }

  //-----------------------------------------------------------------------

  /** scroll to revision
   * @param logData revision to show
   */
  private void scrollTo(final LogData logData)
  {
    if (!dialog.isDisposed())
    {
      display.syncExec(new Runnable()
      {
        public void run()
        {
//TODO:
Dprintf.dprintf("");
        }
      });
    }
  }

  /** show revisions
   * @param revision revision to show
   */
  private void show(String revision)
  {
    // clear
    if (!dialog.isDisposed())
    {
      display.syncExec(new Runnable()
      {
        public void run()
        {
          data.logDataList = null;
          Widgets.modified(data);
         }
      });
    }

    // start show changes log
    Background.run(new BackgroundRunnable(revision)
    {
      public void run(final String revision)
      {
        // get log
        Widgets.setCursor(dialog,Onzen.CURSOR_WAIT);
        repositoryTab.setStatusText("Get changes log"+"\u2026");
        try
        {
          data.logDataList = repositoryTab.repository.getLog();
        }
        catch (RepositoryException exception)
        {
          final String exceptionMessage = exception.getMessage();
          if (!display.isDisposed())
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                if (!dialog.isDisposed())
                {
                  Dialogs.error(dialog,"Get changes log fail: %s",exceptionMessage);
                }
              }
            });
          }
          Onzen.printStacktrace(exception);
          return;
        }
        catch (Throwable throwable)
        {
          display.syncExec(new Runnable()
          {
            public void run()
            {
              dialog.dispose();
            }
          });
          Onzen.printInternalError(throwable);
          return;
        }
        finally
        {
          repositoryTab.clearStatusText();
          Widgets.resetCursor(dialog);
        }

        // get authors
        HashSet<String> authorSet = new HashSet<String>();
        for (LogData logData : data.logDataList)
        {
          authorSet.add(logData.author);
        }
        final ArrayList<String> authorList = new ArrayList<String>(authorSet);
        Collections.sort(authorList);
        display.syncExec(new Runnable()
        {
          public void run()
          {
            if (!widgetAuthor.isDisposed())
            {
              widgetAuthor.add("");
              for (String author : authorList)
              {
                widgetAuthor.add(author);
              }
            }
          }
        });

        // show
        refresh();
      }
    });
  }

  /** refresh log view
   */
  private void refresh()
  {
    if (!widgetChangesLog.isDisposed())
    {
      display.syncExec(new Runnable()
      {
        public void run()
        {
          if (data.logDataList != null)
          {
            String findText = widgetFind.getText().toLowerCase();

            widgetChangesLog.removeAll();
            int     maxWidth = 0;
            int     lineNb   = 1;
            boolean found    = false;
            for (LogData logData : data.logDataList)
            {
              String revisionId    = logData.revisionId;
              String author        = logData.author;
              String date          = Onzen.DATETIME_FORMAT.format(logData.date);
              String commitMessage = StringUtils.join(logData.commitMessageList,'\n');

              if (   (   findText.isEmpty()
                      || (revisionId.toLowerCase().indexOf(findText) >= 0)
                      || (author.toLowerCase().indexOf(findText) >= 0)
                      || (date.toLowerCase().indexOf(findText) >= 0)
                      || (commitMessage.toLowerCase().indexOf(findText) >= 0)
                     )
                  && (   data.author.isEmpty()
                      || data.author.equalsIgnoreCase(author)
                     )
                 )
              {
                Widgets.addTableItem(widgetChangesLog,
                                     logData,
                                     revisionId+StringUtils.repeat('\n',logData.commitMessageList.size()-1),
                                     author+StringUtils.repeat('\n',logData.commitMessageList.size()-1),
                                     date+StringUtils.repeat('\n',logData.commitMessageList.size()-1),
                                     commitMessage
                                    );
                maxWidth = Math.max(maxWidth,Widgets.getTextWidth(widgetChangesLog,logData.commitMessageList));
                lineNb++;
                found = true;
              }
            }
            widgetChangesLogColumn.setWidth(maxWidth);
            if (!found)
            {
              Widgets.flash(widgetFind);
            }
          }
        }
      });
    }
  }

  /** show last revision
   */
  private void show()
  {
    show(repositoryTab.repository.getLastRevision());
  }

  /** select revision
   * @param logData0,logData1 logData
   */
  private void selectRevision(LogData logData0, LogData logData1)
  {
    if (data.selectedLogData0 != null) Widgets.setTableItemColor(widgetChangesLog,data.selectedLogData0,null);
    if (data.selectedLogData0 != null) Widgets.setTableItemColor(widgetChangesLog,data.selectedLogData1,null);
    data.selectedLogData0 = logData0;
    data.selectedLogData1 = logData1;
    if (data.selectedLogData0 != null) Widgets.setTableItemColor(widgetChangesLog,data.selectedLogData0,COLOR_SELECTED0);
    if (data.selectedLogData1 != null) Widgets.setTableItemColor(widgetChangesLog,data.selectedLogData1,COLOR_SELECTED1);
    data.fileData = null;
//Dprintf.dprintf("data.selectedLogData0=%s %s",data.selectedLogData0,COLOR_SELECTED0);
//Dprintf.dprintf("data.selectedLogData1=%s %s",data.selectedLogData1,COLOR_SELECTED1);

    if (data.selectedLogData1 != null)
    {
      if (!dialog.isDisposed())
      {
        display.syncExec(new Runnable()
        {
          public void run()
          {
            // select log entry, show it
            TableItem[] tableItems = widgetChangesLog.getItems();
            for (TableItem tableItem : widgetChangesLog.getItems())
            {
              if ((LogData)tableItem.getData() == data.selectedLogData1)
              {
                widgetChangesLog.setSelection(tableItem);
                widgetChangesLog.showSelection();
                break;
              }
            }

            // set changed files
            widgetFiles.removeAll();
            for (String fileName : data.selectedLogData1.fileNameList)
            {
              widgetFiles.add(fileName);
            }

            if (!data.selectedLogData1.fileNameList.isEmpty())
            {
              // select first file
              widgetFiles.select(0);
              data.fileData = repositoryTab.repository.getFileDataInstance(data.selectedLogData1.fileNameList.get(0));
            }
          }
        });
      }
    }
  }

  /** select revision
   * @param logData log data to select
   */
  private void selectRevision(LogData logData)
  {
    selectRevision(data.selectedLogData1,logData);
  }

  private void editCommitMessage(LogData logData)
  {
    /** dialog data
     */
    class Data
    {
      String message;

      Data()
      {
        this.message = null;
      }
    };

    final Data  data = new Data();
    final Shell subDialog;
    Composite   composite;
    Label       label;
    Button      button;

    // name dialog
    subDialog = Dialogs.openModal(dialog,"Edit commit message",300,300,new double[]{1.0,0.0},1.0);

    final StyledText widgetMessage;
    final Button     widgetSave;
    composite = Widgets.newComposite(subDialog);
    composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      widgetMessage = Widgets.newStyledText(composite,SWT.LEFT|SWT.BORDER|SWT.MULTI|SWT.WRAP|SWT.H_SCROLL|SWT.V_SCROLL);
      widgetMessage.setToolTipText("Commit message.\n"+
                                   Widgets.acceleratorToText(Settings.keyPrevCommitMessage)+"/"+Widgets.acceleratorToText(Settings.keyLastCommitMessage)+"/"+Widgets.acceleratorToText(Settings.keyFirstCommitMessage)+"/"+Widgets.acceleratorToText(Settings.keyLastCommitMessage)+"to select message from history,\n"+
                                   Widgets.acceleratorToText(Settings.keyFormatCommitMessage)+" to format message,\n"+
                                   Widgets.acceleratorToText(Settings.keyCommitMessageDone)+" to commit changed files."
                                  );
      widgetMessage.setText(StringUtils.join(logData.commitMessageList,widgetMessage.getLineDelimiter()));
      Widgets.layout(widgetMessage,0,0,TableLayoutData.NSWE);
    }

    // buttons
    composite = Widgets.newComposite(subDialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetSave = Widgets.newButton(composite,"Save");
      widgetSave.setToolTipText("Save new commit message.");
      Widgets.layout(widgetSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

//          Settings.geometryRepositoryList = subDialog.getSize();

          Dialogs.close(subDialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          Dialogs.close(subDialog,false);
        }
      });
    }

    // listeners

    // show dialog
    Dialogs.show(subDialog);

    // add commit message

    // run
    widgetMessage.setFocus();
    if ((Boolean)Dialogs.run(subDialog,false) && (data.message != null) && !data.message.isEmpty())
    {
Dprintf.dprintf("");
//      repositoryTab.repository.setCommitMessage();
    }
  }
}

/* end of file */
