/***********************************************************************\
*
* $Revision: 800 $
* $Date: 2012-01-28 10:49:16 +0100 (Sat, 28 Jan 2012) $
* $Author: trupp $
* Contents: command merge branch/commit
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.LinkedList;

// graphics
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Widget;

/****************************** Classes ********************************/

/** merge branch/commit command
 */
class CommandMerge
{
  /** dialog data
   */
  class Data
  {
    java.util.List<String>  branchNameList;
    java.util.List<LogData> logDataList;
    String                  branchName;                // branch to merge
    String                  revision;                  // commit to merge
    String[]                message;                   // commit message
    String                  lastMessage;               // last set commit message
    boolean                 messageEdited;             // true iff commit message edited
    boolean                 immediateCommitFlag;       // true for immediate commit

    Data()
    {
      this.branchNameList      = null;
      this.logDataList         = null;
      this.branchName          = null;
      this.revision            = null;
      this.message             = null;
      this.lastMessage         = "";
      this.messageEdited       = false;
      this.immediateCommitFlag = Settings.immediateCommit;
    }
  };

  // --------------------------- constants --------------------------------
  // user events
  private final int USER_EVENT_NEW_MESSAGE         = 0xFFFF+0;
  private final int USER_EVENT_ADD_REPLACE_MESSAGE = 0xFFFF+1;

  // --------------------------- variables --------------------------------

  // global variable references
  private final RepositoryTab        repositoryTab;
  private final Shell                shell;
  private final Display              display;
  private final Clipboard            clipboard;

  // dialog
  private final Data                 data = new Data();
  private final Shell                dialog;
  private final LinkedList<String[]> history;

  // widgets
  private final Button               widgetMergeBranch;
  private final Combo                widgetBranchNames;
  private final Button               widgetMergeCommit;
  private final TreeCombo            widgetRevisions;
  private final List                 widgetHistory;
  private final StyledText           widgetMessage;
  private final Button               widgetImmediateCommit;
  private final Button               widgetMerge;

  private Listener                   enableListener;
  private Listener                   userEventListener;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** set file mode command
   * @param shell shell
   * @param repositoryTab repository tab
   */
  CommandMerge(final Shell shell, final RepositoryTab repositoryTab)
  {
    Composite composite,subComposite;
    Label     label;
    Button    button;

    // initialize variables
    this.repositoryTab = repositoryTab;

    // get shell, display, clipboard
    this.shell     = shell;
    this.display   = shell.getDisplay();
    this.clipboard = new Clipboard(display);

    // get commit history
    history = CommitMessage.getHistory();

    // add files dialog
    dialog = Dialogs.openModal(shell,"Merge branch/commit",new double[]{1.0,0.0},1.0);

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{0.0,0.0,1.0,0.0,1.0,0.0},1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0}));
      Widgets.layout(subComposite,0,0,TableLayoutData.WE);
      {
        widgetMergeBranch = Widgets.newRadio(subComposite,"Branch name:");
        widgetMergeBranch.setEnabled(repositoryTab.repository.supportMergeBranch());
        widgetMergeBranch.setSelection(repositoryTab.repository.supportMergeBranch());
        Widgets.layout(widgetMergeBranch,0,0,TableLayoutData.W);

        widgetBranchNames = Widgets.newCombo(subComposite);
        widgetBranchNames.setToolTipText("Branch to merge. For CVS this is a branch name, for SVN, HG and GIT this is a directory name, usually below 'branches' or 'tags'.\n\nNote: HG use the term 'branch' different. Nevertheless a 'branch' follow here the common understanding: a branch is a fork of the checked-in files.");
        widgetBranchNames.setEnabled(widgetMergeBranch.getSelection());
        Widgets.layout(widgetBranchNames,0,1,TableLayoutData.WE);

        widgetMergeCommit = Widgets.newRadio(subComposite,"Revision:");
        widgetMergeCommit.setEnabled(repositoryTab.repository.supportMergeCommit());
        widgetMergeCommit.setSelection(repositoryTab.repository.supportMergeCommit() && !repositoryTab.repository.supportMergeBranch());
        Widgets.layout(widgetMergeCommit,1,0,TableLayoutData.W);

        widgetRevisions = new TreeCombo(subComposite,100,TreeCombo.DEFAULT_HEIGHT,SWT.NONE);
        widgetRevisions.setToolTipText("Revision to merge. For CVS this is a tag name, for SVN, HG and GIT this is a commit number, a hash value, or a symbolic name.");
        widgetRevisions.addTreeColumn("Revision",SWT.RIGHT, 80, false);
        widgetRevisions.addTreeColumn("Date",    SWT.LEFT,  150,true );
        widgetRevisions.addTreeColumn("Author",  SWT.LEFT,  80, true );
        widgetRevisions.addTreeColumn("Message", SWT.LEFT,  0,  true );
        widgetRevisions.setText(repositoryTab.repository.getLastRevision());
        widgetRevisions.setEnabled(widgetMergeCommit.getSelection());
        widgetRevisions.setLayout(new double[]{0.0,0.25,0.25,0.5});
        Widgets.layout(widgetRevisions,1,1,TableLayoutData.WE);
      }

      label = Widgets.newLabel(composite,"History:");
      Widgets.layout(label,1,0,TableLayoutData.W);

      widgetHistory = Widgets.newList(composite);
      widgetHistory.setToolTipText("Commit message history.");
      widgetHistory.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetHistory,2,0,TableLayoutData.NSWE);

      label = Widgets.newLabel(composite,"Message:");
      Widgets.layout(label,3,0,TableLayoutData.W);

      widgetMessage = Widgets.newStyledText(composite,SWT.LEFT|SWT.BORDER|SWT.MULTI|SWT.WRAP|SWT.H_SCROLL|SWT.V_SCROLL);
      widgetMessage.setToolTipText("Commit message. Use:\n"+
                                   Widgets.acceleratorToText(Settings.keyPrevCommitMessage)+"/"+Widgets.acceleratorToText(Settings.keyLastCommitMessage)+"/"+Widgets.acceleratorToText(Settings.keyFirstCommitMessage)+"/"+Widgets.acceleratorToText(Settings.keyLastCommitMessage)+"to select message from history,\n"+
                                   Widgets.acceleratorToText(Settings.keyFormatCommitMessage)+" to format message,\n"+
                                   Widgets.acceleratorToText(Settings.keyCommitMessageDone)+" to merge."
                                  );
      widgetMessage.setEnabled(Settings.immediateCommit);
      Widgets.layout(widgetMessage,4,0,TableLayoutData.NSWE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetMessage,data)
      {
        public void modified(Control control)
        {
          if (!control.isDisposed())
          {
            control.setEnabled(data.immediateCommitFlag);
            control.setBackground(data.immediateCommitFlag ? null : Onzen.COLOR_GRAY);
            if (data.immediateCommitFlag) control.setFocus();
          }
        }
      });

      widgetImmediateCommit = Widgets.newCheckbox(composite,"immediate commit");
      widgetImmediateCommit.setToolTipText("Select this checkbox to commit added files immediately.");
      widgetImmediateCommit.setSelection(Settings.immediateCommit);
      Widgets.layout(widgetImmediateCommit,5,0,TableLayoutData.E);
      widgetImmediateCommit.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.immediateCommitFlag = widgetImmediateCommit.getSelection();
          Widgets.modified(data);
        }
      });
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetMerge = Widgets.newButton(composite,"Merge");
      widgetMerge.setToolTipText("Merge branch/commit.");
      widgetMerge.setEnabled(false);
      Widgets.layout(widgetMerge,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetMerge.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.branchName          = widgetMergeBranch.getSelection() ? widgetBranchNames.getText().trim() : null;
          data.revision            = widgetMergeCommit.getSelection() ? widgetRevisions.getText()          : null;
          data.message             = widgetMessage.getText().trim().split(widgetMessage.getLineDelimiter());
          data.immediateCommitFlag = widgetImmediateCommit.getSelection();

          if (merge())
          {
            Settings.geometryMerge   = dialog.getSize();
            Settings.immediateCommit = widgetImmediateCommit.getSelection();

            Dialogs.close(dialog,true);
          }
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetMergeBranch.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Button widget = (Button)selectionEvent.widget;

        widgetBranchNames.setEnabled(widget.getSelection());
        widgetRevisions.setEnabled(!widget.getSelection());
      }
    });
    widgetBranchNames.addModifyListener(new ModifyListener()
    {
      public void modifyText(ModifyEvent modifyEvent)
      {
        widgetMerge.setEnabled(   (widgetMergeBranch.getSelection() && !widgetBranchNames.getText().trim().isEmpty())
                               || (widgetMergeCommit.getSelection() && !widgetRevisions.getText().trim().isEmpty())
                              );
      }
    });
    widgetMergeCommit.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Button widget = (Button)selectionEvent.widget;

        widgetBranchNames.setEnabled(!widget.getSelection());
        widgetRevisions.setEnabled(widget.getSelection());
      }
    });
    widgetRevisions.addSelectionListener(new SelectionListener()
    {
      @Override
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      @Override
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        TreeCombo widget = (TreeCombo)selectionEvent.widget;

        LogData logData = widget.getSelection();
        if (logData != null)
        {
          widget.setText(logData.revisionId);
        }
      }
    });
    widgetRevisions.addModifyListener(new ModifyListener()
    {
      public void modifyText(ModifyEvent modifyEvent)
      {
Dprintf.dprintf("");
        widgetMerge.setEnabled(   (widgetMergeBranch.getSelection() && !widgetBranchNames.getText().trim().isEmpty())
                               || (widgetMergeCommit.getSelection() && !widgetRevisions.getText().trim().isEmpty())
                              );
      }
    });
    widgetHistory.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        List widget = (List)selectionEvent.widget;

        int i = widget.getSelectionIndex();
        if (i >= 0)
        {
          Widgets.notify(dialog,USER_EVENT_NEW_MESSAGE,StringUtils.join(history.get(i),widgetMessage.getLineDelimiter()).trim());
        }
      }
    });
    widgetHistory.addMouseListener(new MouseListener()
    {
      public void mouseDoubleClick(MouseEvent mouseEvent)
      {
        List widget = (List)mouseEvent.widget;

        int i = widget.getSelectionIndex();
        if (i >= 0)
        {
          Widgets.notify(dialog,USER_EVENT_NEW_MESSAGE,StringUtils.join(history.get(i),widgetMessage.getLineDelimiter()).trim());
        }
      }
      public void mouseDown(MouseEvent mouseEvent)
      {
      }
      public void mouseUp(MouseEvent mouseEvent)
      {
      }
    });
    widgetMessage.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        int i = widgetHistory.getSelectionIndex();
        if (i < 0) i = history.size();

        if      (Widgets.isAccelerator(keyEvent,Settings.keyPrevCommitMessage))
        {
          // previous history entry
          if (i > 0)
          {
            widgetHistory.setSelection(i-1);
            display.update();
            widgetHistory.showSelection();

            Widgets.notify(dialog,USER_EVENT_NEW_MESSAGE,StringUtils.join(history.get(i-1),widgetMessage.getLineDelimiter()).trim());
          }
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyNextCommitMessage))
        {
          // next history entry
          if (i < history.size()-1)
          {
            widgetHistory.setSelection(i+1);
            display.update();
            widgetHistory.showSelection();

            Widgets.notify(dialog,USER_EVENT_NEW_MESSAGE,StringUtils.join(history.get(i+1),widgetMessage.getLineDelimiter()).trim());
          }
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyFirstCommitMessage))
        {
          // first history entry
          if (history.size() > 0)
          {
            widgetHistory.setSelection(0);
            display.update();
            widgetHistory.showSelection();

            Widgets.notify(dialog,USER_EVENT_NEW_MESSAGE,StringUtils.join(history.getFirst(),widgetMessage.getLineDelimiter()).trim());
          }
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyLastCommitMessage))
        {
          // last history entry
          if (history.size() > 0)
          {
            widgetHistory.setSelection(history.size()-1);
            display.update();
            widgetHistory.showSelection();

            Widgets.notify(dialog,USER_EVENT_NEW_MESSAGE,StringUtils.join(history.getLast(),widgetMessage.getLineDelimiter()).trim());
          }
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyFormatCommitMessage))
        {
          // format message
          String oldMessage = widgetMessage.getText().trim();
          String newMessage = CommitMessage.format(oldMessage,widgetMessage.getLineDelimiter());
          if (!newMessage.equals(oldMessage))
          {
            widgetMessage.setText(newMessage);
            Widgets.setFocus(widgetMessage);
          }
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyCommitMessageDone))
        {
          // invoke merge branch-button
          Widgets.invoke(widgetMerge);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'a'))
        {
          widgetMessage.setSelection(0,widgetMessage.getCharCount());
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'c'))
        {
          Widgets.setClipboard(clipboard,widgetMessage.getSelectionText());
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    });

    enableListener = new Listener()
    {
      public void handleEvent(Event event)
      {
        // auto-enable branch names/revision on click
        if (event.widget instanceof Control)
        {
          Control control = (Control)event.widget;
          Point   p;
          Point   size;

          p    = widgetBranchNames.toControl(control.toDisplay(event.x,event.y));
          size = widgetBranchNames.getSize();
          if      (   repositoryTab.repository.supportMergeBranch()
                   && (p.x >= 0) && (p.x <= size.x)
                   && (p.y >= 0) && (p.y <= size.y)
                  )
          {
            widgetMergeBranch.setSelection(true);
            widgetMergeCommit.setSelection(false);
            widgetBranchNames.setEnabled(true);
            widgetRevisions.setEnabled(false);
            Widgets.setFocus(widgetBranchNames);
            return;
          }

          p    = widgetRevisions.toControl(control.toDisplay(event.x,event.y));
          size = widgetRevisions.getSize();
          if      (   repositoryTab.repository.supportMergeCommit()
                   && (p.x >= 0) && (p.x <= size.x)
                   && (p.y >= 0) && (p.y <= size.y)
                  )
          {
            widgetMergeBranch.setSelection(false);
            widgetMergeCommit.setSelection(true);
            widgetBranchNames.setEnabled(false);
            widgetRevisions.setEnabled(true);
            Widgets.setFocus(widgetRevisions);
            return;
          }
        }
      }
    };
    display.addFilter(SWT.MouseDown,enableListener);

    userEventListener = new Listener()
    {
      public void handleEvent(Event event)
      {
        String currentMessage = widgetMessage.getText().trim();

        String newMessage;
        if (   (!currentMessage.isEmpty() && !currentMessage.equals(data.lastMessage))
            || (!currentMessage.isEmpty() && (event.type == USER_EVENT_ADD_REPLACE_MESSAGE))
            || data.messageEdited
           )
        {
          switch (Dialogs.select(dialog,"Confirmation","Replace or add message to existing commit message?",new String[]{"Replace","Add","Cancel"},1))
          {
            case 0:
              // replace
              newMessage         = event.text;
              data.lastMessage   = newMessage;
              data.messageEdited = false;
              break;
            case 1:
              // add
              newMessage         = CommitMessage.format(currentMessage+widgetMessage.getLineDelimiter()+event.text,widgetMessage.getLineDelimiter());
              data.messageEdited = true;
              break;
            default:
              return;
          }
        }
        else
        {
          newMessage       = event.text;
          data.lastMessage = newMessage;
        }

        widgetMessage.setText(newMessage);
        Widgets.setFocus(widgetMessage);
      }
    };
    dialog.addListener(USER_EVENT_NEW_MESSAGE,userEventListener);
    dialog.addListener(USER_EVENT_ADD_REPLACE_MESSAGE,userEventListener);

    Widgets.setNextFocus(widgetBranchNames,
                         widgetRevisions,
                         widgetMessage,
                         widgetMerge
                        );

    // show dialog
    Dialogs.show(dialog,Settings.geometryMerge,Settings.setWindowLocation);

    // add history
    if (!widgetHistory.isDisposed())
    {
      for (String[] lines : history)
      {
        widgetHistory.add(StringUtils.join(lines,", "));
      }
      widgetHistory.setSelection(widgetHistory.getItemCount()-1);
      display.update();
      widgetHistory.showSelection();
      widgetHistory.deselectAll();
    }

    // add existing branch names
    Background.run(new BackgroundRunnable()
    {
      public void run()
      {
        if (!dialog.isDisposed())
        {
          // get branch names
          Widgets.setCursor(dialog,Onzen.CURSOR_WAIT);
          repositoryTab.setStatusText("Get branch names...");
          try
          {
            data.branchNameList = repositoryTab.repository.getBranchTagNames();
            data.logDataList    = repositoryTab.repository.getLog();
          }
          catch (RepositoryException exception)
          {
            final String exceptionMessage = exception.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(dialog,String.format("Getting branch names fail: %s",exceptionMessage));
              }
            });
            Onzen.printStacktrace(exception);
            return;
          }
          catch (Throwable throwable)
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                dialog.dispose();
              }
            });
            Onzen.printInternalError(throwable);
            return;
          }
          finally
          {
            repositoryTab.clearStatusText();
            Widgets.resetCursor(dialog);
          }
        }

        if (!display.isDisposed())
        {
          // add branch names to widget
          display.syncExec(new Runnable()
          {
            public void run()
            {
              if (!widgetBranchNames.isDisposed())
              {
                for (String branchName : data.branchNameList)
                {
                  widgetBranchNames.add(branchName);
                }
              }
              if (!widgetRevisions.isDisposed())
              {
                widgetRevisions.addTreeItem(null,repositoryTab.repository.getLastRevision(),"","","");
                for (LogData logData : data.logDataList)
                {
                  widgetRevisions.addTreeItem(logData,
                                              logData.revisionId,
                                              Onzen.DATETIME_FORMAT.format(logData.date),
                                              logData.author,
                                              logData.getFirstCommitMessage()
                                             );
                }
              }
            }
          });
        }
      }
    });
  }

  /** run dialog
   */
  public void run()
  {
    Widgets.setFocus(widgetBranchNames);
    Dialogs.run(dialog);
    display.removeFilter(SWT.MouseDown,enableListener);
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "CommandMerge {}";
  }

  //-----------------------------------------------------------------------

  /** merge branch/commit
   * @return true iff branch/commit merged
   */
  private boolean merge()
  {
    if      (data.branchName != null)
    {
      final BusyDialog busyDialog = new BusyDialog(dialog,
                                                   "Merge branch",
                                                   "Merge branch '"+
                                                   data.branchName+
                                                   "':",
                                                   BusyDialog.TEXT0|BusyDialog.LIST|BusyDialog.AUTO_ANIMATE
                                                  );

      repositoryTab.setStatusText("Merge branch '"+data.branchName+"'...");
      CommitMessage commitMessage = null;
      try
      {
        // merge and add message to history
        if (data.immediateCommitFlag)
        {
          commitMessage = new CommitMessage(data.message);
          commitMessage.addToHistory();
        }

        // merge branch
        repositoryTab.repository.mergeBranch(data.branchName);

        // free resources
        if (commitMessage != null) { commitMessage.done(); commitMessage = null; }
      }
      catch (RepositoryException exception)
      {
        final String   exceptionMessage = exception.getMessage();
        final String[] extendedMessage  = exception.getExtendedMessage();
        display.syncExec(new Runnable()
        {
          public void run()
          {
            busyDialog.close();

            Dialogs.error(dialog,extendedMessage,"Merge branch fail:\n%s",exceptionMessage);
          }
        });
        Onzen.printStacktrace(exception);
        return false;
      }
      finally
      {
        if (commitMessage != null) commitMessage.done();
        display.syncExec(new Runnable()
        {
          public void run()
          {
            busyDialog.close();
          }
        });
        repositoryTab.clearStatusText();
      }
    }
    else if (data.revision!= null)
    {
      final BusyDialog busyDialog = new BusyDialog(dialog,
                                                   "Merge commit",
                                                   "Merge commit '"+
                                                   data.revision+
                                                   "':",
                                                   BusyDialog.TEXT0|BusyDialog.LIST|BusyDialog.AUTO_ANIMATE
                                                  );

      repositoryTab.setStatusText("Merge commit '"+data.revision+"'...");
      CommitMessage commitMessage = null;
      try
      {
        // merge and add message to history
        if (data.immediateCommitFlag)
        {
          commitMessage = new CommitMessage(data.message);
          commitMessage.addToHistory();
        }

        // merge commit
        repositoryTab.repository.mergeCommit(data.revision);

        // free resources
        if (commitMessage != null) { commitMessage.done(); commitMessage = null; }
      }
      catch (RepositoryException exception)
      {
        final String   exceptionMessage = exception.getMessage();
        final String[] extendedMessage  = exception.getExtendedMessage();
        display.syncExec(new Runnable()
        {
          public void run()
          {
            busyDialog.close();

            Dialogs.error(dialog,extendedMessage,"Merge commit fail:\n%s",exceptionMessage);
          }
        });
        Onzen.printStacktrace(exception);
        return false;
      }
      finally
      {
        if (commitMessage != null) commitMessage.done();
        display.syncExec(new Runnable()
        {
          public void run()
          {
            busyDialog.close();
          }
        });
        repositoryTab.clearStatusText();
      }
    }

    // update file states
    repositoryTab.asyncUpdateAllFileStates();

    return true;
  }
}

/* end of file */
