/***********************************************************************\
*
* $Revision: 2094 $
* $Date: 2017-07-04 17:02:17 +0200 (Tue, 04 Jul 2017) $
* $Author: torsten $
* Contents: command store patch in internal database
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.util.Date;
import java.util.LinkedHashSet;
import java.util.Set;

// graphics
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Widget;

/****************************** Classes ********************************/

/** store patch command
 */
class CommandStorePatch
{
  /** dialog data
   */
  class Data
  {
    String[]              revisionNames;        // revision names
    String[]              lines;                // patch lines
    String[]              linesNoWhitespaces;   // patch lines (without whitespaces)
    String                summary;              // summary for patch
    String[]              message;              // message for patch
    LinkedHashSet<String> testSet;              // tests for patch

    Data()
    {
      this.revisionNames      = null;
      this.lines              = null;
      this.linesNoWhitespaces = null;
      this.summary            = null;
      this.message            = null;
      this.testSet            = null;
    }
  };

  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------
  public String                summary;
  public String[]              message;
  public LinkedHashSet<String> testSet;

  // global variable references
  private final RepositoryTab  repositoryTab;
  private final Patch          patch;
  private final Date           date;
  private final Display        display;

  // dialog
  private final Data           data = new Data();
  private final Shell          dialog;

  // widgets
  private final Text           widgetPatch;
  private final Text           widgetSummary;
  private final StyledText     widgetMessage;
  private final Button         widgetStore;
  private final Button         widgetCancel;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** store patch command
   * @param shell shell
   * @param repositoryTab repository tab
   * @param fileDataSet file data set
   * @param patch patch
   * @param summary summary text
   * @param message message text lines
   */
  CommandStorePatch(final Shell shell, final RepositoryTab repositoryTab, Set<FileData> fileDataSet, final Patch patch, String summary, String[] message)
  {
    Composite         composite,subComposite,subSubComposite,subSubSubComposite;
    Label             label;
    TabFolder         tabFolder;
    Button            button;
    ScrolledComposite scrolledComposite;
    SelectionListener selectionListener;
    Listener          listener;

    // initialize variables
    this.summary       = summary;
    this.message       = message;
    this.testSet       = new LinkedHashSet<String>();
    this.repositoryTab = repositoryTab;
    this.patch         = patch;
    this.date          = new Date();

    // get display
    display = shell.getDisplay();

    // add files dialog
    dialog = Dialogs.openModal(shell,"Store patch",new double[]{1.0,0.0},1.0);

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{0.0,1.0,1.0},1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Patch:");
      Widgets.layout(label,0,0,TableLayoutData.W);

      widgetPatch = Widgets.newText(composite,SWT.LEFT|SWT.BORDER|SWT.MULTI|SWT.H_SCROLL|SWT.V_SCROLL|SWT.READ_ONLY);
      widgetPatch.setFont(Onzen.FONT_CHANGES);
      widgetPatch.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetPatch,1,0,TableLayoutData.NSWE);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(new double[]{0.0,1.0},new double[]{0.0,1.0}));
      Widgets.layout(subComposite,2,0,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(subComposite,"Summary:");
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetSummary = Widgets.newText(subComposite);
        widgetSummary.setText(summary);
        Widgets.layout(widgetSummary,0,1,TableLayoutData.WE);

        label = Widgets.newLabel(subComposite,"Message:");
        Widgets.layout(label,1,0,TableLayoutData.NW);

        widgetMessage = Widgets.newStyledText(subComposite,SWT.LEFT|SWT.BORDER|SWT.MULTI|SWT.WRAP|SWT.H_SCROLL|SWT.V_SCROLL);
        widgetMessage.setText(StringUtils.join(message,widgetMessage.getLineDelimiter()));
        Widgets.layout(widgetMessage,1,1,TableLayoutData.NSWE);
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0}));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetStore = Widgets.newButton(composite,"Store");
      Widgets.layout(widgetStore,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetStore.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.summary = widgetSummary.getText();
          data.message = widgetMessage.getText().trim().split(widgetMessage.getLineDelimiter());
//???
Dprintf.dprintf("");
//          data.testSet = widgetSummary.getText();

          Settings.geometryStorePatch = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      widgetCancel = Widgets.newButton(composite,"Cancel");
      Widgets.layout(widgetCancel,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetCancel.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners

    // show dialog
    Dialogs.show(dialog,Settings.geometryStorePatch,Settings.setWindowLocation);

    // update
    widgetPatch.setText(StringUtils.join(patch.getLines(),widgetPatch.getLineDelimiter()));
  }

  /** store patch command
   * @param shell shell
   * @param repositoryTab repository tab
   * @param fileDataSet file data set
   * @param patch patch
   */
  CommandStorePatch(Shell shell, RepositoryTab repositoryTab, Set<FileData> fileDataSet, Patch patch)
  {
    this(shell,repositoryTab,fileDataSet,patch,"",null);
  }

  /** run dialog
   */
  public void run()
  {
    if (!dialog.isDisposed())
    {
      Widgets.setFocus(widgetSummary);
      if ((Boolean)Dialogs.run(dialog,false))
      {
        summary = data.summary;
        message = data.message;
//???
Dprintf.dprintf("");
//          data.testSet = widgetSummary.getText();
      }
    }
  }

  /** run and wait for dialog
   */
  public boolean execute()
  {
    if (!dialog.isDisposed())
    {
      Widgets.setFocus(widgetSummary);
      if ((Boolean)Dialogs.run(dialog,false))
      {
        summary = data.summary;
        message = data.message;
//???
Dprintf.dprintf("");
//          data.testSet = widgetSummary.getText();

        return true;
      }
      else
      {
        return false;
      }
    }
    else
    {
      return false;
    }
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "CommandStorePatch {}";
  }

  //-----------------------------------------------------------------------
}

/* end of file */
