/***********************************************************************\
*
* $Revision: 2024 $
* $Date: 2017-04-17 03:42:27 +0200 (Mon, 17 Apr 2017) $
* $Author: torsten $
* Contents: command line option
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/

/****************************** Classes ********************************/

/** option
 */
public class Option
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------
  public String              name;
  public String              shortName;
  public Options.Types       type;
  public String              fieldName;
  public Object[]            units;
  public OptionEnumeration[] enumeration;
  public OptionSpecial       special;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create option
   * @param name name
   * @param shortName short name
   * @param type option type
   * @param fieldName field name to store value
   * @param units units (array {name,factor})
   */
  public Option(String name, String shortName, Options.Types type, String fieldName, Object[] units)
  {
    assert (units == null) || (units.length%2 == 0);

    this.name        = name;
    this.shortName   = shortName;
    this.type        = type;
    this.fieldName   = fieldName;
    this.units       = units;
  }

  /** create option
   * @param name name
   * @param shortName short name
   * @param type option type
   * @param fieldName field name to store value
   */
  public Option(String name, String shortName, Options.Types type, String fieldName)
  {
    this(name,shortName,type,fieldName,(Object[])null);
  }

  /** create enumeration option
   * @param name name
   * @param shortName short name
   * @param type option type
   * @param fieldName field name to store value
   * @param enumeration enumeration values
   */
  public Option(String name, String shortName, Options.Types type, String fieldName, OptionEnumeration[] enumeration)
  {
    this.name        = name;
    this.shortName   = shortName;
    this.type        = type;
    this.fieldName   = fieldName;
    this.enumeration = enumeration;
  }

  /** create special option
   * @param name name
   * @param shortName short name
   * @param type option type
   * @param fieldName field name to store value
   * @param enumeration enumeration values
   */
  public Option(String name, String shortName, Options.Types type, String fieldName, OptionSpecial special)
  {
    this.name      = name;
    this.shortName = shortName;
    this.type      = type;
    this.fieldName = fieldName;
    this.special   = special;
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "{"+name+", "+shortName+", "+type+", "+fieldName+"}";
  }
}

/* end of file */
