/***********************************************************************\
*
* $Revision: 2075 $
* $Date: 2017-05-12 10:22:47 +0200 (Fri, 12 May 2017) $
* $Author: torsten $
* Contents: set preferences
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Constructor;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.HashSet;
import java.util.Properties;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import javax.mail.Authenticator;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.PasswordAuthentication;
import javax.mail.Session;
import javax.mail.Transport;

// graphics
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.ColorDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.FontDialog;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.ScrollBar;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;

/****************************** Classes ********************************/

/** set preferences
 */
class Preferences
{
  /** dialog data
   */
  class Data
  {

    Data()
    {
    }
  };

  /** SVN command
   */
  class SVNCommand
  {
    public String version;
    public String command;

    SVNCommand(String version, String command)
    {
      this.version = version;
      this.command = command;
    }

    SVNCommand()
    {
      this("","");
    }

    /** convert to string
    * @return string
    */
    public String toString()
    {
      return "SVNCommand {"+version+", "+command+"}";
    }
  }

  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------

  // global variable references
  private final Shell         shell;
  private final Display       display;

  // dialog
  private final Data          data = new Data();
  private final Shell         dialog;

  // widgets
  private final Table         widgetKeys;
  private final Table         widgetColors;
  private final Table         widgetFonts;

  private final Table         widgetEditors;
  private final Table         widgetShellCommands;
  private final Text          widgetCommandLineMail;
  private final Text          widgetCommandLineMailAttachment;

  private final Text          widgetMailSMTPHost;
  private final Spinner       widgetMailSMTPPort;
  private final Button        widgetMailSMTPSSL;
  private final Text          widgetMailLogin;
  private final Text          widgetMailPassword;
  private final Text          widgetMailFrom;

  private final Text          widgetReviewServerHost;
  private final Text          widgetReviewServerLogin;
  private final Text          widgetCommandLinePostReviewServer;
  private final Text          widgetCommandLineUpdateReviewServer;

  private final Text          widgetCVSCommand;
  private final Button        widgetCVSPruneEmptyDirectories;

  private final Table         widgetSVNCommands;
  private final Text          widgetSVNDiffCommand;
  private final Text          widgetSVNDiffCommandOptions;
  private final Text          widgetSVNDiffCommandOptionsIgnoreWhitespaces;
  private final Button        widgetSVNAlwaysTrustServerCertificate;

  private final Text          widgetHGCommand;
  private final Text          widgetHGDiffCommand;
  private final Text          widgetHGDiffCommandOptions;
  private final Text          widgetHGDiffCommandOptionsIgnoreWhitespaces;
  private final Button        widgetHGUseForestExtension;
  private final Button        widgetHGUseQueueExtension;
  private final Button        widgetHGUpdateWithFetch;
  private final Button        widgetHGSafeUpdate;
  private final Button        widgetHGSingleLineCommitMessages;
  private final Spinner       widgetHGSingleLineMaxCommitMessageLength;
  private final Button        widgetHGRelativePatchPaths;

  private final Text          widgetGITCommand;
  private final Button        widgetGITShortCommitNumbers;
  private final Button        widgetGITStashOnUpdate;
  private final Button        widgetGITAutoPullSubModules;

  private final List          widgetStoredPasswords;
  private final List          widgetRepositoryHistoryURLs;
  private final List          widgetRepositoryURLs;

  private final Button        widgetEOLAuto;
  private final Button        widgetEOLUnix;
  private final Button        widgetEOLMac;
  private final Button        widgetEOLWindows;
  private final Button        widgetCheckTABs;
  private final Button        widgetCheckTrailingWhitespaces;
  private final Button        widgetAutoTrimTrailingWhitespaces;
  private final List          widgetSkipWhitespaceCheckFilePatterns;
  private final List          widgetHiddenFilePatterns;
  private final List          widgetHiddenDirectoryPatterns;

  private final Text          widgetShowDiffCommand;
  private final Text          widgetTmpDirectory;
  private final Text          widgetBackupFileSuffix;
  private final Text          widgetDateFormat;
  private final Text          widgetTimeFormat;
  private final Text          widgetDateTimeFormat;
  private final Spinner       widgetConvertSpacesPerTAB;
  private final Spinner       widgetMaxBackgroundTasks;
  private final Spinner       widgetMaxMessageHistory;
  private final Text          widgetMessageBroadcastAddress;
  private final Spinner       widgetMessageBroadcastPort;
  private final List          widgetAutoSummaryPatterns;
  private final Button        widgetSetWindowLocation;
  private final Table         widgetShowFlags;

  private final Button        widgetButtonSave;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** view command
   * @param shell shell
   * @param onzen Onzen instance
   */
  Preferences(final Shell shell, final Onzen onzen)
  {
    TabFolder   tabFolder,subTabFolder;
    Composite   composite,subComposite,subSubComposite,subSubSubComposite;
    Label       label;
    Button      button;
    Listener    listener;
    TableItem   tableItem;
    TableColumn tableColumn;

    // initialize variables
    this.shell = shell;

    // get display
    display = shell.getDisplay();

    // add files dialog
    dialog = Dialogs.openModal(shell,"Preferences",new double[]{1.0,0.0},1.0);

    // create tab
    tabFolder = Widgets.newTabFolder(dialog);
    Widgets.layout(tabFolder,0,0,TableLayoutData.NSWE);
    {
      composite = Widgets.addTab(tabFolder,"Keys");
      composite.setLayout(new TableLayout(1.0,1.0,2));
      Widgets.layout(composite,0,0,TableLayoutData.NSWE);
      {
        widgetKeys = Widgets.newTable(composite);
        widgetKeys.setToolTipText("Keyboard short cuts list.");
        Widgets.layout(widgetKeys,0,0,TableLayoutData.NSWE);
        tableColumn = Widgets.addTableColumn(widgetKeys,0,"Name",SWT.LEFT,200,true);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        tableColumn = Widgets.addTableColumn(widgetKeys,1,"Key", SWT.LEFT,100,true);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        widgetKeys.addMouseListener(new MouseListener()
        {
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
            Table widget = (Table)mouseEvent.widget;

            int index = widget.getSelectionIndex();
            if (index >= 0)
            {
              TableItem tableItem = widget.getItem(index);
              String    name      = tableItem.getText(0);
              int[]     key       = (int[])tableItem.getData();

              key[0] = editKey(name,key[0]);
              Widgets.updateTableItem(widgetKeys,key,null,Widgets.menuAcceleratorToText(key[0]));
            }
          }
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          public void mouseUp(MouseEvent mouseEvent)
          {
          }
        });
        addKeys();
      }

      composite = Widgets.addTab(tabFolder,"Colors");
      composite.setLayout(new TableLayout(1.0,1.0,2));
      Widgets.layout(composite,0,1,TableLayoutData.NSWE);
      {
        widgetColors = Widgets.newTable(composite);
        widgetColors.setToolTipText("Colors list.");
        Widgets.layout(widgetColors,0,0,TableLayoutData.NSWE);
        tableColumn = Widgets.addTableColumn(widgetColors,0,"Name", SWT.LEFT,320,true);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        tableColumn = Widgets.addTableColumn(widgetColors,1,"Foreground",SWT.LEFT,100,true);
        tableColumn = Widgets.addTableColumn(widgetColors,2,"Background",SWT.LEFT,100,true);
        widgetColors.addMouseListener(new MouseListener()
        {
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
            Table widget = (Table)mouseEvent.widget;

            int index = widget.getSelectionIndex();
            if (index >= 0)
            {
              TableItem      tableItem = widget.getItem(index);
              String         name      = tableItem.getText(0);
              Settings.Color color     = (Settings.Color)tableItem.getData();

              if (editColor(name,color))
              {
                if (color.foreground != null) Widgets.setTableItemColor(widgetColors,color,1,new Color(null,color.foreground));
                if (color.background != null) Widgets.setTableItemColor(widgetColors,color,2,new Color(null,color.background));
                widgetColors.deselectAll();
              }
            }
          }
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          public void mouseUp(MouseEvent mouseEvent)
          {
          }
        });
        addColors();
      }

      composite = Widgets.addTab(tabFolder,"Fonts");
      composite.setLayout(new TableLayout(1.0,1.0,2));
      Widgets.layout(composite,0,2,TableLayoutData.NSWE);
      {
        widgetFonts = Widgets.newTable(composite);
        widgetFonts.setToolTipText("Fonts list.");
        Widgets.layout(widgetFonts,0,0,TableLayoutData.NSWE);
        tableColumn = Widgets.addTableColumn(widgetFonts,0,"Name",SWT.LEFT,200,true);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        tableColumn = Widgets.addTableColumn(widgetFonts,1,"Font",SWT.LEFT,100,true);
        widgetFonts.addMouseListener(new MouseListener()
        {
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
            Table widget = (Table)mouseEvent.widget;

            int index = widget.getSelectionIndex();
            if (index >= 0)
            {
              TableItem tableItem = widget.getItem(index);
              String    name      = tableItem.getText(0);
              FontData  fontData  = (FontData)tableItem.getData();

              FontDialog fontDialog = new FontDialog(dialog);
              fontDialog.setFontList(new FontData[]{fontData});
              FontData newFontData = fontDialog.open();
              if (newFontData != null)
              {
                fontData.setName(newFontData.getName());
                fontData.setHeight(newFontData.getHeight());
                fontData.setStyle(newFontData.getStyle());
                Widgets.updateTableItem(widgetFonts,fontData,null,Widgets.fontDataToText(fontData));
                Widgets.setTableItemFont(widgetFonts,fontData,1,fontData);
              }
            }
          }
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          public void mouseUp(MouseEvent mouseEvent)
          {
          }
        });
        addFonts();
      }

      composite = Widgets.addTab(tabFolder,"Editors");
      composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,2));
      Widgets.layout(composite,0,3,TableLayoutData.NSWE);
      {
        widgetEditors = Widgets.newTable(composite);
        widgetEditors.setToolTipText("Editor list.");
        Widgets.layout(widgetEditors,0,0,TableLayoutData.NSWE);
        tableColumn = Widgets.addTableColumn(widgetEditors,0,"Name",     SWT.LEFT,100,false);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        tableColumn = Widgets.addTableColumn(widgetEditors,1,"Mime type",SWT.LEFT,200,false);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        tableColumn = Widgets.addTableColumn(widgetEditors,2,"File name",SWT.LEFT,100,false);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        tableColumn = Widgets.addTableColumn(widgetEditors,3,"Command",  SWT.LEFT,300,true );
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        widgetEditors.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
            Table widget = (Table)selectionEvent.widget;

            int index = widget.getSelectionIndex();
            if (index >= 0)
            {
              Settings.Editor editor = (Settings.Editor)(widget.getItem(index).getData());

              if (editEditor(editor,"Edit editor","Save"))
              {
                Widgets.updateTableItem(widgetEditors,editor,editor.name,editor.mimeType,editor.fileName,editor.commandLine);
              }
            }
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
          }
        });
        for (Settings.Editor editor : Settings.editors)
        {
          Widgets.addTableItem(widgetEditors,editor.clone(),editor.name,editor.mimeType,editor.fileName,editor.commandLine);
        }

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(null,null));
        Widgets.layout(subComposite,1,0,TableLayoutData.E);
        {
          button = Widgets.newButton(subComposite,"Add"+"\u2026");
          Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              Settings.Editor editor = new Settings.Editor();

              if (editEditor(editor,"Add editor","Add"))
              {
                Widgets.addTableItem(widgetEditors,editor,editor.name,editor.mimeType,editor.fileName,editor.commandLine);
              }
            }
          });

          button = Widgets.newButton(subComposite,"Remove");
          Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              int index = widgetEditors.getSelectionIndex();
              if (index >= 0)
              {
                widgetEditors.remove(index);
              }
            }
          });
        }
      }

      composite = Widgets.addTab(tabFolder,"Shell commands");
      composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,2));
      Widgets.layout(composite,0,3,TableLayoutData.NSWE);
      {
        widgetShellCommands = Widgets.newTable(composite);
        widgetShellCommands.setToolTipText("Shell command list.");
        Widgets.layout(widgetShellCommands,0,0,TableLayoutData.NSWE);
        tableColumn = Widgets.addTableColumn(widgetShellCommands,0,"Name",    SWT.LEFT, 200,true);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        tableColumn = Widgets.addTableColumn(widgetShellCommands,1,"Command", SWT.LEFT, 200,true);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
        tableColumn = Widgets.addTableColumn(widgetShellCommands,2,"Exitcode",SWT.RIGHT, 20,true);
        tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_INT);
        widgetShellCommands.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
            Table widget = (Table)selectionEvent.widget;

            int index = widget.getSelectionIndex();
            if (index >= 0)
            {
              Settings.ShellCommand shellCommand = (Settings.ShellCommand)(widget.getItem(index).getData());

              if (editShellCommand(shellCommand,"Edit shell commands","Save"))
              {
                Widgets.updateTableItem(widgetShellCommands,shellCommand,shellCommand.name,shellCommand.commandLine,Integer.toString(shellCommand.validExitcode));
              }
            }
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
          }
        });
        for (Settings.ShellCommand shellCommand : Settings.shellCommands)
        {
          Widgets.addTableItem(widgetShellCommands,shellCommand.clone(),shellCommand.name,shellCommand.commandLine,Integer.toString(shellCommand.validExitcode));
        }

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(null,null));
        Widgets.layout(subComposite,1,0,TableLayoutData.E);
        {
          button = Widgets.newButton(subComposite,"Add"+"\u2026");
          Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              Settings.ShellCommand shellCommand = new Settings.ShellCommand();

              if (editShellCommand(shellCommand,"Add shell command","Add"))
              {
                Widgets.addTableItem(widgetShellCommands,shellCommand,shellCommand.name,shellCommand.commandLine,Integer.toString(shellCommand.validExitcode));
              }
            }
          });

          button = Widgets.newButton(subComposite,"Edit"+"\u2026");
          Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              int index = widgetShellCommands.getSelectionIndex();
              if (index >= 0)
              {
                Settings.ShellCommand shellCommand = (Settings.ShellCommand)(widgetShellCommands.getItem(index).getData());

                if (editShellCommand(shellCommand,"Edit shell commands","Save"))
                {
                  Widgets.updateTableItem(widgetShellCommands,shellCommand,shellCommand.name,shellCommand.commandLine,Integer.toString(shellCommand.validExitcode));
                }
              }
            }
          });

          button = Widgets.newButton(subComposite,"Clone"+"\u2026");
          Widgets.layout(button,0,2,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              int index = widgetShellCommands.getSelectionIndex();
              if (index >= 0)
              {
                Settings.ShellCommand shellCommand = ((Settings.ShellCommand)(widgetShellCommands.getItem(index).getData())).clone();

                if (editShellCommand(shellCommand,"Add shell command","Add"))
                {
                  Widgets.addTableItem(widgetShellCommands,shellCommand,shellCommand.name,shellCommand.commandLine,Integer.toString(shellCommand.validExitcode));
                }
              }
            }
          });

          button = Widgets.newButton(subComposite,"Remove");
          Widgets.layout(button,0,3,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              int index = widgetShellCommands.getSelectionIndex();
              if (index >= 0)
              {
                widgetShellCommands.remove(index);
              }
            }
          });
        }
      }

      composite = Widgets.addTab(tabFolder,"Mail");
      composite.setLayout(new TableLayout(0.0,1.0,2));
      Widgets.layout(composite,0,3,TableLayoutData.NSWE);
      {
        subComposite = Widgets.newGroup(composite,"External Mail");
        subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0},8));
        Widgets.layout(subComposite,0,0,TableLayoutData.WE);
        {
          label = Widgets.newLabel(subComposite,"Command:");
          Widgets.layout(label,0,0,TableLayoutData.W);
          widgetCommandLineMail = Widgets.newText(subComposite);
          widgetCommandLineMail.setToolTipText("External mail command.\nMacros:\n  ${to} - to address\n  ${cc} - CC address\n  ${subject} - subject line\n");
          widgetCommandLineMail.setText(Settings.commandMail);
          Widgets.layout(widgetCommandLineMail,0,1,TableLayoutData.WE);

          label = Widgets.newLabel(subComposite,"Command with attachment:");
          Widgets.layout(label,1,0,TableLayoutData.W);
          widgetCommandLineMailAttachment = Widgets.newText(subComposite);
          widgetCommandLineMailAttachment.setToolTipText("External mail command with an attachment.\nMacros:\n  ${to} - to address\n  ${cc} - CC address\n  ${subject} - subject line\n  ${file} - attachment file name");
          widgetCommandLineMailAttachment.setText(Settings.commandMailAttachment);
          Widgets.layout(widgetCommandLineMailAttachment,1,1,TableLayoutData.WE);
        }

        subComposite = Widgets.newGroup(composite,"SMTP Mail");
        subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0},8));
        Widgets.layout(subComposite,1,0,TableLayoutData.WE);
        {
          label = Widgets.newLabel(subComposite,"Server:");
          Widgets.layout(label,0,0,TableLayoutData.W,0,0,2);

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,new double[]{0.7,0.0,0.3,0.0}));
          Widgets.layout(subSubComposite,0,1,TableLayoutData.WE,0,0,2);
          {
            widgetMailSMTPHost = Widgets.newText(subSubComposite);
            widgetMailSMTPHost.setToolTipText("Mail SMTP server host name.");
            widgetMailSMTPHost.setText(Settings.mailSMTPHost);
            Widgets.layout(widgetMailSMTPHost,0,0,TableLayoutData.WE);

            label = Widgets.newLabel(subSubComposite,"Port:");
            Widgets.layout(label,0,1,TableLayoutData.W);

            widgetMailSMTPPort = Widgets.newSpinner(subSubComposite,0,65535);
            widgetMailSMTPPort.setToolTipText("Mail SMTP server port number.");
            widgetMailSMTPPort.setTextLimit(5);
            widgetMailSMTPPort.setSelection(Settings.mailSMTPPort);
            Widgets.layout(widgetMailSMTPPort,0,2,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);

            widgetMailSMTPSSL = Widgets.newCheckbox(subSubComposite,"SSL");
            widgetMailSMTPSSL.setToolTipText("Use SMTP with SSL encryption.");
            widgetMailSMTPSSL.setSelection(Settings.mailSMTPSSL);
            Widgets.layout(widgetMailSMTPSSL,0,3,TableLayoutData.E);
          }

          label = Widgets.newLabel(subComposite,"Login:");
          Widgets.layout(label,1,0,TableLayoutData.W,0,0,2);

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0,1.0}));
          Widgets.layout(subSubComposite,1,1,TableLayoutData.WE,0,0,2);
          {
            widgetMailLogin = Widgets.newText(subSubComposite);
            widgetMailLogin.setToolTipText("Mail server login name.");
            widgetMailLogin.setText(Settings.mailLogin);
            Widgets.layout(widgetMailLogin,0,0,TableLayoutData.WE);

            label = Widgets.newLabel(subSubComposite,"Password:");
            Widgets.layout(label,0,1,TableLayoutData.W);

            widgetMailPassword = Widgets.newPassword(subSubComposite);
            widgetMailPassword.setToolTipText("Mail server login password.");
            String password = onzen.getPassword(Settings.mailLogin,Settings.mailSMTPHost,false);
            if (password != null) widgetMailPassword.setText(password);
            Widgets.layout(widgetMailPassword,0,2,TableLayoutData.WE);
          }

          label = Widgets.newLabel(subComposite,"From name:");
          Widgets.layout(label,2,0,TableLayoutData.W,0,0,2);

          widgetMailFrom = Widgets.newText(subComposite);
          widgetMailFrom.setToolTipText("Mail from address.");
          widgetMailFrom.setText(Settings.mailFrom);
          Widgets.layout(widgetMailFrom,2,1,TableLayoutData.WE,0,0,2);

          button = Widgets.newButton(subComposite,"Send test mail");
          Widgets.layout(button,3,1,TableLayoutData.E);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              String toAddress = Dialogs.string(dialog,"Mail address","Mail address:",widgetMailFrom.getText().trim(),"Send","Cancel");
              if ((toAddress != null) && !toAddress.trim().isEmpty())
              {
                sendTestMail(widgetMailSMTPHost.getText().trim(),
                             Integer.parseInt(widgetMailSMTPPort.getText()),
                             widgetMailSMTPSSL.getSelection(),
                             widgetMailLogin.getText().trim(),
                             widgetMailPassword.getText().trim(),
                             widgetMailFrom.getText().trim(),
                             toAddress.trim()
                            );
              }
           }
          });
        }

        subComposite = Widgets.newGroup(composite,"Post review server");
        subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0},8));
        Widgets.layout(subComposite,2,0,TableLayoutData.WE);
        {
          label = Widgets.newLabel(subComposite,"Server:");
          Widgets.layout(label,0,0,TableLayoutData.W);
          widgetReviewServerHost = Widgets.newText(subComposite);
          widgetReviewServerHost.setToolTipText("Default post review server name.\n");
          widgetReviewServerHost.setText(Settings.reviewServerHost);
          Widgets.layout(widgetReviewServerHost,0,1,TableLayoutData.WE);

          label = Widgets.newLabel(subComposite,"Login:");
          Widgets.layout(label,1,0,TableLayoutData.W);
          widgetReviewServerLogin = Widgets.newText(subComposite);
          widgetReviewServerLogin.setToolTipText("Default review server login name.\n");
          widgetReviewServerLogin.setText(Settings.reviewServerLogin);
          Widgets.layout(widgetReviewServerLogin,1,1,TableLayoutData.WE);

          label = Widgets.newLabel(subComposite,"Post command:");
          Widgets.layout(label,2,0,TableLayoutData.W);
          widgetCommandLinePostReviewServer = Widgets.newText(subComposite);
          widgetCommandLinePostReviewServer.setToolTipText("Post review server command.\nMacros:\n  ${server} - review server name\n  ${login} - login name\n  ${password} - password\n  ${summary} - summary line\n  ${description} - description\n  ${tests} - tests done\n  ${file} - diff file name\n");
          widgetCommandLinePostReviewServer.setText(Settings.commandPostReviewServer);
          Widgets.layout(widgetCommandLinePostReviewServer,2,1,TableLayoutData.WE);

          label = Widgets.newLabel(subComposite,"Update command:");
          Widgets.layout(label,3,0,TableLayoutData.W);
          widgetCommandLineUpdateReviewServer = Widgets.newText(subComposite);
          widgetCommandLineUpdateReviewServer.setToolTipText("Update review server command.\nMacros:\n  ${server} - review server name\n  ${login} - login name\n  ${password} - password\n  ${reference} - reference\n  ${summary} - summary line\n  ${description} - description\n  ${tests} - tests done\n  ${file} - diff file name\n");
          widgetCommandLineUpdateReviewServer.setText(Settings.commandUpdateReviewServer);
          Widgets.layout(widgetCommandLineUpdateReviewServer,3,1,TableLayoutData.WE);
        }
      }

      composite = Widgets.addTab(tabFolder,"Repository");
      composite.setLayout(new TableLayout(null,1.0,2));
      Widgets.layout(composite,0,4,TableLayoutData.NSWE);
      {
        subTabFolder = Widgets.newTabFolder(composite);
        Widgets.layout(subTabFolder,0,0,TableLayoutData.NSWE);
        {

          // --- CVS -----------------------------------------------------

          subComposite = Widgets.addTab(subTabFolder,"CVS");
          subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0},8));
          Widgets.layout(subComposite,0,0,TableLayoutData.NSWE);
          {
            label = Widgets.newLabel(subComposite,"Command:");
            Widgets.layout(label,0,0,TableLayoutData.W);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
            Widgets.layout(subSubComposite,0,1,TableLayoutData.WE);
            {
              widgetCVSCommand = Widgets.newText(subSubComposite);
              widgetCVSCommand.setToolTipText("CVS command.");
              widgetCVSCommand.setText(Settings.cvsCommand);
              Widgets.layout(widgetCVSCommand,0,0,TableLayoutData.WE);

              button = Widgets.newButton(subSubComposite,Onzen.IMAGE_DIRECTORY);
              Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String command = Dialogs.fileOpen(shell,
                                                    "Select command",
                                                    widgetCVSCommand.getText(),
                                                    new String[]{"Scripts",    "*.sh",
                                                                 "Batch files","*.cmd",
                                                                 "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                                 "All files",  Onzen.ALL_FILE_EXTENSION
                                                                }
                                                   );
                  if (command != null)
                  {
                    widgetCVSCommand.setText(command);
                  }
                }
              });
            }

            label = Widgets.newLabel(subComposite,"Miscellaneous:");
            Widgets.layout(label,1,0,TableLayoutData.W);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,null));
            Widgets.layout(subSubComposite,1,1,TableLayoutData.WE);
            {
              widgetCVSPruneEmptyDirectories = Widgets.newCheckbox(subSubComposite,"prune empty directories");
              widgetCVSPruneEmptyDirectories.setToolTipText("Delete empty directories after check-out.");
              widgetCVSPruneEmptyDirectories.setSelection(Settings.cvsPruneEmtpyDirectories);
              Widgets.layout(widgetCVSPruneEmptyDirectories,0,0,TableLayoutData.WE);
            }
          }

          // --- SVN -----------------------------------------------------

          subComposite = Widgets.addTab(subTabFolder,"SVN");
          subComposite.setLayout(new TableLayout(new double[]{1.0,0.0,0.0},new double[]{0.0,1.0},8));
          Widgets.layout(subComposite,1,0,TableLayoutData.NSWE);
          {
            label = Widgets.newLabel(subComposite,"Commands:");
            Widgets.layout(label,0,0,TableLayoutData.NW);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(new double[]{1.0,0.0},new double[]{1.0}));
            Widgets.layout(subSubComposite,0,1,TableLayoutData.NSWE);
            {
              widgetSVNCommands = Widgets.newTable(subSubComposite);
              widgetSVNCommands.setToolTipText("SVN command list.");
              Widgets.layout(widgetSVNCommands,0,0,TableLayoutData.NSWE);
              tableColumn = Widgets.addTableColumn(widgetSVNCommands,0,"Version",SWT.LEFT,50,true);
              tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
              tableColumn = Widgets.addTableColumn(widgetSVNCommands,1,"Command", SWT.LEFT);
              tableColumn.addSelectionListener(Widgets.DEFAULT_TABLE_SELECTION_LISTENER_STRING);
              for (String name : Settings.svnCommands.names())
              {
                SVNCommand svnCommand = new SVNCommand(name,Settings.svnCommands.get(name));
                Widgets.addTableItem(widgetSVNCommands,svnCommand,svnCommand.version,svnCommand.command);
              }
              widgetSVNCommands.addMouseListener(new MouseListener()
              {
                public void mouseDoubleClick(MouseEvent mouseEvent)
                {
                  Table widget = (Table)mouseEvent.widget;

                  int index = widget.getSelectionIndex();
                  if (index >= 0)
                  {
                    TableItem tableItem   = widget.getItem(index);
                    SVNCommand svnCommand = (SVNCommand)tableItem.getData();
                    if (editSVNCommand(svnCommand,"Edit SVN command","Save"))
                    {
                      Widgets.updateTableItem(widgetSVNCommands,svnCommand,svnCommand.version,svnCommand.command);
                    }
                  }
                }
                public void mouseDown(MouseEvent mouseEvent)
                {
                }
                public void mouseUp(MouseEvent mouseEvent)
                {
                }
              });

              subSubSubComposite = Widgets.newComposite(subSubComposite);
              subSubSubComposite.setLayout(new TableLayout(null,null));
              Widgets.layout(subSubSubComposite,1,0,TableLayoutData.E);
              {
                button = Widgets.newButton(subSubSubComposite,"Add"+"\u2026");
                Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
                button.addSelectionListener(new SelectionListener()
                {
                  public void widgetDefaultSelected(SelectionEvent selectionEvent)
                  {
                  }
                  public void widgetSelected(SelectionEvent selectionEvent)
                  {
                    SVNCommand svnCommand = new SVNCommand();
                    if (editSVNCommand(svnCommand,"Add SVN command","Add"))
                    {
                      Widgets.addTableItem(widgetSVNCommands,svnCommand,svnCommand.version,svnCommand.command);
                    }
                  }
                });

                button = Widgets.newButton(subSubSubComposite,"Remove");
                Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
                button.addSelectionListener(new SelectionListener()
                {
                  public void widgetDefaultSelected(SelectionEvent selectionEvent)
                  {
                  }
                  public void widgetSelected(SelectionEvent selectionEvent)
                  {
                    int index = widgetSVNCommands.getSelectionIndex();
                    if (index >= 0)
                    {
                      widgetSVNCommands.remove(index);
                    }
                  }
                });
              }
            }
/*
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
            Widgets.layout(subSubComposite,0,1,TableLayoutData.WE);
            {
              widgetSVNCommand = Widgets.newText(subSubComposite);
              widgetSVNCommand.setToolTipText("SVN command.");
              for (
              widgetHGCommand.setText(Settings.hgCommand);
              widgetSVNCommand.setText(Settings.svnCommand.get());
              Widgets.layout(widgetSVNCommand,0,0,TableLayoutData.WE);

              button = Widgets.newButton(subSubComposite,Onzen.IMAGE_DIRECTORY);
              Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String command = Dialogs.fileOpen(shell,
                                                    "Select command",
                                                    widgetCVSCommand.getText(),
                                                    new String[]{"Scripts",    "*.sh",
                                                                 "Batch files","*.cmd",
                                                                 "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                                 "All files",  Onzen.ALL_FILE_EXTENSION
                                                                }
                                                   );
                  if (command != null)
                  {
                    widgetSVNCommand.setText(command);
                  }
                }
              });
            }
*/

            label = Widgets.newLabel(subComposite,"External diff:");
            Widgets.layout(label,1,0,TableLayoutData.NW);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
            Widgets.layout(subSubComposite,1,1,TableLayoutData.WE);
            {
              widgetSVNDiffCommand = Widgets.newText(subSubComposite);
              widgetSVNDiffCommand.setToolTipText("External SVN diff command.");
              widgetSVNDiffCommand.setText(Settings.svnDiffCommand);
              Widgets.layout(widgetSVNDiffCommand,0,0,TableLayoutData.WE);

              button = Widgets.newButton(subSubComposite,Onzen.IMAGE_DIRECTORY);
              Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String command = Dialogs.fileOpen(shell,
                                                    "Select command",
                                                    widgetSVNDiffCommand.getText(),
                                                    new String[]{"Scripts",    "*.sh",
                                                                 "Batch files","*.cmd",
                                                                 "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                                 "All files",  Onzen.ALL_FILE_EXTENSION
                                                                }
                                                   );
                  if (command != null)
                  {
                    widgetSVNDiffCommand.setText(command);
                  }
                }
              });
            }

            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0,0.0,1.0}));
            Widgets.layout(subSubComposite,2,1,TableLayoutData.WE);
            {
              label = Widgets.newLabel(subSubComposite,"Options:");
              Widgets.layout(label,0,0,TableLayoutData.W);
              widgetSVNDiffCommandOptions = Widgets.newText(subSubComposite);
              widgetSVNDiffCommandOptions.setToolTipText("Options for external diff command. Leave it empty for using internal diff of SVN.");
              widgetSVNDiffCommandOptions.setText(Settings.svnDiffCommandOptions);
              Widgets.layout(widgetSVNDiffCommandOptions,0,1,TableLayoutData.WE);

              label = Widgets.newLabel(subSubComposite,"No whitespaces:");
              Widgets.layout(label,0,2,TableLayoutData.W);
              widgetSVNDiffCommandOptionsIgnoreWhitespaces = Widgets.newText(subSubComposite);
              widgetSVNDiffCommandOptionsIgnoreWhitespaces.setToolTipText("Options for external diff command with ignoring whitespace changes.");
              widgetSVNDiffCommandOptionsIgnoreWhitespaces.setText(Settings.svnDiffCommandOptionsIgnoreWhitespaces);
              Widgets.layout(widgetSVNDiffCommandOptionsIgnoreWhitespaces,0,3,TableLayoutData.WE);
            }

            label = Widgets.newLabel(subComposite,"Miscellaneous:");
            Widgets.layout(label,3,0,TableLayoutData.NW);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,1.0));
            Widgets.layout(subSubComposite,3,1,TableLayoutData.WE);
            {
              widgetSVNAlwaysTrustServerCertificate = Widgets.newCheckbox(subSubComposite,"always trust server certificate");
              widgetSVNAlwaysTrustServerCertificate.setToolTipText("Always trust server certificate.");
              widgetSVNAlwaysTrustServerCertificate.setSelection(Settings.svnAlwaysTrustServerCertificate);
              Widgets.layout(widgetSVNAlwaysTrustServerCertificate,0,0,TableLayoutData.WE);
            }
          }

          // --- HG ------------------------------------------------------

          subComposite = Widgets.addTab(subTabFolder,"HG");
          subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0},8));
          Widgets.layout(subComposite,2,0,TableLayoutData.NSWE);
          {
            label = Widgets.newLabel(subComposite,"Command:");
            Widgets.layout(label,0,0,TableLayoutData.W);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
            Widgets.layout(subSubComposite,0,1,TableLayoutData.WE);
            {
              widgetHGCommand = Widgets.newText(subSubComposite);
              widgetHGCommand.setToolTipText("HG command.");
              widgetHGCommand.setText(Settings.hgCommand);
              Widgets.layout(widgetHGCommand,0,0,TableLayoutData.WE);

              button = Widgets.newButton(subSubComposite,Onzen.IMAGE_DIRECTORY);
              Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String command = Dialogs.fileOpen(shell,
                                                    "Select command",
                                                    widgetCVSCommand.getText(),
                                                    new String[]{"Scripts",    "*.sh",
                                                                 "Batch files","*.cmd",
                                                                 "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                                 "All files",  Onzen.ALL_FILE_EXTENSION
                                                                }
                                                   );
                  if (command != null)
                  {
                    widgetHGCommand.setText(command);
                  }
                }
              });
            }

            label = Widgets.newLabel(subComposite,"External diff:");
            Widgets.layout(label,1,0,TableLayoutData.NW);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
            Widgets.layout(subSubComposite,1,1,TableLayoutData.WE);
            {
              widgetHGDiffCommand = Widgets.newText(subSubComposite);
              widgetHGDiffCommand.setToolTipText("External HG diff command. Leave it empty for using internal diff of HG.");
              widgetHGDiffCommand.setText(Settings.hgDiffCommand);
              Widgets.layout(widgetHGDiffCommand,0,0,TableLayoutData.WE);

              button = Widgets.newButton(subSubComposite,Onzen.IMAGE_DIRECTORY);
              Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String command = Dialogs.fileOpen(shell,
                                                    "Select command",
                                                    widgetHGDiffCommand.getText(),
                                                    new String[]{"Scripts",    "*.sh",
                                                                 "Batch files","*.cmd",
                                                                 "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                                 "All files",  Onzen.ALL_FILE_EXTENSION
                                                                }
                                                   );
                  if (command != null)
                  {
                    widgetHGDiffCommand.setText(command);
                  }
                }
              });
            }

            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0,0.0,1.0}));
            Widgets.layout(subSubComposite,2,1,TableLayoutData.WE);
            {
              label = Widgets.newLabel(subSubComposite,"Options:");
              Widgets.layout(label,0,0,TableLayoutData.W);
              widgetHGDiffCommandOptions = Widgets.newText(subSubComposite);
              widgetHGDiffCommandOptions.setToolTipText("Options for external diff command.");
              widgetHGDiffCommandOptions.setText(Settings.hgDiffCommandOptions);
              Widgets.layout(widgetHGDiffCommandOptions,0,1,TableLayoutData.WE);

              label = Widgets.newLabel(subSubComposite,"No whitespaces:");
              Widgets.layout(label,0,2,TableLayoutData.W);
              widgetHGDiffCommandOptionsIgnoreWhitespaces = Widgets.newText(subSubComposite);
              widgetHGDiffCommandOptionsIgnoreWhitespaces.setToolTipText("Options for external diff command with ignoring whitespace changes.");
              widgetHGDiffCommandOptionsIgnoreWhitespaces.setText(Settings.hgDiffCommandOptionsIgnoreWhitespaces);
              Widgets.layout(widgetHGDiffCommandOptionsIgnoreWhitespaces,0,3,TableLayoutData.WE);
            }

            label = Widgets.newLabel(subComposite,"Miscellaneous:");
            Widgets.layout(label,3,0,TableLayoutData.NW);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,1.0));
            Widgets.layout(subSubComposite,3,1,TableLayoutData.WE);
            {
              widgetHGUseForestExtension = Widgets.newCheckbox(subSubComposite,"use forest extension");
              widgetHGUseForestExtension.setToolTipText("Use HG forest extension commands.");
              widgetHGUseForestExtension.setSelection(Settings.hgUseForestExtension);
              Widgets.layout(widgetHGUseForestExtension,0,0,TableLayoutData.WE);

              widgetHGUseQueueExtension = Widgets.newCheckbox(subSubComposite,"use queue extension");
              widgetHGUseQueueExtension.setToolTipText("Use HG queue extension commands.");
              widgetHGUseQueueExtension.setSelection(Settings.hgUseQueueExtension);
              Widgets.layout(widgetHGUseQueueExtension,1,0,TableLayoutData.WE);

              widgetHGUpdateWithFetch = Widgets.newCheckbox(subSubComposite,"update with fetch extension");
              widgetHGUpdateWithFetch.setToolTipText("Use HG fetch extension for update (fetch+fpush).");
              widgetHGUpdateWithFetch.setSelection(Settings.hgUpdateWithFetch);
              Widgets.layout(widgetHGUpdateWithFetch,2,0,TableLayoutData.WE);

              widgetHGSafeUpdate = Widgets.newCheckbox(subSubComposite,"'safe' update");
              widgetHGSafeUpdate.setToolTipText("Do 'safe' update. Allow fetch update with not-commited local changes: save local changes, revert, update and restore local changes with merge if needed.");
              widgetHGSafeUpdate.setSelection(Settings.hgSafeUpdate);
              Widgets.layout(widgetHGSafeUpdate,3,0,TableLayoutData.WE);

              subSubSubComposite = Widgets.newComposite(subSubComposite);
              subSubSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0,0.0}));
              Widgets.layout(subSubSubComposite,4,0,TableLayoutData.WE);
              {
                widgetHGSingleLineCommitMessages = Widgets.newCheckbox(subSubSubComposite,"single-line commit messages");
                widgetHGSingleLineCommitMessages.setToolTipText("Concat commit message into single-line to make commit message visible in the one-line-only commit message log of Mercurial.");
                widgetHGSingleLineCommitMessages.setSelection(Settings.hgSingleLineCommitMessages);
                Widgets.layout(widgetHGSingleLineCommitMessages,0,0,TableLayoutData.WE);
                widgetHGSingleLineCommitMessages.addSelectionListener(new SelectionListener()
                {
                  public void widgetDefaultSelected(SelectionEvent selectionEvent)
                  {
                  }
                  public void widgetSelected(SelectionEvent selectionEvent)
                  {
                    // notify modification
                    Widgets.modified(Settings.hgSingleLineCommitMessages);
                  }
                });

                label = Widgets.newLabel(subSubSubComposite,"Max:");
                Widgets.layout(label,0,1,TableLayoutData.W);

                widgetHGSingleLineMaxCommitMessageLength = Widgets.newSpinner(subSubSubComposite);
                widgetHGSingleLineMaxCommitMessageLength.setToolTipText("Max. length of single-line commit message. If the length exceed this limit a warning is shown.");
                widgetHGSingleLineMaxCommitMessageLength.setEnabled(Settings.hgSingleLineCommitMessages);
                widgetHGSingleLineMaxCommitMessageLength.setTextLimit(5);
                widgetHGSingleLineMaxCommitMessageLength.setSelection(Settings.hgSingleLineMaxCommitMessageLength);
                Widgets.layout(widgetHGSingleLineMaxCommitMessageLength,0,2,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
                Widgets.addModifyListener(new WidgetModifyListener(widgetHGSingleLineMaxCommitMessageLength,Settings.hgSingleLineCommitMessages)
                {
                  public void modified(Control control)
                  {
                    control.setEnabled(widgetHGSingleLineCommitMessages.getSelection());
                  }
                });
              }

              widgetHGRelativePatchPaths = Widgets.newCheckbox(subSubComposite,"relative patch paths");
              widgetHGRelativePatchPaths.setToolTipText("Remove path prefixes 'a/' and 'b/' Mercurial is adding to file names in patches and convert to a relative path.");
              widgetHGRelativePatchPaths.setSelection(Settings.hgRelativePatchPaths);
              Widgets.layout(widgetHGRelativePatchPaths,5,0,TableLayoutData.WE);
            }
          }

          // --- GIT -----------------------------------------------------

          subComposite = Widgets.addTab(subTabFolder,"GIT");
          subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0},8));
          Widgets.layout(subComposite,3,0,TableLayoutData.NSWE);
          {
            label = Widgets.newLabel(subComposite,"Command:");
            Widgets.layout(label,0,0,TableLayoutData.W);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
            Widgets.layout(subSubComposite,0,1,TableLayoutData.WE);
            {
              widgetGITCommand = Widgets.newText(subSubComposite);
              widgetGITCommand.setToolTipText("GIT command.");
              widgetGITCommand.setText(Settings.gitCommand);
              Widgets.layout(widgetGITCommand,0,0,TableLayoutData.WE);

              button = Widgets.newButton(subSubComposite,Onzen.IMAGE_DIRECTORY);
              Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
              button.addSelectionListener(new SelectionListener()
              {
                public void widgetDefaultSelected(SelectionEvent selectionEvent)
                {
                }
                public void widgetSelected(SelectionEvent selectionEvent)
                {
                  String command = Dialogs.fileOpen(shell,
                                                    "Select command",
                                                    widgetCVSCommand.getText(),
                                                    new String[]{"Scripts",    "*.sh",
                                                                 "Batch files","*.cmd",
                                                                 "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                                 "All files",  Onzen.ALL_FILE_EXTENSION
                                                                }
                                                   );
                  if (command != null)
                  {
                    widgetGITCommand.setText(command);
                  }
                }
              });
            }

            label = Widgets.newLabel(subComposite,"Miscellaneous:");
            Widgets.layout(label,3,0,TableLayoutData.NW);
            subSubComposite = Widgets.newComposite(subComposite);
            subSubComposite.setLayout(new TableLayout(null,1.0));
            Widgets.layout(subSubComposite,3,1,TableLayoutData.WE);
            {
              widgetGITShortCommitNumbers = Widgets.newCheckbox(subSubComposite,"short commit numbers");
              widgetGITShortCommitNumbers.setToolTipText("Show short commit numbers.");
              widgetGITShortCommitNumbers.setSelection(Settings.gitShortCommitNumbers);
              Widgets.layout(widgetGITShortCommitNumbers,0,0,TableLayoutData.WE);

              widgetGITStashOnUpdate = Widgets.newCheckbox(subSubComposite,"stash on update");
              widgetGITStashOnUpdate.setToolTipText("Automatic save/pop stash on update.");
              widgetGITStashOnUpdate.setSelection(Settings.gitStashOnUpdate);
              Widgets.layout(widgetGITStashOnUpdate,1,0,TableLayoutData.WE);

              widgetGITAutoPullSubModules = Widgets.newCheckbox(subSubComposite,"auto pull submodules");
              widgetGITAutoPullSubModules.setToolTipText("Automatic pull sub-modules on pull.");
              widgetGITAutoPullSubModules.setSelection(Settings.gitAutoPullSubModules);
              Widgets.layout(widgetGITAutoPullSubModules,2,0,TableLayoutData.WE);
            }
          }
        }
        subTabFolder.setSelection(0);
      }

      composite = Widgets.addTab(tabFolder,"Data");
      composite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0},2));
      Widgets.layout(composite,0,5,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(composite,"Stored passwords:");
        Widgets.layout(label,0,0,TableLayoutData.NW);

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
        Widgets.layout(subComposite,0,1,TableLayoutData.NSWE);
        {
          widgetStoredPasswords = Widgets.newList(subComposite);
          widgetStoredPasswords.setToolTipText("Stored password list.");
          Widgets.layout(widgetStoredPasswords,0,0,TableLayoutData.NSWE);
          widgetStoredPasswords.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
              int index = widgetStoredPasswords.getSelectionIndex();
              if (index >= 0)
              {
                String name = (String)Widgets.getListItem(widgetStoredPasswords,index);
                String password = editPassword(name,"Edit password");
                if (password != null)
                {
                  onzen.setPassword(name,password);
                }
              }
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
            }
          });
          widgetStoredPasswords.addKeyListener(new KeyListener()
          {
            public void keyPressed(KeyEvent keyEvent)
            {
              if      (Widgets.isAccelerator(keyEvent,SWT.DEL))
              {
                int index = widgetStoredPasswords.getSelectionIndex();
                if (index >= 0)
                {
                  Widgets.removeListItem(widgetStoredPasswords,index);
                }
              }
            }
            public void keyReleased(KeyEvent keyEvent)
            {
            }
          });
          for (String name : onzen.getPasswordNames())
          {
            Widgets.addListItem(widgetStoredPasswords,name,name);
          }

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,null));
          Widgets.layout(subSubComposite,1,0,TableLayoutData.E);
          {
            button = Widgets.newButton(subSubComposite,"Remove");
            Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetStoredPasswords.getSelectionIndex();
                if (index >= 0)
                {
                  Widgets.removeListItem(widgetStoredPasswords,index);
                }
              }
            });
          }
        }

        label = Widgets.newLabel(composite,"Repository history:");
        Widgets.layout(label,1,0,TableLayoutData.NW);

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
        Widgets.layout(subComposite,1,1,TableLayoutData.NSWE);
        {
          widgetRepositoryHistoryURLs = Widgets.newList(subComposite);
          widgetRepositoryHistoryURLs.setToolTipText("Repository URL history list.");
          Widgets.layout(widgetRepositoryHistoryURLs,0,0,TableLayoutData.NSWE);
          widgetRepositoryHistoryURLs.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
              int index = widgetRepositoryHistoryURLs.getSelectionIndex();
              if (index >= 0)
              {
                RepositoryURL repositoryURL = (RepositoryURL)Widgets.getListItem(widgetRepositoryHistoryURLs,index);
                if (editRepositoryURL(repositoryURL,"Edit repository URL","Save"))
                {
                  Widgets.updateListItem(widgetRepositoryHistoryURLs,repositoryURL,repositoryURL.path);
                  Widgets.sortList(widgetRepositoryHistoryURLs);
                }
              }
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
            }
          });
          widgetRepositoryHistoryURLs.addKeyListener(new KeyListener()
          {
            public void keyPressed(KeyEvent keyEvent)
            {
              if      (Widgets.isAccelerator(keyEvent,SWT.DEL))
              {
                int index = widgetRepositoryHistoryURLs.getSelectionIndex();
                if (index >= 0)
                {
                  Widgets.removeListItem(widgetRepositoryHistoryURLs,index);
                }
              }
            }
            public void keyReleased(KeyEvent keyEvent)
            {
            }
          });
          for (RepositoryURL repositoryURL : RepositoryURL.getHistoryURLs())
          {
            Widgets.addListItem(widgetRepositoryHistoryURLs,repositoryURL,repositoryURL.path);
          }

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,null));
          Widgets.layout(subSubComposite,1,0,TableLayoutData.E);
          {
            button = Widgets.newButton(subSubComposite,"Remove");
            Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetRepositoryHistoryURLs.getSelectionIndex();
                if (index >= 0)
                {
                  Widgets.removeListItem(widgetRepositoryHistoryURLs,index);
                }
              }
            });
          }
        }

        label = Widgets.newLabel(composite,"Additional repositories:");
        Widgets.layout(label,2,0,TableLayoutData.NW);

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
        Widgets.layout(subComposite,2,1,TableLayoutData.NSWE);
        {
          widgetRepositoryURLs = Widgets.newList(subComposite);
          widgetRepositoryURLs.setToolTipText("Additional repository path list.");
          Widgets.layout(widgetRepositoryURLs,0,0,TableLayoutData.NSWE);
          widgetRepositoryURLs.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
              List widget = (List)selectionEvent.widget;

              int index = widget.getSelectionIndex();
              if (index >= 0)
              {
                RepositoryURL repositoryURL = (RepositoryURL)Widgets.getListItem(widget,index);
                if (editRepositoryURL(repositoryURL,"Edit repository URL","Save"))
                {
                  Widgets.updateListItem(widget,repositoryURL,repositoryURL.path);
                }
              }
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
            }
          });
          widgetRepositoryURLs.addKeyListener(new KeyListener()
          {
            public void keyPressed(KeyEvent keyEvent)
            {
              if      (Widgets.isAccelerator(keyEvent,SWT.DEL))
              {
                int index = widgetRepositoryURLs.getSelectionIndex();
                if (index >= 0)
                {
                  Widgets.removeListItem(widgetRepositoryURLs,index);
                }
              }
            }
            public void keyReleased(KeyEvent keyEvent)
            {
            }
          });
          for (RepositoryURL repositoryURL : Settings.repositoryURLs)
          {
            Widgets.addListItem(widgetRepositoryURLs,repositoryURL,repositoryURL.path);
          }

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,null));
          Widgets.layout(subSubComposite,1,0,TableLayoutData.E);
          {
            button = Widgets.newButton(subSubComposite,"Add"+"\u2026");
            Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                RepositoryURL repositoryURL = new RepositoryURL();
                if (editRepositoryURL(repositoryURL,"Add repository URL","Add"))
                {
                  Widgets.addListItem(widgetRepositoryURLs,repositoryURL,repositoryURL.path);
                  Widgets.sortList(widgetRepositoryURLs);
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Remove");
            Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetRepositoryURLs.getSelectionIndex();
                if (index >= 0)
                {
                  Widgets.removeListItem(widgetRepositoryURLs,index);
                }
              }
            });
          }
        }
      }

      composite = Widgets.addTab(tabFolder,"Files");
      composite.setLayout(new TableLayout(new double[]{0.0,0.0,1.0,1.0,1.0},new double[]{0.0,1.0},2));
      Widgets.layout(composite,0,6,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(composite,"End-Of-Line type:");
        Widgets.layout(label,0,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(null,0.0));
        Widgets.layout(subComposite,0,1,TableLayoutData.WE);
        {
          widgetEOLAuto = Widgets.newRadio(subComposite,"auto");
          widgetEOLAuto.setToolTipText("Set system dependent end-of-line type.");
          widgetEOLAuto.setSelection(Settings.eolType == Settings.EOLTypes.AUTO);
          Widgets.layout(widgetEOLAuto,0,0,TableLayoutData.W);

          widgetEOLUnix = Widgets.newRadio(subComposite,"Unix (LF)");
          widgetEOLUnix.setToolTipText("Set system dependent end-of-line type.");
          widgetEOLUnix.setSelection(Settings.eolType == Settings.EOLTypes.UNIX);
          Widgets.layout(widgetEOLUnix,0,1,TableLayoutData.W);

          widgetEOLMac = Widgets.newRadio(subComposite,"Mac (CR)");
          widgetEOLMac.setToolTipText("Set system dependent end-of-line type.");
          widgetEOLMac.setSelection(Settings.eolType == Settings.EOLTypes.MAC);
          Widgets.layout(widgetEOLMac,0,2,TableLayoutData.W);

          widgetEOLWindows = Widgets.newRadio(subComposite,"Windows (CR+LF)");
          widgetEOLWindows.setToolTipText("Set system dependent end-of-line type.");
          widgetEOLWindows.setSelection(Settings.eolType == Settings.EOLTypes.WINDOWS);
          Widgets.layout(widgetEOLWindows,0,3,TableLayoutData.W);
        }

        label = Widgets.newLabel(composite,"Whitespaces:");
        Widgets.layout(label,1,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(null,1.0));
        Widgets.layout(subComposite,1,1,TableLayoutData.WE);
        {
          widgetCheckTABs = Widgets.newCheckbox(subComposite,"TABs check");
          widgetCheckTABs.setToolTipText("Check if file contain TABs before file is added or a commit is done.");
          widgetCheckTABs.setSelection(Settings.checkTABs);
          Widgets.layout(widgetCheckTABs,0,0,TableLayoutData.W);

          widgetCheckTrailingWhitespaces = Widgets.newCheckbox(subComposite,"whitespaces check");
          widgetCheckTrailingWhitespaces.setToolTipText("Check if a file contain trailing whitespaces before file is added or a commit is done.");
          widgetCheckTrailingWhitespaces.setSelection(Settings.checkTrailingWhitespaces);
          Widgets.layout(widgetCheckTrailingWhitespaces,1,0,TableLayoutData.W);

          widgetAutoTrimTrailingWhitespaces = Widgets.newCheckbox(subComposite,"auto trim trailing whitespaces");
          widgetAutoTrimTrailingWhitespaces.setToolTipText("Automatic remove trailing whitespaces before file is added or a commit is done.");
          widgetAutoTrimTrailingWhitespaces.setSelection(Settings.autoTrimTrailingWhitespaces);
          Widgets.layout(widgetAutoTrimTrailingWhitespaces,2,0,TableLayoutData.W);
        }

        label = Widgets.newLabel(composite,"Skip whitespace-check files:");
        Widgets.layout(label,2,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
        Widgets.layout(subComposite,2,1,TableLayoutData.NSWE);
        {
          widgetSkipWhitespaceCheckFilePatterns = Widgets.newList(subComposite);
          widgetSkipWhitespaceCheckFilePatterns.setToolTipText("Patterns for files where whitespace-check should be skipped.");
          Widgets.layout(widgetSkipWhitespaceCheckFilePatterns,0,0,TableLayoutData.NSWE);
          widgetSkipWhitespaceCheckFilePatterns.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
Dprintf.dprintf("");
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
Dprintf.dprintf("");
            }
          });
          widgetSkipWhitespaceCheckFilePatterns.addMouseListener(new MouseListener()
          {
            public void mouseDoubleClick(MouseEvent mouseEvent)
            {
              List widget = (List)mouseEvent.widget;

              int index = widget.getSelectionIndex();
              if (index >= 0)
              {
                String string = Dialogs.string(dialog,"Edit file pattern","Pattern:",widget.getItem(index),"Save","Cancel","File pattern: * and ?");
                if (string != null)
                {
                  widget.setItem(index,string);
                }
              }
            }
            public void mouseDown(MouseEvent mouseEvent)
            {
            }
            public void mouseUp(MouseEvent mouseEvent)
            {
            }
          });
          for (Settings.FilePattern filePattern : Settings.skipWhitespaceCheckFilePatterns)
          {
            widgetSkipWhitespaceCheckFilePatterns.add(filePattern.string);
          }

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,null));
          Widgets.layout(subSubComposite,1,0,TableLayoutData.E);
          {
            button = Widgets.newButton(subSubComposite,"Add"+"\u2026");
            Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                String string = Dialogs.string(dialog,"Add file pattern","Pattern:","","Add","Cancel","File pattern: * and ?");
                if (string != null)
                {
                  widgetSkipWhitespaceCheckFilePatterns.add(string);
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Edit"+"\u2026");
            Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetSkipWhitespaceCheckFilePatterns.getSelectionIndex();
                if (index >= 0)
                {
                  String string = Dialogs.string(dialog,"Edit file pattern","Pattern:",widgetSkipWhitespaceCheckFilePatterns.getItem(index),"Save","Cancel","File pattern: * and ?");
                  if (string != null)
                  {
                    widgetHiddenFilePatterns.setItem(index,string);
                  }
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Clone"+"\u2026");
            Widgets.layout(button,0,2,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetSkipWhitespaceCheckFilePatterns.getSelectionIndex();
                if (index >= 0)
                {
                  String string = Dialogs.string(dialog,"Clone file pattern","Pattern:",widgetSkipWhitespaceCheckFilePatterns.getItem(index),"Add","Cancel","File pattern: * and ?");
                  if (string != null)
                  {
                    widgetHiddenFilePatterns.add(string);
                  }
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Remove");
            Widgets.layout(button,0,3,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetSkipWhitespaceCheckFilePatterns.getSelectionIndex();
                if (index >= 0)
                {
                  widgetSkipWhitespaceCheckFilePatterns.remove(index);
                }
              }
            });
          }
        }

        label = Widgets.newLabel(composite,"Hidden files:");
        Widgets.layout(label,3,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
        Widgets.layout(subComposite,3,1,TableLayoutData.NSWE);
        {
          widgetHiddenFilePatterns = Widgets.newList(subComposite);
          widgetHiddenFilePatterns.setToolTipText("Patterns for hidden files in tree view.");
          Widgets.layout(widgetHiddenFilePatterns,0,0,TableLayoutData.NSWE);
          widgetHiddenFilePatterns.addMouseListener(new MouseListener()
          {
            public void mouseDoubleClick(MouseEvent mouseEvent)
            {
              List widget = (List)mouseEvent.widget;

              int index = widget.getSelectionIndex();
              if (index >= 0)
              {
                String string = Dialogs.string(dialog,"Edit hidden file pattern","Pattern:",widget.getItem(index),"Save","Cancel","File pattern: * and ?");
                if (string != null)
                {
                  widget.setItem(index,string);
                }
              }
            }
            public void mouseDown(MouseEvent mouseEvent)
            {
            }
            public void mouseUp(MouseEvent mouseEvent)
            {
            }
          });
          for (Settings.FilePattern filePattern : Settings.hiddenFilePatterns)
          {
            widgetHiddenFilePatterns.add(filePattern.string);
          }

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,null));
          Widgets.layout(subSubComposite,1,0,TableLayoutData.E);
          {
            button = Widgets.newButton(subSubComposite,"Add"+"\u2026");
            Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                String string = Dialogs.string(dialog,"Add hidden file pattern","Pattern:","","Add","Cancel","File pattern: * and ?");
                if (string != null)
                {
                  widgetHiddenFilePatterns.add(string);
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Edit"+"\u2026");
            Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetHiddenFilePatterns.getSelectionIndex();
                if (index >= 0)
                {
                  String string = Dialogs.string(dialog,"Edit hidden file pattern","Pattern:",widgetHiddenFilePatterns.getItem(index),"Save","Cancel","File pattern: * and ?");
                  if (string != null)
                  {
                    widgetHiddenFilePatterns.setItem(index,string);
                  }
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Clone"+"\u2026");
            Widgets.layout(button,0,2,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetHiddenFilePatterns.getSelectionIndex();
                if (index >= 0)
                {
                  String string = Dialogs.string(dialog,"Clone hidden file pattern","Pattern:",widgetHiddenFilePatterns.getItem(index),"Add","Cancel","File pattern: * and ?");
                  if (string != null)
                  {
                    widgetHiddenFilePatterns.add(string);
                  }
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Remove");
            Widgets.layout(button,0,3,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetHiddenFilePatterns.getSelectionIndex();
                if (index >= 0)
                {
                  widgetHiddenFilePatterns.remove(index);
                }
              }
            });
          }
        }

        label = Widgets.newLabel(composite,"Hidden directories:");
        Widgets.layout(label,4,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
        Widgets.layout(subComposite,4,1,TableLayoutData.NSWE);
        {
          widgetHiddenDirectoryPatterns = Widgets.newList(subComposite);
          widgetHiddenDirectoryPatterns.setToolTipText("Patterns for hidden directories in tree view.");
          Widgets.layout(widgetHiddenDirectoryPatterns,0,0,TableLayoutData.NSWE);
          widgetHiddenDirectoryPatterns.addMouseListener(new MouseListener()
          {
            public void mouseDoubleClick(MouseEvent mouseEvent)
            {
              List widget = (List)mouseEvent.widget;

              int index = widget.getSelectionIndex();
              if (index >= 0)
              {
                String string = Dialogs.string(dialog,"Edit hidden file directory","Pattern:",widget.getItem(index),"Save","Cancel","File pattern: * and ?");
                if (string != null)
                {
                  widget.setItem(index,string);
                }
              }
            }
            public void mouseDown(MouseEvent mouseEvent)
            {
            }
            public void mouseUp(MouseEvent mouseEvent)
            {
            }
          });
          for (Settings.FilePattern filePattern : Settings.hiddenDirectoryPatterns)
          {
            widgetHiddenDirectoryPatterns.add(filePattern.string);
          }

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,null));
          Widgets.layout(subSubComposite,1,0,TableLayoutData.E);
          {
            button = Widgets.newButton(subSubComposite,"Add"+"\u2026");
            Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                String string = Dialogs.string(dialog,"Add hidden directory pattern","Pattern:","","Add","Cancel","Directory pattern: * and ?");
                if (string != null)
                {
                  widgetHiddenDirectoryPatterns.add(string);
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Edit"+"\u2026");
            Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetHiddenDirectoryPatterns.getSelectionIndex();
                if (index >= 0)
                {
                  String string = Dialogs.string(dialog,"Edit hidden file directory","Pattern:",widgetHiddenFilePatterns.getItem(index),"Save","Cancel","File pattern: * and ?");
                  if (string != null)
                  {
                    widgetHiddenDirectoryPatterns.setItem(index,string);
                  }
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Clone"+"\u2026");
            Widgets.layout(button,0,2,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetHiddenDirectoryPatterns.getSelectionIndex();
                if (index >= 0)
                {
                  String string = Dialogs.string(dialog,"Clone hidden directory pattern","Pattern:",widgetHiddenDirectoryPatterns.getItem(index),"Add","Cancel","Directory pattern: * and ?");
                  if (string != null)
                  {
                    widgetHiddenDirectoryPatterns.add(string);
                  }
                }
              }
            });

            button = Widgets.newButton(subSubComposite,"Remove");
            Widgets.layout(button,0,3,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetHiddenDirectoryPatterns.getSelectionIndex();
                if (index >= 0)
                {
                  widgetHiddenDirectoryPatterns.remove(index);
                }
              }
            });
          }
        }
      }

      composite = Widgets.addTab(tabFolder,"Misc");
      composite.setLayout(new TableLayout(new double[]{0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,1.0,1.0,0.0},new double[]{0.0,1.0},2));
      Widgets.layout(composite,0,7,TableLayoutData.NSWE);
      {
        label = Widgets.newLabel(composite,"Show diff command:");
        Widgets.layout(label,0,0,TableLayoutData.W);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(1.0,new double[]{1.0,0.0}));
        Widgets.layout(subComposite,0,1,TableLayoutData.WE);
        {
          widgetShowDiffCommand = Widgets.newText(subComposite);
          widgetShowDiffCommand.setToolTipText("Temporary directory.");
          widgetShowDiffCommand.setText(Settings.showDiffCommand);
          Widgets.layout(widgetShowDiffCommand,0,0,TableLayoutData.WE);

          button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
          Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              String fileName = Dialogs.fileOpen(shell,
                                                 "Select program",
                                                 widgetShowDiffCommand.getText(),
                                                 new String[]{"Scripts",    "*.sh",
                                                              "Batch files","*.cmd",
                                                              "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                              "All files",  Onzen.ALL_FILE_EXTENSION
                                                             }
                                                 );
              if (fileName != null)
              {
                widgetShowDiffCommand.setText(fileName);
              }
            }
          });
        }

        label = Widgets.newLabel(composite,"Temporary directory:");
        Widgets.layout(label,1,0,TableLayoutData.W);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(1.0,new double[]{1.0,0.0}));
        Widgets.layout(subComposite,1,1,TableLayoutData.WE);
        {
          widgetTmpDirectory = Widgets.newText(subComposite);
          widgetTmpDirectory.setToolTipText("Temporary directory.");
          widgetTmpDirectory.setText(Settings.tmpDirectory);
          Widgets.layout(widgetTmpDirectory,0,0,TableLayoutData.WE);

          button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
          Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              String directoryPath = Dialogs.directory(shell,
                                                       "Select temporary directory",
                                                       widgetTmpDirectory.getText()
                                                      );
              if (directoryPath != null)
              {
                widgetTmpDirectory.setText(directoryPath);
              }
            }
          });
        }

        label = Widgets.newLabel(composite,"Backup file suffix:");
        Widgets.layout(label,2,0,TableLayoutData.W);
        widgetBackupFileSuffix = Widgets.newText(composite);
        widgetBackupFileSuffix.setToolTipText("Backup file name suffix.");
        widgetBackupFileSuffix.setText(Settings.backupFileSuffix);
        Widgets.layout(widgetBackupFileSuffix,2,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Date format:");
        Widgets.layout(label,3,0,TableLayoutData.W);
        widgetDateFormat = Widgets.newText(composite);
        widgetDateFormat.setToolTipText("Date format.\nPatterns:\n  y - year digit\n  M - month digit\n  d - day digit\n  E - week day name");
        widgetDateFormat.setText(Settings.dateFormat);
        Widgets.layout(widgetDateFormat,3,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Time format:");
        Widgets.layout(label,4,0,TableLayoutData.W);
        widgetTimeFormat = Widgets.newText(composite);
        widgetTimeFormat.setToolTipText("Time format.\nPatterns:\n  H - hour digit\n  m - minute digit\n  s - second digit");
        widgetTimeFormat.setText(Settings.timeFormat);
        Widgets.layout(widgetTimeFormat,4,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Date/Time format:");
        Widgets.layout(label,5,0,TableLayoutData.W);
        widgetDateTimeFormat = Widgets.newText(composite);
        widgetDateTimeFormat.setToolTipText("Date/time format.\nPatterns:\n  y - year digit\n  M - month digit\n  d - day digit\n  E - week day name\n  H - hour digit\n  m - minute digit\n  s - second digit");
        widgetDateTimeFormat.setText(Settings.dateTimeFormat);
        Widgets.layout(widgetDateTimeFormat,5,1,TableLayoutData.WE);

        label = Widgets.newLabel(composite,"Spaces per TAB:");
        Widgets.layout(label,6,0,TableLayoutData.W);
        widgetConvertSpacesPerTAB = Widgets.newSpinner(composite,1,256);
        widgetConvertSpacesPerTAB.setToolTipText("Number of spaces when converting TABs.");
        widgetConvertSpacesPerTAB.setSelection(Settings.convertSpacesPerTAB);
        Widgets.layout(widgetConvertSpacesPerTAB,6,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Max. background tasks:");
        Widgets.layout(label,7,0,TableLayoutData.W);
        widgetMaxBackgroundTasks = Widgets.newSpinner(composite,1,256);
        widgetMaxBackgroundTasks.setToolTipText("Max. number of background tasks.");
        widgetMaxBackgroundTasks.setSelection(Settings.maxBackgroundTasks);
        Widgets.layout(widgetMaxBackgroundTasks,7,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Max. message history:");
        Widgets.layout(label,8,0,TableLayoutData.W);
        widgetMaxMessageHistory = Widgets.newSpinner(composite,0);
        widgetMaxMessageHistory.setToolTipText("Max. length of commit message history.");
        widgetMaxMessageHistory.setSelection(Settings.maxMessageHistory);
        Widgets.layout(widgetMaxMessageHistory,8,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);

        label = Widgets.newLabel(composite,"Message broadcast:");
        Widgets.layout(label,9,0,TableLayoutData.W);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0,0.0}));
        Widgets.layout(subComposite,9,1,TableLayoutData.WE);
        {
          widgetMessageBroadcastAddress = Widgets.newText(subComposite);
          widgetMessageBroadcastAddress.setToolTipText("Commit message broadcast address.");
          widgetMessageBroadcastAddress.setText(Settings.messageBroadcastAddress);
          Widgets.layout(widgetMessageBroadcastAddress,0,0,TableLayoutData.WE);

          label = Widgets.newLabel(subComposite,":");
          Widgets.layout(label,0,1,TableLayoutData.W);

          widgetMessageBroadcastPort = Widgets.newSpinner(subComposite,1,65535);
          widgetMessageBroadcastPort.setToolTipText("Commit message broadcast port number.");
          widgetMessageBroadcastPort.setSelection(Settings.messageBroadcastPort);
          Widgets.layout(widgetMessageBroadcastPort,0,2,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);
        }

        label = Widgets.newLabel(composite,"Auto-summary patterns:");
        Widgets.layout(label,10,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
        Widgets.layout(subComposite,10,1,TableLayoutData.NSWE);
        {
          widgetAutoSummaryPatterns = Widgets.newList(subComposite);
          widgetAutoSummaryPatterns.setToolTipText("Patterns for creating summary message line.");
          Widgets.layout(widgetAutoSummaryPatterns,0,0,TableLayoutData.NSWE);
          widgetAutoSummaryPatterns.addMouseListener(new MouseListener()
          {
            public void mouseDoubleClick(MouseEvent mouseEvent)
            {
              List widget = (List)mouseEvent.widget;

              int index = widget.getSelectionIndex();
              if (index >= 0)
              {
                String  string    = widget.getItem(index);
                boolean retryFlag = false;
                do
                {
                  string = Dialogs.string(dialog,"Add auto-summary pattern","Pattern:",string,"Add","Cancel","Regular expression pattern.");
                  if (string != null)
                  {
                    retryFlag = false;
                    try
                    {
                      Pattern.compile(string);
                    }
                    catch (PatternSyntaxException exception)
                    {
                      if (Dialogs.confirm(dialog,"Confirm","Invalid pattern: '"+string+"' - re-edit?"))
                      {
                        retryFlag = true;
                      }
                      else
                      {
                        string = null;
                      }
                    }
                  }
                }
                while (retryFlag);

                if (string != null)
                {
                  widget.setItem(index,string);
                }
              }
            }
            public void mouseDown(MouseEvent mouseEvent)
            {
            }
            public void mouseUp(MouseEvent mouseEvent)
            {
            }
          });
          for (String pattern : Settings.autoSummaryPatterns)
          {
            widgetAutoSummaryPatterns.add(pattern);
          }

          subSubComposite = Widgets.newComposite(subComposite);
          subSubComposite.setLayout(new TableLayout(null,null));
          Widgets.layout(subSubComposite,1,0,TableLayoutData.E);
          {
            button = Widgets.newButton(subSubComposite,"Add"+"\u2026");
            Widgets.layout(button,0,0,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                String  string    = "(.*)";
                boolean retryFlag = false;
                do
                {
                  string = Dialogs.string(dialog,"Add auto-summary pattern","Pattern:",string,"Add","Cancel","Regular expression pattern.");
                  if (string != null)
                  {
                    retryFlag = false;
                    try
                    {
                      Pattern.compile(string);
                    }
                    catch (PatternSyntaxException exception)
                    {
                      if (Dialogs.confirm(dialog,"Confirm","Invalid pattern: '"+string+"' - re-edit?"))
                      {
                        retryFlag = true;
                      }
                      else
                      {
                        string = null;
                      }
                    }
                  }
                }
                while (retryFlag);

                if (string != null)
                {
                  widgetAutoSummaryPatterns.add(string);
                }
              }
            });
            button = Widgets.newButton(subSubComposite,"Remove");
            Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
            button.addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                int index = widgetAutoSummaryPatterns.getSelectionIndex();
                if (index >= 0)
                {
                  widgetAutoSummaryPatterns.remove(index);
                }
              }
            });
          }
        }

        label = Widgets.newLabel(composite,"Show dialog flags:");
        Widgets.layout(label,11,0,TableLayoutData.NW);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(1.0,1.0));
        Widgets.layout(subComposite,11,1,TableLayoutData.NSWE);
        {
          widgetShowFlags = Widgets.newTable(subComposite,SWT.CHECK);
          widgetShowFlags.setToolTipText("Show dialogs flags.");
          Widgets.layout(widgetShowFlags,0,0,TableLayoutData.NSWE);
          Widgets.addTableColumn(widgetShowFlags,0,"",SWT.LEFT,500,true);
          widgetShowFlags.setLinesVisible(false);
          widgetShowFlags.setHeaderVisible(false);
          widgetShowFlags.addListener(SWT.Resize, new Listener()
          {
            public void handleEvent(Event event)
            {
              Table       table = (Table)event.widget;
              TableColumn tableColumn = table.getColumn(0);

              tableColumn.setWidth(table.getClientArea().width);
            }
          });

          Widgets.addTableItem(widgetShowFlags,Settings.singleChangedFilesInstance,   "single changed-files dialog instance").setChecked(Settings.singleChangedFilesInstance   );
          Widgets.addTableItem(widgetShowFlags,Settings.showQuitConfirmation,         "quit confirmation"                   ).setChecked(Settings.showQuitConfirmation         );
          Widgets.addTableItem(widgetShowFlags,Settings.showUpdateStatusErrors,       "show update status errors"           ).setChecked(Settings.showUpdateStatusErrors       );
          Widgets.addTableItem(widgetShowFlags,Settings.showExecuteErrors,            "show execute errors"                 ).setChecked(Settings.showExecuteErrors            );
          Widgets.addTableItem(widgetShowFlags,Settings.showRestartAfterConfigChanged,"restart after configuration changed" ).setChecked(Settings.showRestartAfterConfigChanged);
          Widgets.addTableItem(widgetShowFlags,Settings.showUpdateInfo,               "show update information"             ).setChecked(Settings.showUpdateInfo               );
          Widgets.addTableItem(widgetShowFlags,Settings.showNewVersionInformation,    "show new version information"        ).setChecked(Settings.showNewVersionInformation    );
        }

        label = Widgets.newLabel(composite,"Miscellaneous:");
        Widgets.layout(label,12,0,TableLayoutData.W);
        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(0.0,0.0));
        Widgets.layout(subComposite,12,1,TableLayoutData.WE);
        {
          widgetSetWindowLocation = Widgets.newCheckbox(subComposite,"set window location");
          widgetSetWindowLocation.setToolTipText("Set window location relative to cursor location.");
          widgetSetWindowLocation.setSelection(Settings.setWindowLocation);
          Widgets.layout(widgetSetWindowLocation,0,1,TableLayoutData.W);
        }
      }
    }
    tabFolder.setSelection(0);

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetButtonSave = Widgets.newButton(composite,"Save");
      Widgets.layout(widgetButtonSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetButtonSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          saveKeys();
          saveColors();
          saveFonts();

          Settings.editors                                = getEditors();
          Settings.shellCommands                          = getShellCommands();
          Settings.showDiffCommand                        = widgetShowDiffCommand.getText();

          Settings.commandMail                            = widgetCommandLineMail.getText();
          Settings.commandMailAttachment                  = widgetCommandLineMailAttachment.getText().trim();

          Settings.mailSMTPHost                           = widgetMailSMTPHost.getText().trim();
          Settings.mailSMTPPort                           = Integer.parseInt(widgetMailSMTPPort.getText());
          Settings.mailSMTPSSL                            = widgetMailSMTPSSL.getSelection();
          Settings.mailLogin                              = widgetMailLogin.getText().trim();
          onzen.setPassword(Settings.mailLogin,Settings.mailSMTPHost,widgetMailPassword.getText());
          Settings.mailFrom                               = widgetMailFrom.getText().trim();

          Settings.cvsCommand                             = widgetCVSCommand.getText().trim();
          Settings.cvsPruneEmtpyDirectories               = widgetCVSPruneEmptyDirectories.getSelection();

          Settings.svnCommands.clear();
          for (TableItem tableItem : widgetSVNCommands.getItems())
          {
            SVNCommand svnCommand = (SVNCommand)tableItem.getData();
            Settings.svnCommands.set(svnCommand.version,svnCommand.command);
          }

          Settings.svnDiffCommand                         = widgetSVNDiffCommand.getText().trim();
          Settings.svnDiffCommandOptions                  = widgetSVNDiffCommandOptions.getText().trim();
          Settings.svnDiffCommandOptionsIgnoreWhitespaces = widgetSVNDiffCommandOptionsIgnoreWhitespaces.getText().trim();
          Settings.svnAlwaysTrustServerCertificate        = widgetSVNAlwaysTrustServerCertificate.getSelection();

          Settings.hgCommand                              = widgetHGCommand.getText().trim();
          Settings.hgDiffCommand                          = widgetHGDiffCommand.getText().trim();
          Settings.hgDiffCommandOptions                   = widgetHGDiffCommandOptions.getText().trim();
          Settings.hgDiffCommandOptionsIgnoreWhitespaces  = widgetHGDiffCommandOptionsIgnoreWhitespaces.getText().trim();
          Settings.hgUseForestExtension                   = widgetHGUseForestExtension.getSelection();
          Settings.hgUseQueueExtension                    = widgetHGUseQueueExtension.getSelection();
          Settings.hgUpdateWithFetch                      = widgetHGUpdateWithFetch.getSelection();
          Settings.hgSafeUpdate                           = widgetHGSafeUpdate.getSelection();
          Settings.hgSingleLineCommitMessages             = widgetHGSingleLineCommitMessages.getSelection();
          Settings.hgSingleLineCommitMessages             = widgetHGSingleLineCommitMessages.getSelection();
          Settings.hgRelativePatchPaths                   = widgetHGRelativePatchPaths.getSelection();

          Settings.gitCommand                             = widgetGITCommand.getText().trim();
          Settings.gitShortCommitNumbers                  = widgetGITShortCommitNumbers.getSelection();
          Settings.gitStashOnUpdate                       = widgetGITStashOnUpdate.getSelection();
          Settings.gitAutoPullSubModules                  = widgetGITAutoPullSubModules.getSelection();

          Settings.tmpDirectory                           = widgetTmpDirectory.getText().trim();
          Settings.backupFileSuffix                       = widgetBackupFileSuffix.getText().trim();
          Settings.dateFormat                             = widgetDateFormat.getText().trim();
          Settings.timeFormat                             = widgetTimeFormat.getText().trim();
          Settings.dateTimeFormat                         = widgetDateTimeFormat.getText().trim();
          Settings.convertSpacesPerTAB                    = Integer.parseInt(widgetConvertSpacesPerTAB.getText());
          Settings.maxBackgroundTasks                     = Integer.parseInt(widgetMaxBackgroundTasks.getText());
          Settings.maxMessageHistory                      = Integer.parseInt(widgetMaxMessageHistory.getText());

          String[] passwordNames = widgetStoredPasswords.getItems();
          for (String name : onzen.getPasswordNames())
          {
            if (!ArrayUtils.contains(passwordNames,name)) onzen.removePassword(name);
          }
          RepositoryURL.setHistoryURLs(Widgets.getListItems(widgetRepositoryHistoryURLs,new RepositoryURL[0]));
          Settings.repositoryURLs                         = Widgets.getListItems(widgetRepositoryURLs,new RepositoryURL[0]);

          if      (widgetEOLAuto.getSelection()   ) Settings.eolType = Settings.EOLTypes.AUTO;
          else if (widgetEOLUnix.getSelection()   ) Settings.eolType = Settings.EOLTypes.UNIX;
          else if (widgetEOLMac.getSelection()    ) Settings.eolType = Settings.EOLTypes.MAC;
          else if (widgetEOLWindows.getSelection()) Settings.eolType = Settings.EOLTypes.WINDOWS;
          else                                      Settings.eolType = Settings.EOLTypes.AUTO;
          Settings.checkTABs                              = widgetCheckTABs.getSelection();
          Settings.checkTrailingWhitespaces               = widgetCheckTrailingWhitespaces.getSelection();
          Settings.autoTrimTrailingWhitespaces            = widgetAutoTrimTrailingWhitespaces.getSelection();

          Settings.messageBroadcastAddress                = widgetMessageBroadcastAddress.getText().trim();
          Settings.messageBroadcastPort                   = Integer.parseInt(widgetMessageBroadcastPort.getText());

          Settings.reviewServerHost                       = widgetReviewServerHost.getText().trim();
          Settings.reviewServerLogin                      = widgetReviewServerLogin.getText().trim();
//          onzen.setPassword(Settings.reviewServerLogin,Settings.reviewServerHost,widgetReviewServerPassword.getText());
          Settings.commandPostReviewServer                = widgetCommandLinePostReviewServer.getText().trim();
          Settings.commandUpdateReviewServer              = widgetCommandLineUpdateReviewServer.getText().trim();

          Settings.skipWhitespaceCheckFilePatterns        = getSkipWhitespaceCheckFilePatterns();
          Settings.hiddenFilePatterns                     = getHiddenFilePatterns();
          Settings.hiddenDirectoryPatterns                = getHiddenDirectoryPatterns();
          Settings.autoSummaryPatterns                    = getAutoSummaryPatterns();
          for (TableItem tableItem : widgetShowFlags.getItems())
          {
            Object data = tableItem.getData();
            if      (data == Settings.singleChangedFilesInstance   ) Settings.singleChangedFilesInstance    = tableItem.getChecked();
            else if (data == Settings.showQuitConfirmation         ) Settings.showQuitConfirmation          = tableItem.getChecked();
            else if (data == Settings.showUpdateStatusErrors       ) Settings.showUpdateStatusErrors        = tableItem.getChecked();
            else if (data == Settings.showExecuteErrors            ) Settings.showExecuteErrors             = tableItem.getChecked();
            else if (data == Settings.showRestartAfterConfigChanged) Settings.showRestartAfterConfigChanged = tableItem.getChecked();
            else if (data == Settings.showUpdateInfo               ) Settings.showUpdateInfo                = tableItem.getChecked();
            else if (data == Settings.showNewVersionInformation    ) Settings.showNewVersionInformation     = tableItem.getChecked();
          }

          Settings.setWindowLocation                      = widgetSetWindowLocation.getSelection();

          Settings.geometryPreferences = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetKeys.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });
    widgetColors.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });
    widgetFonts.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });

    // show dialog
    Dialogs.show(dialog,Settings.geometryPreferences,Settings.setWindowLocation);
  }

  /** run dialog
   */
  public void run()
  {
    if (!dialog.isDisposed())
    {
      if ((Boolean)Dialogs.run(dialog,false))
      {
        boolean saveSettings = true;
        if (Settings.isFileModified())
        {
          saveSettings = Dialogs.confirm(shell,
                                         "Confirmation",
                                         "Settings were modified externally.\nOverwrite settings?",
                                         "Overwrite",
                                         "Cancel",
                                         false
                                        );
        }
        if (saveSettings)
        {
          Settings.save();
        }

        synchronized(Settings.showRestartAfterConfigChanged)
        {
          if (Settings.showRestartAfterConfigChanged)
          {
// NYI ??? return showRestartAfterConfigChanged flag?
            if (Dialogs.confirm(shell,
                                "Confirmation",
                                "Some settings may become active only after restarting Onzen.\nRestart now?",
                                "Now",
                                "Later",
                                true
                               )
               )
            {
              Widgets.notify(shell,SWT.Close,Onzen.EXITCODE_RESTART);
            }
          }
        }
      }
    }
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "Preferences {}";
  }

  //-----------------------------------------------------------------------

  /** add shortcut keys to keys list
   */
  private void addKeys()
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor  = Settings.ValueAdapterKey.class.getDeclaredConstructor();
      SettingUtils.ValueAdapter valueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance();

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.ValueAdapterKey.class.isAssignableFrom(configValue.type()))
              {
                // get key
                int key = field.getInt(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // convert to string
                String string = (String)valueAdapter.toString(field.get(null));

                // add entry
                Widgets.addTableItem(widgetKeys,new int[]{key},name.substring(3),string);
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Onzen.printInternalError(exception);
    }
  }

  /** add colors to colors list
   */
  private void addColors()
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor  = Settings.ValueAdapterColor.class.getDeclaredConstructor();
      SettingUtils.ValueAdapter valueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance();

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.ValueAdapterColor.class.isAssignableFrom(configValue.type()))
              {
                // get color
                Settings.Color color = (Settings.Color)field.get(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // add entry
                color = color.clone();
                Widgets.addTableItem(widgetColors,color,name.substring(5));   // Note: remove prefix "Color"
                if (color.foreground != null) Widgets.setTableItemColor(widgetColors,color,1,new Color(null,color.foreground));
                if (color.background != null) Widgets.setTableItemColor(widgetColors,color,2,new Color(null,color.background));
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Onzen.printInternalError(exception);
    }
  }

  /** add fonts to fonts list
   */
  private void addFonts()
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor  = Settings.ValueAdapterFontData.class.getDeclaredConstructor();
      SettingUtils.ValueAdapter valueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance();

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.ValueAdapterFontData.class.isAssignableFrom(configValue.type()))
              {
                // get font data
                FontData fontData = (FontData)field.get(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // convert to string
                String string = (String)valueAdapter.toString(field.get(null));

                // add entry
                fontData = (fontData != null) ? new FontData(fontData.getName(),(int)fontData.getHeight(),fontData.getStyle()) : new FontData();
                Widgets.addTableItem(widgetFonts,fontData,name.substring(4),string);
                Widgets.setTableItemFont(widgetFonts,fontData,1,fontData);
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Onzen.printInternalError(exception);
    }
  }

  private void saveKeys()
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor  = Settings.ValueAdapterKey.class.getDeclaredConstructor();
      SettingUtils.ValueAdapter valueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance();

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.ValueAdapterKey.class.isAssignableFrom(configValue.type()))
              {
                // get key
                int key = field.getInt(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // set key
                name = name.substring(3);
                boolean found = false;
                for (TableItem tableItem : widgetKeys.getItems())
                {
                  if (name.equals(tableItem.getText(0)))
                  {
                    field.setInt(null,((int[])tableItem.getData())[0]);
                    found = true;
                    break;
                  }
                }
                if (!found)
                {
                  Onzen.printInternalError("Key %s not found in table!",name);
                }
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Onzen.printInternalError(exception);
    }
  }

  private void saveColors()
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor  = Settings.ValueAdapterColor.class.getDeclaredConstructor();
      SettingUtils.ValueAdapter valueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance();

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.ValueAdapterColor.class.isAssignableFrom(configValue.type()))
              {
                // get color
                Settings.Color color = (Settings.Color)field.get(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // set color
                name = name.substring(5);
                boolean found = false;
                for (TableItem tableItem : widgetColors.getItems())
                {
                  if (name.equals(tableItem.getText(0)))
                  {
                    field.set(null,tableItem.getData());
                    found = true;
                    break;
                  }
                }
                if (!found)
                {
                  Onzen.printInternalError("Color %s not found in table!",name);
                }
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Onzen.printInternalError(exception);
    }
  }

  private void saveFonts()
  {
    try
    {
      // instantiate config adapter class
      Constructor               constructor  = Settings.ValueAdapterFontData.class.getDeclaredConstructor();
      SettingUtils.ValueAdapter valueAdapter = (SettingUtils.ValueAdapter)constructor.newInstance();

      // get setting classes
      Class[] settingClasses = Settings.getSettingClasses();
      for (Class clazz : settingClasses)
      {
        for (Field field : clazz.getDeclaredFields())
        {
          for (Annotation annotation : field.getDeclaredAnnotations())
          {
            if      (annotation instanceof SettingValue)
            {
              SettingValue configValue = (SettingValue)annotation;
              if (Settings.ValueAdapterFontData.class.isAssignableFrom(configValue.type()))
              {
                // get font data
                FontData fontData = (FontData)field.get(null);

                // get name
                String name = (!configValue.name().isEmpty()) ? configValue.name() : field.getName();

                // set font
                name = name.substring(4);
                boolean found = false;
                for (TableItem tableItem : widgetFonts.getItems())
                {
                  if (name.equals(tableItem.getText(0)))
                  {
                    field.set(null,tableItem.getData());
                    found = true;
                    break;
                  }
                }
                if (!found)
                {
                  Onzen.printInternalError("Font %s not found in table!",name);
                }
              }
            }
          }
        }
      }
    }
    catch (Exception exception)
    {
      // cannot happen
      Onzen.printInternalError(exception);
    }
  }

  /** get editors array from widget
   * @return editors array
   */
  private Settings.Editor[] getEditors()
  {
    Settings.Editor[] editors = new Settings.Editor[widgetEditors.getItemCount()];
    for (int z = 0; z < widgetEditors.getItemCount(); z++)
    {
      editors[z] = (Settings.Editor)(widgetEditors.getItem(z).getData());
    }

    return editors;
  }

  /** get shell commands array from widget
   * @return shell commands array
   */
  private Settings.ShellCommand[] getShellCommands()
  {
    Settings.ShellCommand[] shellCommands = new Settings.ShellCommand[widgetShellCommands.getItemCount()];
    for (int z = 0; z < widgetShellCommands.getItemCount(); z++)
    {
      shellCommands[z] = (Settings.ShellCommand)(widgetShellCommands.getItem(z).getData());
    }

    Arrays.sort(shellCommands,new Comparator<Settings.ShellCommand>()
    {
      public int compare(Settings.ShellCommand shellCommand1, Settings.ShellCommand shellCommand2)
      {
        return shellCommand1.name.compareTo(shellCommand2.name);
      }
    });

    return shellCommands;
  }

  /** get skip whitepace-check file pattern array from widget
   * @return skip whitespae-check file pattern array
   */
  private Settings.FilePattern[] getSkipWhitespaceCheckFilePatterns()
  {
    Settings.FilePattern[] filePatterns = new Settings.FilePattern[widgetSkipWhitespaceCheckFilePatterns.getItemCount()];
    for (int z = 0; z < widgetSkipWhitespaceCheckFilePatterns.getItemCount(); z++)
    {
      filePatterns[z] = new Settings.FilePattern(widgetSkipWhitespaceCheckFilePatterns.getItem(z));
    }

    return filePatterns;
  }

  /** get hidden file pattern array from widget
   * @return hidden file pattern array
   */
  private Settings.FilePattern[] getHiddenFilePatterns()
  {
    Settings.FilePattern[] filePatterns = new Settings.FilePattern[widgetHiddenFilePatterns.getItemCount()];
    for (int z = 0; z < widgetHiddenFilePatterns.getItemCount(); z++)
    {
      filePatterns[z] = new Settings.FilePattern(widgetHiddenFilePatterns.getItem(z));
    }

    return filePatterns;
  }

  /** get hidden directory pattern array from widget
   * @return hidden directory pattern array
   */
  private Settings.FilePattern[] getHiddenDirectoryPatterns()
  {
    Settings.FilePattern[] filePatterns = new Settings.FilePattern[widgetHiddenDirectoryPatterns.getItemCount()];
    for (int z = 0; z < widgetHiddenDirectoryPatterns.getItemCount(); z++)
    {
      filePatterns[z] = new Settings.FilePattern(widgetHiddenDirectoryPatterns.getItem(z));
    }

    return filePatterns;
  }

  /** get auto-summary pattern array from widget
   * @return auto-summary pattern array
   */
  private String[] getAutoSummaryPatterns()
  {
    String[] autoSummaryPatterns = new String[widgetAutoSummaryPatterns.getItemCount()];
    for (int z = 0; z < widgetAutoSummaryPatterns.getItemCount(); z++)
    {
      autoSummaryPatterns[z] = widgetAutoSummaryPatterns.getItem(z);
    }

    return autoSummaryPatterns;
  }

  /** edit keyboard shortcut
   * @param key key code
   * @return new key code
   */
  private int editKey(String name, int key)
  {
    /** dialog data
     */
    class Data
    {
      int key;

      Data()
      {
        this.key = 0;
      }
    };

    final Data data = new Data();

    Composite  composite,subComposite;
    Label      label;
    Text       text;
    Button     button;

    // set data
    data.key = key;

    // add editor dialog
    final Shell dialog = Dialogs.openModal(this.dialog,"Edit keyboard shortcut",300,SWT.DEFAULT,new double[]{1.0,0.0,0.0},1.0);

    final Text   widgetKey;
    final Button widgetSave;
    Listener     keyFilter;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Name:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      text = Widgets.newStringView(composite);
      text.setText(name);
      Widgets.layout(text,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Key:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,1,1,TableLayoutData.WE);
      {
        widgetKey = Widgets.newStringView(subComposite);
        widgetKey.setText(Widgets.menuAcceleratorToText(key));
        Widgets.layout(widgetKey,0,0,TableLayoutData.WE);

        button = Widgets.newButton(subComposite,"Clear");
        Widgets.layout(button,0,1,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            data.key = 0;
            widgetKey.setText("");
          }
        });
      }
    }

    label = Widgets.newLabel(dialog,"Note: type key, then select 'Save'");
    Widgets.layout(label,1,0,TableLayoutData.DEFAULT);

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,2,0,TableLayoutData.WE,0,0,4);
    {
      widgetSave = Widgets.newButton(composite,"Save");
      Widgets.layout(widgetSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    keyFilter = new Listener()
    {
      public void handleEvent(Event event)
      {
        data.key = event.stateMask+event.keyCode;
        widgetKey.setText(Widgets.menuAcceleratorToText(data.key));
      }
    };
    display.addFilter(SWT.KeyDown,keyFilter);

    if ((Boolean)Dialogs.run(dialog,false))
    {
      display.removeFilter(SWT.KeyDown,keyFilter);

      return data.key;
    }
    else
    {
      display.removeFilter(SWT.KeyDown,keyFilter);
      return key;
    }
  }

  /** edit color
   * @param name name
   * @param color color
   * @return true if edit OK, false on cancel
   */
  private boolean editColor(String name, final Settings.Color color)
  {
    Composite composite,subComposite;
    Label     label;
    Text      text;
    Button    button;

    // add editor dialog
    final Shell dialog = Dialogs.openModal(this.dialog,"Edit color",100,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Canvas widgetColorForeground;
    final Canvas widgetColorBackground;
    final Label  widgetValueForeground;
    final Label  widgetValueBackground;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Name:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      text = Widgets.newStringView(composite);
      text.setText(name);
      Widgets.layout(text,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Foreground:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,1,1,TableLayoutData.WE);
      {
        widgetColorForeground = Widgets.newCanvas(subComposite,SWT.BORDER);
        widgetColorForeground.setForeground((color.foreground != null) ? new Color(null,color.foreground) : null);
        widgetColorForeground.setBackground((color.foreground != null) ? new Color(null,color.foreground) : null);
        Widgets.layout(widgetColorForeground,0,0,TableLayoutData.WE,0,0,0,0,60,20);
        String colorName = (color.foreground != null)
                             ? String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue)
                             : "";
        widgetValueForeground = Widgets.newLabel(subComposite,colorName);
        Widgets.layout(widgetValueForeground,0,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
        widgetColorForeground.addMouseListener(new MouseListener()
        {
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
          }
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          public void mouseUp(MouseEvent mouseEvent)
          {
            ColorDialog colorDialog = new ColorDialog(dialog);
            colorDialog.setRGB(color.foreground);
            RGB rgb = colorDialog.open();
            if (rgb != null)
            {
              color.foreground = rgb;
              widgetColorForeground.setForeground(new Color(null,rgb));
              widgetColorForeground.setBackground(new Color(null,rgb));
              widgetValueForeground.setText(String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue));
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Background:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,2,1,TableLayoutData.WE);
      {
        widgetColorBackground = Widgets.newCanvas(subComposite,SWT.BORDER);
        widgetColorBackground.setForeground((color.background != null) ? new Color(null,color.background) : null);
        widgetColorBackground.setBackground((color.background != null) ? new Color(null,color.background) : null);
        Widgets.layout(widgetColorBackground,0,0,TableLayoutData.WE,0,0,0,0,60,20);
        String colorName = (color.background != null)
                             ? String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue)
                             : "";
        widgetValueBackground = Widgets.newLabel(subComposite,colorName);
        Widgets.layout(widgetValueBackground,0,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
        widgetColorBackground.addMouseListener(new MouseListener()
        {
          public void mouseDoubleClick(MouseEvent mouseEvent)
          {
          }
          public void mouseDown(MouseEvent mouseEvent)
          {
          }
          public void mouseUp(MouseEvent mouseEvent)
          {
            ColorDialog colorDialog = new ColorDialog(dialog);
            colorDialog.setRGB(color.background);
            RGB rgb = colorDialog.open();
            if (rgb != null)
            {
              color.background = rgb;
              widgetColorBackground.setForeground(new Color(null,rgb));
              widgetColorBackground.setBackground(new Color(null,rgb));
              widgetValueBackground.setText(String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue));
            }
          }
        });
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      button = Widgets.newButton(composite,"Save");
      Widgets.layout(button,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Default");
      Widgets.layout(button,0,1,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          color.foreground = color.DEFAULT_FOREGROUND;
          color.background = color.DEFAULT_BACKGROUND;

          widgetColorForeground.setForeground((color.foreground != null) ? new Color(null,color.foreground) : null);
          widgetColorForeground.setBackground((color.foreground != null) ? new Color(null,color.foreground) : null);
          widgetColorBackground.setForeground((color.background != null) ? new Color(null,color.background) : null);
          widgetColorBackground.setBackground((color.background != null) ? new Color(null,color.background) : null);

          String foregroundColorName = (color.foreground != null) ? String.format("#%02x%02x%02x",color.foreground.red,color.foreground.green,color.foreground.blue) : "";
          String backgroundColorName = (color.background != null) ? String.format("#%02x%02x%02x",color.background.red,color.background.green,color.background.blue) : "";
          widgetValueForeground.setText(foregroundColorName);
          widgetValueBackground.setText(backgroundColorName);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,2,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners

    return (Boolean)Dialogs.run(dialog,false);
  }

  /** edit editor command
   * @param editor editor command
   * @param title title text
   * @param buttonText button text
   * @return true if edit OK, false on cancel
   */
  private boolean editEditor(final Settings.Editor editor, String title, String buttonText)
  {
    Composite composite,subComposite;
    Label     label;
    Button    button;

    // add editor dialog
    final Shell dialog = Dialogs.openModal(this.dialog,title,500,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Text    widgetName;
    final Text    widgetMimeType;
    final Text    widgetFileName;
    final Text    widgetCommandLine;
    final Spinner widgetValidExitcode;
    final Button  widgetAddSave;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Name:");
      Widgets.layout(label,0,0,TableLayoutData.W);

      widgetName = Widgets.newText(composite);
      widgetName.setToolTipText("Name of editor or empty.\n");
      widgetName.setText(editor.name);
      Widgets.layout(widgetName,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Mime type:");
      Widgets.layout(label,1,0,TableLayoutData.W);

      widgetMimeType = Widgets.newText(composite);
      widgetMimeType.setToolTipText("Mime type pattern. Format: <type>/<sub-type>\n");
      widgetMimeType.setText(editor.mimeType);
      Widgets.layout(widgetMimeType,1,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"File name:");
      Widgets.layout(label,2,0,TableLayoutData.W);

      widgetFileName = Widgets.newText(composite);
      widgetFileName.setToolTipText("Simple file file name pattern, e. g. *.pdf.\n");
      widgetFileName.setText(editor.fileName);
      Widgets.layout(widgetFileName,2,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Command:");
      Widgets.layout(label,3,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,3,1,TableLayoutData.WE);
      {
        widgetCommandLine = Widgets.newText(subComposite);
        widgetCommandLine.setToolTipText("Command to run.\nMacros:\n  %file% - file name\n  %directory% - directory name\n  %directories% - directory names\n  %n% - line number\n  %% - %");
        widgetCommandLine.setText(editor.commandLine);
        Widgets.layout(widgetCommandLine,0,0,TableLayoutData.WE);

        button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
        Widgets.layout(button,0,1,TableLayoutData.E);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            String fileName = Dialogs.fileOpen(shell,
                                               "Select program",
                                               widgetCommandLine.getText(),
                                               new String[]{"Scripts",    "*.sh",
                                                            "Batch files","*.cmd",
                                                            "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                            "All files",Onzen.ALL_FILE_EXTENSION,
                                                           }
                                              );
            if (fileName != null)
            {
              widgetCommandLine.setText(fileName);
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Valid exitcode:");
      Widgets.layout(label,4,0,TableLayoutData.W);

      widgetValidExitcode = Widgets.newSpinner(composite);
      widgetValidExitcode.setSelection(editor.validExitcode);
      widgetValidExitcode.setMinimum(0);
      widgetValidExitcode.setMaximum(255);
      Widgets.layout(widgetValidExitcode,4,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetAddSave = Widgets.newButton(composite,buttonText);
      Widgets.layout(widgetAddSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetAddSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          editor.name          = widgetName.getText().trim();
          editor.mimeType      = widgetMimeType.getText().trim();
          editor.fileName      = widgetFileName.getText().trim();
          editor.commandLine   = widgetCommandLine.getText().trim();
          editor.validExitcode = widgetValidExitcode.getSelection();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners

    // focus traversal
    Widgets.setNextFocus(widgetName,
                         widgetMimeType,
                         widgetFileName,
                         widgetValidExitcode,
                         widgetCommandLine,
                         widgetAddSave
                        );

    // run dialog
    Widgets.setFocus(widgetName);
    return (Boolean)Dialogs.run(dialog,false);
  }

  /** edit shell command
   * @param editor editor command
   * @return true if edit OK, false on cancel
   */
  private boolean editShellCommand(final Settings.ShellCommand shellCommand, String title, String buttonText)
  {
    Composite composite,subComposite;
    Label     label;
    Button    button;

    // add editor dialog
    final Shell dialog = Dialogs.openModal(this.dialog,title,300,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Text    widgetName;
    final Text    widgetCommandLine;
    final Spinner widgetValidExitcode;
    final Button  widgetAddSave;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Name:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      widgetName = Widgets.newText(composite);
      widgetName.setToolTipText("Name of command");
      widgetName.setText(shellCommand.name);
      Widgets.layout(widgetName,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Command:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,1,1,TableLayoutData.WE);
      {
        widgetCommandLine = Widgets.newText(subComposite);
        widgetCommandLine.setToolTipText("Command to run.\nMacros:\n  %file% - file name\n  %directory% - directory name\n  %rootPath% - root path\n  %% - %");
        widgetCommandLine.setText(shellCommand.commandLine);
        Widgets.layout(widgetCommandLine,0,0,TableLayoutData.WE);

        button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
        Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            String command = Dialogs.fileOpen(shell,
                                              "Select command",
                                              widgetCommandLine.getText(),
                                              new String[]{"Scripts",    "*.sh",
                                                           "Batch files","*.cmd",
                                                           "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                           "All files",  Onzen.ALL_FILE_EXTENSION
                                                          }
                                             );
            if (command != null)
            {
              widgetCommandLine.setText(command);
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Valid exitcode:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0}));
      Widgets.layout(subComposite,2,1,TableLayoutData.WE);
      {
        label = Widgets.newLabel(subComposite,"<=");
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetValidExitcode = Widgets.newSpinner(subComposite);
        widgetValidExitcode.setSelection(shellCommand.validExitcode);
        widgetValidExitcode.setMinimum(0);
        widgetValidExitcode.setMaximum(255);
        Widgets.layout(widgetValidExitcode,0,1,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetAddSave = Widgets.newButton(composite,buttonText);
      Widgets.layout(widgetAddSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetAddSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          shellCommand.name          = widgetName.getText().trim();
          shellCommand.commandLine   = widgetCommandLine.getText();
          shellCommand.validExitcode = widgetValidExitcode.getSelection();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetName.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        widgetCommandLine.setFocus();
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });
    widgetCommandLine.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        widgetAddSave.setFocus();
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });

    // focus traversal
    Widgets.setNextFocus(widgetName,
                         widgetValidExitcode,
                         widgetAddSave
                        );

    // run dialog
    Widgets.setFocus(widgetName);
    return (Boolean)Dialogs.run(dialog,false);
  }

  /** edit SVN command
   * @param svnCommandData svn command
   * @param title title text
   * @param buttonText button text
   * @return true if edit OK, false on cancel
   */
  private boolean editSVNCommand(final SVNCommand svnCommand, String title, String buttonText)
  {
    Composite composite,subComposite;
    Label     label;
    Button    button;

    // add editor dialog
    final Shell dialog = Dialogs.openModal(this.dialog,title,500,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Text   widgetVersion;
    final Text   widgetCommand;
    final Button widgetAddSave;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Version:");
      Widgets.layout(label,0,0,TableLayoutData.W);

      widgetVersion = Widgets.newText(composite);
      widgetVersion.setToolTipText("Version of SVN command.\n");
      widgetVersion.setText(svnCommand.version);
      Widgets.layout(widgetVersion,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"Command:");
      Widgets.layout(label,1,0,TableLayoutData.W);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,1,1,TableLayoutData.WE);
      {
        widgetCommand = Widgets.newText(subComposite);
        widgetCommand.setText(svnCommand.command);
        Widgets.layout(widgetCommand,0,0,TableLayoutData.WE);

        button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
        Widgets.layout(button,0,1,TableLayoutData.E);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            String fileName = Dialogs.fileOpen(shell,
                                               "Select program",
                                               widgetCommand.getText(),
                                               new String[]{"Scripts",    "*.sh",
                                                            "Batch files","*.cmd",
                                                            "Executables",Onzen.EXECUTABLE_FILE_EXTENSION,
                                                            "All files",Onzen.ALL_FILE_EXTENSION
                                                           }
                                              );
            if (fileName != null)
            {
              widgetCommand.setText(fileName);
            }
          }
        });
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetAddSave = Widgets.newButton(composite,buttonText);
      Widgets.layout(widgetAddSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetAddSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          svnCommand.version = widgetVersion.getText().trim();
          svnCommand.command = widgetCommand.getText().trim();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners

    // focus traversal
    Widgets.setNextFocus(widgetVersion,
                         widgetCommand
                        );

    // run dialog
    Widgets.setFocus(widgetVersion);
    return (Boolean)Dialogs.run(dialog,false);
  }

  /** edit password
   * @param name name string
   * @param title dialog title
   * @return new password or null
   */
  private String editPassword(final String name, String title)
  {
    /** dialog data
     */
    class Data
    {
      String password;

      Data()
      {
        this.password = null;
      }
    }

    Composite composite;
    Label     label;
    Button    button;

    final Data  data   = new Data();
    final Shell dialog = Dialogs.openModal(this.dialog,title,300,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Label  widgetName;
    final Text   widgetPassword;
    final Button widgetSave;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Name:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      widgetName = Widgets.newView(composite);
      widgetName.setText(name);
      Widgets.layout(widgetName,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"New password:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      widgetPassword = Widgets.newText(composite);
      widgetPassword.setToolTipText("New password to store");
      Widgets.layout(widgetPassword,1,1,TableLayoutData.WE);
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetSave = Widgets.newButton(composite,"Save");
      Widgets.layout(widgetSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.password = widgetPassword.getText().trim();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners

    // focus traversal
    Widgets.setNextFocus(widgetPassword,
                         widgetSave
                        );

    // run dialog
    Widgets.setFocus(widgetPassword);
    if ((Boolean)Dialogs.run(dialog,false))
    {
      return data.password;
    }
    else
    {
      return null;
    }
  }

  /** edit repository URL
   * @param repositoryURL repository URL
   * @param title dialog title
   * @param buttonText button text
   * @return true if edit OK, false on cancel
   */
  private boolean editRepositoryURL(final RepositoryURL repositoryURL, String title, String buttonText)
  {
    /** dialog data
     */
    class Data
    {
      Repository.Types type;
      String           path;
      String           moduleName;
      String           userName;
      String           password;

      Data()
      {
        this.type       = Repository.Types.NONE;
        this.path       = "";
        this.moduleName = "";
        this.userName   = "";
        this.password   = "";
      }
    }

    Composite composite,subComposite;
    Label     label;
    Button    button;

    final Data  data   = new Data();
    final Shell dialog = Dialogs.openModal(this.dialog,title,300,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Text   widgetPath;
    final Text   widgetModuleName;
    final Text   widgetUserName;
    final Text   widgetPassword;
    final Button widgetAddSave;

    data.type       = repositoryURL.repositoryType;
    data.path       = repositoryURL.path;
    data.moduleName = repositoryURL.moduleName;
    data.userName   = repositoryURL.userName;
//    data.password   = repositoryURL.password;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Type:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,null));
      Widgets.layout(subComposite,0,1,TableLayoutData.WE);
      {
        button = Widgets.newRadio(subComposite,"CVS");
        button.setSelection(data.type == Repository.Types.CVS);
        Widgets.layout(button,0,0,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;

            if (widget.getSelection())
            {
              synchronized(data)
              {
                data.type = Repository.Types.CVS;
                Widgets.modified(data);
              }
            }
          }
        });

        button = Widgets.newRadio(subComposite,"SVN");
        button.setSelection(data.type == Repository.Types.SVN);
        Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;

            if (widget.getSelection())
            {
              synchronized(data)
              {
                data.type = Repository.Types.SVN;
                Widgets.modified(data);
              }
            }
          }
        });

        button = Widgets.newRadio(subComposite,"HG");
        button.setSelection(data.type == Repository.Types.HG);
        Widgets.layout(button,0,2,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;

            if (widget.getSelection())
            {
              synchronized(data)
              {
                data.type = Repository.Types.HG;
                Widgets.modified(data);
              }
            }
          }
        });

        button = Widgets.newRadio(subComposite,"GIT");
        button.setSelection(data.type == Repository.Types.GIT);
        Widgets.layout(button,0,3,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button widget = (Button)selectionEvent.widget;

            if (widget.getSelection())
            {
              synchronized(data)
              {
                data.type = Repository.Types.GIT;
                Widgets.modified(data);
              }
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Path:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,1,1,TableLayoutData.WE);
      {
        widgetPath = Widgets.newText(subComposite);
        widgetPath.setToolTipText("Path to repository");
        widgetPath.setText(data.path);
        Widgets.layout(widgetPath,0,0,TableLayoutData.WE);

        button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
        Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            String path = Dialogs.path(dialog,"Edit repository path","Path:",widgetPath.getText());
            if (path != null)
            {
              widgetPath.setText(path.trim());
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Module:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      widgetModuleName = Widgets.newText(composite);
      widgetModuleName.setToolTipText("Branch name");
      widgetModuleName.setText(data.moduleName);
      Widgets.layout(widgetModuleName,2,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"User name:");
      Widgets.layout(label,3,0,TableLayoutData.W);
      widgetUserName = Widgets.newText(composite);
      widgetUserName.setToolTipText("User name");
      widgetUserName.setText(data.userName);
      Widgets.layout(widgetUserName,3,1,TableLayoutData.WE);

      /*
      label = Widgets.newLabel(composite,"Password:");
      Widgets.layout(label,4,0,TableLayoutData.W);
      widgetPassword = Widgets.newText(composite);
      widgetPassword.setToolTipText("Repository access password.");
//      widgetPassword.setText(data.name);
      Widgets.layout(widgetPassword,4,1,TableLayoutData.WE);
      */
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetAddSave = Widgets.newButton(composite,buttonText);
      widgetAddSave.setEnabled(   (data.type != Repository.Types.NONE)
                               && !data.path.isEmpty()
                              );
      Widgets.layout(widgetAddSave,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetAddSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.path       = widgetPath.getText().trim();
          data.moduleName = widgetModuleName.getText().trim();
          data.userName   = widgetUserName.getText().trim();
//          data.password   = widgetPassword.getText();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    Widgets.addModifyListener(new WidgetModifyListener(widgetAddSave,data)
    {
      public void modified(Control control)
      {
        control.setEnabled(   (data.type != Repository.Types.NONE)
                           && !widgetPath.getText().trim().isEmpty()
                          );
      }
    });
    widgetPath.addModifyListener(new ModifyListener()
    {
      public void modifyText(ModifyEvent modifyEvent)
      {
        widgetAddSave.setEnabled(   (data.type != Repository.Types.NONE)
                                 && !widgetPath.getText().trim().isEmpty()
                                );
      }
    });

    // focus traversal
    Widgets.setNextFocus(widgetPath,
                         widgetModuleName,
                         widgetUserName,
                         widgetAddSave
                        );

    // run dialog
    Widgets.setFocus(widgetPath);
    if ((Boolean)Dialogs.run(dialog,false))
    {
      repositoryURL.repositoryType = data.type;
      repositoryURL.path           = data.path;
      repositoryURL.moduleName     = data.moduleName;
      repositoryURL.userName       = data.userName;
//      repositoryURL.password = data.type;

      return true;
    }
    else
    {
      return false;
    }
  }

  /** send test mail to mail server
   * @param mailSMTPHost mail server name
   * @param mailSMTPPort mail server port number
   * @param mailSMTPSSL true iff SSL encrypted connection
   * @param mailLogin login name
   * @param mailPassword login password
   * @param mailFrom mail from-address
   * @param toAddress mail to-address
   * @return true if edit OK, false on cancel
   */
  private void sendTestMail(String        mailSMTPHost,
                            int           mailSMTPPort,
                            boolean       mailSMTPSSL,
                            final String  mailLogin,
                            final String  mailPassword,
                            String        mailFrom,
                            String        toAddress
                           )
  {
    final String TEST_MAIL_TEMPLATE =
    "Test mail:\n" +
    "\n" +
    "  SMTP host: ${SMTPHost}\n" +
    "  SMTP port: ${SMTPPort}\n" +
    "  SMTP SSL: ${SMTPSSL}\n" +
    "  Login name: ${login}\n" +
    "  Mail from: ${from}\n" +
    "\n" +
    "";

    // create mail text
    Macro macro = new Macro(TEST_MAIL_TEMPLATE);
    macro.expand("SMTPHost", mailSMTPHost                  );
    macro.expand("SMTPPort", Integer.toString(mailSMTPPort));
    macro.expand("SMTPSSL",  mailSMTPSSL ? "yes" : "no"    );
    macro.expand("login",    mailLogin                     );
    macro.expand("from",     mailFrom                      );

    try
    {
      // create mail session
      Properties properties = new Properties();
      properties.put("mail.transport.protocol","smtp");
      properties.put("mail.smtp.host",mailSMTPHost);
      properties.put("mail.smtp.port",Integer.toString(mailSMTPPort));
      if (!mailFrom.isEmpty()) properties.put("mail.from",mailFrom);
//properties.put("mail.smtp.starttls.enable","true");
      properties.put("mail.smtp.auth",(mailPassword != null) && !mailPassword.isEmpty());
      if (mailSMTPSSL)
      {
        properties.put("mail.smtp.socketFactory.port",Integer.toString(mailSMTPPort));
        properties.put("mail.smtp.socketFactory.class","javax.net.ssl.SSLSocketFactory");
        properties.put("mail.smtp.socketFactory.fallback","false");
      }

//properties.put("mail.smtps.starttls.enable","true");
//                  if (Settings.debugFlag) properties.put("mail.debug","true");
      Authenticator auth = new Authenticator()
      {
        public PasswordAuthentication getPasswordAuthentication()
        {
          return new PasswordAuthentication(mailLogin,mailPassword);
        }
      };
      Session session = Session.getInstance(properties,auth);

      // create message
      MimeMultipart mimeMultipart = new MimeMultipart();

      MimeBodyPart text = new MimeBodyPart();
      text.setText(macro.getValue());
      text.setDisposition(MimeBodyPart.INLINE);
      mimeMultipart.addBodyPart(text);

      Message message = new MimeMessage(session);
      if (!mailFrom.isEmpty()) message.setFrom(new InternetAddress(mailFrom));
      message.setSubject("Onzen test mail");
      message.setRecipient(Message.RecipientType.TO,new InternetAddress(toAddress));
      message.setSentDate(new Date());
      message.setContent(mimeMultipart);
      message.saveChanges();

      // send message
      Transport.send(message);

      Dialogs.info(dialog,"Information","Test mail successfully sent!");
    }
    catch (MessagingException exception)
    {
exception.printStackTrace();
Dprintf.dprintf("exception=%s",exception);
      Dialogs.error(dialog,"Cannot send test mail (error: %s)",exception.getMessage());
    }
  }
}

/* end of file */
