/***********************************************************************\
*
* $Revision: 1040 $
* $Date: 2012-04-16 11:28:09 +0200 (Mo, 16. Apr 2012) $
* $Author: trupp $
* Contents: repository functions for "directory"
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Set;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlList;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.bind.annotation.XmlValue;
import javax.xml.bind.annotation.XmlSeeAlso;
import javax.xml.bind.annotation.adapters.XmlAdapter;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

/****************************** Classes ********************************/

/** directory repository
 */
class RepositoryDirectory extends Repository
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------
 private final static RepositoryDirectory staticInstance = new RepositoryDirectory();

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get static instance
   * @return static instance
   */
  public final static RepositoryDirectory getInstance()
  {
    return staticInstance;
  }

  /** create repository
   * @param rootPath root path
   */
  RepositoryDirectory(String rootPath, String comment)
  {
    super(rootPath,comment);

    assert !rootPath.isEmpty();
  }

  /** create repository
   */
  RepositoryDirectory()
  {
    super();
  }

  /** get repository type
   * @return repository type
   */
  @Override
  public Types getType()
  {
    return Types.DIRECTORY;
  }

  /** set commit message
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void setCommitMessage(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
  }

  /** check if repository support setting file mode
   * @return true iff file modes are supported
   */
  @Override
  public boolean supportSetFileMode()
  {
    return true;
  }

  /** list repository content
   * @param name name
   * @return repository file list
   */
  @Override
  public List<RepositoryFile> list(String name)
    throws RepositoryException
  {
return null;
  }

  /** create new repository module
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param importPath import directory
   */
  @Override
  public void create(String repositoryURL, String moduleName, String importPath)
    throws RepositoryException
  {
  }

  /** checkout repository from server
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param revisionId revision to checkout
   * @param userName user name or ""
   * @param destinationPath destination path
   * @param busyDialog busy dialog or null
   */
  @Override
  public void checkout(String repositoryURL, String moduleName, String revisionId, String userName, String destinationPath, BusyDialog busyDialog)
    throws RepositoryException
  {
  }

  /** update file states
   * @param fileDataSet file data set to update
   * @param fileDirectorySet relative directory name set to check for new/missing files
   * @param newFileDataSet new file data set or null
   */
  @Override
  public void updateStates(Set<FileData> fileDataSet, Set<String> fileDirectorySet, Set<FileData> newFileDataSet)
    throws RepositoryException
  {
  }

  /** get repository URL
   * @return repository URL
   */
  @Override
  public String getRepositoryURL()
  {
    return "";
  }

  /** get first revision name
   * @return first revision name
   */
  @Override
  public String getFirstRevision()
  {
    return "";
  }

  /** get last revision name
   * @return last revision name
   */
  @Override
  public String getLastRevision()
  {
    return "";
  }

  /** get current revision
   * @paran repositoryURL repository server URL
   * @return current revision or null
   */
  public String getCurrentRevision(String repositoryURL)
    throws RepositoryException
  {
    return null;
  }

  /** get revision names of file
   * @param name file name or URL
   * @return array with revision names
   */
  @Override
  public List<String> getRevisionNames(String name)
    throws RepositoryException
  {
    return new ArrayList<String>();
  }

  /** get revision data
   * @param fileData file data
   * @param revisionId revision
   * @return always null
   */
  @Override
  public RevisionData getRevisionData(FileData fileData, String revisionId)
    throws RepositoryException
  {
    assert revisionId != null;

    return null;
  }

  /** get revision data tree
   * @param fileData file data
   * @param busyDialog busy dialog or null
   * @return revision data tree
   */
  @Override
  public RevisionDataTree getRevisionDataTree(FileData fileData, BusyDialog busyDialog)
    throws RepositoryException
  {
    return new RevisionDataTree();
  }

  /** get file data (text lines)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  @Override
  public String[] getFileLines(FileData fileData, String revisionId)
    throws RepositoryException
  {
    return new String[0];
  }

  /** get file data (byte array)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  @Override
  public byte[] getFileBytes(FileData fileData, String revisionId)
    throws RepositoryException
  {
    return new byte[0];
  }

  /** get all changed/unknown files
   * @param stateSet state set
   * @return fileDataSet file data set with modified files
   */
  @Override
  public Set<FileData> getChangedFiles(EnumSet<FileData.States> stateSet)
    throws RepositoryException
  {
    return new HashSet<FileData>();
  }

  /** get diff of file
   * @param fileData file data
   * @param revisionId1,revisionId2 revisions to get diff for
   * @return diff data
   */
  @Override
  public List<DiffData> getDiff(FileData fileData, String oldRevision, String newRevision)
    throws RepositoryException
  {
    return new ArrayList<DiffData>();
  }

  /** get unified patch lines for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @param output patch output or null
   * @param lineList patch data lines or null
   */
  @Override
  public void getPatch(Set<FileData> fileDataSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces, PrintWriter output, List<String> lineList)
    throws RepositoryException
  {
  }

  /** get patch data for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @return patch data bytes
   */
  @Override
  public byte[] getPatchBytes(Set<FileData> fileDataSet, String revisionId1, String revisionId2)
    throws RepositoryException
  {
    return new byte[0];
  }

  /** get log to file
   * @param fileData file data
   * @return log array
   */
  @Override
  public List<LogData> getLog(FileData fileData)
    throws RepositoryException
  {
    return new ArrayList<LogData>();
  }

  /** get annotations to file
   * @param fileData file data
   * @param revisionId revision
   * @return annotation list
   */
  @Override
  public List<AnnotationData> getAnnotations(FileData fileData, String revisionId)
    throws RepositoryException
  {
    return new ArrayList<AnnotationData>();
  }

  /** update file from respository
   * @param fileDataSet file data set or null
   * @param busyDialog busy dialog or null
   */
  @Override
  public void update(Set<FileData> fileDataSet, BusyDialog busyDialog)
    throws RepositoryException
  {
  }

  /** commit files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void commit(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
  }

  /** add files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   * @param binaryFlag true to add file as binary files, false otherwise
   */
  @Override
  public void add(Set<FileData> fileDataSet, CommitMessage commitMessage, boolean binaryFlag)
    throws RepositoryException
  {
  }

  /** remove files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void remove(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
  }

  /** copy files
   * @param fileDataSet files to copy
   * @param destination destination
   * @param commitMessage commit message
   */
  @Override
  public void copy(Set<FileData> fileDataSet, String destination, CommitMessage commitMessage)
    throws RepositoryException
  {
Dprintf.dprintf("TODO");
  }

  /** rename file
   * @param fileData file data to rename
   * @param newName new name
   * @param commitMessage commit message
   */
  @Override
  public void rename(FileData fileData, String newName, CommitMessage commitMessage)
    throws RepositoryException
  {
    // rename local file
    File oldFile = fileData.getFile();
    File newFile = new File(rootPath,newName);
    if (!newFile.exists())
    {
      if (!oldFile.renameTo(newFile))
      {
        throw new RepositoryException("Rename file '%s' to '%s' fail",oldFile.getName(),newFile.getName());
      }
    }
    else
    {
      throw new RepositoryException("File '%s' already exists",newFile.getName());
    }
  }

  /** revert files
   * @param fileDataSet file data set or null for all files
   * @param revisionId revision to revert to
   * @param recursive true for recursive revert, false otherwise
   */
  @Override
  public void revert(Set<FileData> fileDataSet, String revisionId, boolean recursive)
    throws RepositoryException
  {
Dprintf.dprintf("");
throw new Error("NYI");
  }

  /** set conflicts resolved
   * @param fileDataSet file data set or null for all files
   */
  @Override
  public void resolve(Set<FileData> fileDataSet)
    throws RepositoryException
  {
  }

  /** set files mode
   * @param fileDataSet file data set
   * @param mode file mode
   * @param commitMessage commit message
   */
  @Override
  public void setFileMode(Set<FileData> fileDataSet, FileData.Modes mode, CommitMessage commitMessage)
    throws RepositoryException
  {
  }

  /** get default name of root
   * @return always null (no root name)
   */
  @Override
  public String getDefaultRootName()
  {
    return null;
  }

  /** get default branch/tag name
   * @return default branch/tag name
   */
  @Override
  public String getDefaultBranchTagName()
  {
    return "";
  }

  /** get names of existing branches/tags
   * @return branch/tag name list (always empty)
   */
  @Override
  public List<String> getBranchTagNames(String pathName)
    throws RepositoryException
  {
    return new ArrayList<String>();
  }

  /** create new branch
   * @param rootName root name (source)
   * @param branchName branch name
   * @param fromRevision from revision name or null
   * @param commitMessage commit message
   * @param buysDialog busy dialog or null
   */
  @Override
  public void createBranch(String rootName, String branchName, String fromRevision, CommitMessage commitMessage, BusyDialog busyDialog)
    throws RepositoryException
  {
    // nothing to do
  }

  //-----------------------------------------------------------------------

}

/* end of file */
