/***********************************************************************\
*
* $Revision: 2176 $
* $Date: 2017-11-02 08:20:25 +0100 (Thu, 02 Nov 2017) $
* $Author: torsten $
* Contents: repository tab
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;

import java.text.SimpleDateFormat;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Set;
import java.util.WeakHashMap;

// graphics
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.TreeEvent;
import org.eclipse.swt.events.TreeListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeColumn;
import org.eclipse.swt.widgets.TreeItem;

/****************************** Classes ********************************/

/** repository tab
 */
class RepositoryTab
{
  /** file data comparator
   */
  static class FileDataComparator implements Comparator<FileData>
  {
//TODO
    enum X
    {
      a,
      b
    };
    // Note: enum in inner classes are not possible in Java, thus use the old way...
    private final static int SORTMODE_NAME = 0;
    private final static int SORTMODE_TYPE = 1;
    private final static int SORTMODE_SIZE = 2;
    private final static int SORTMODE_DATE = 3;

    private int sortMode;

    /** create file data comparator
     * @param sortMode sort mode
     */
    FileDataComparator(int sortMode)
    {
      this.sortMode = sortMode;
    }

    /** create file data comparator
     * @param tree file tree
     * @param sortColumn column to sort
     */
    FileDataComparator(Tree tree, TreeColumn sortColumn)
    {
      if      (tree.getColumn(0) == sortColumn) sortMode = SORTMODE_NAME;
      else if (tree.getColumn(1) == sortColumn) sortMode = SORTMODE_TYPE;
      else if (tree.getColumn(2) == sortColumn) sortMode = SORTMODE_SIZE;
      else if (tree.getColumn(3) == sortColumn) sortMode = SORTMODE_DATE;
      else                                      sortMode = SORTMODE_NAME;
    }

    /** create file data comparator
     * @param tree file tree
     */
    FileDataComparator(Tree tree)
    {
      this(tree,tree.getSortColumn());
    }

    /** compare file tree data without taking care about type
     * @param fileData1, fileData2 file tree data to compare
     * @return -1 iff fileData1 < fileData2,
                0 iff fileData1 = fileData2,
                1 iff fileData1 > fileData2
     */
    private int compareWithoutType(FileData fileData1, FileData fileData2)
    {
      switch (sortMode)
      {
        case SORTMODE_NAME:
          return fileData1.getName().compareTo(fileData2.getName());
        case SORTMODE_TYPE:
          return fileData1.type.compareTo(fileData2.type);
        case SORTMODE_SIZE:
          if      (fileData1.size < fileData2.size) return -1;
          else if (fileData1.size > fileData2.size) return  1;
          else                                      return  0;
        case SORTMODE_DATE:
          if      (fileData1.date.before(fileData2.date)) return -1;
          else if (fileData1.date.after(fileData2.date))  return  1;
          else                                            return  0;
        default:
          return 0;
      }
    }

    /** compare file tree data
     * @param fileData1, fileData2 file tree data to compare
     * @return -1 iff fileData1 < fileData2,
                0 iff fileData1 = fileData2,
                1 iff fileData1 > fileData2
     */
    public int compare(FileData fileData1, FileData fileData2)
    {
      if ((fileData1 != null) && (fileData2 != null))
      {
        if (fileData1.type == FileData.Types.DIRECTORY)
        {
          if (fileData2.type == FileData.Types.DIRECTORY)
          {
            return compareWithoutType(fileData1,fileData2);
          }
          else
          {
            return -1;
          }
        }
        else
        {
          if (fileData2.type == FileData.Types.DIRECTORY)
          {
            return 1;
          }
          else
          {
            return compareWithoutType(fileData1,fileData2);
          }
        }
      }
      else
      {
        return 0;
      }
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      return "FileDataComparator {"+sortMode+"}";
    }
  }

  // --------------------------- constants --------------------------------

  // colors
  public static Color COLOR_OK;
  public static Color COLOR_UNKNOWN;
  public static Color COLOR_MODIFIED;
  public static Color COLOR_CHECKOUT;
  public static Color COLOR_UPDATE;
  public static Color COLOR_MERGE;
  public static Color COLOR_CONFLICT;
  public static Color COLOR_ADDED;
  public static Color COLOR_REMOVED;
  public static Color COLOR_NOT_EXISTS;
  public static Color COLOR_WAITING;
  public static Color COLOR_ERROR;
  public static Color COLOR_UPDATE_STATUS;
  public static Color COLOR_IGNORE;

  // convert whitespaces modes
  public enum ConvertWhitespacesModes
  {
    TABS,
    TRAILING
  };

  // convert whitespaces results
  public enum ConvertWhitespacesResult
  {
    OK,
    ALL,
    FAIL,
    ABORTED
  };

  // --------------------------- variables --------------------------------
  // repository
  public  final Repository    repository;           // repository instance of tab

  // global variable references
  private final Shell         shell;
  private final Display       display;
  private final Clipboard     clipboard;

  protected final Onzen       onzen;

  // widgets
  private final TabFolder     widgetTabFolder;
  public  final Composite     widgetComposite;
  private final Tree          widgetFileTree;
  private final Menu          widgetHeaderMenu,widgetBodyMenu;
  private final Menu          menuOpenFilesWithCommands;
  private final Menu          menuShellCommands;

  // last push/pull repository URL
  private RepositoryURL lastRepositoryURL = new RepositoryURL();

  // last copy destination directory
  private static String       lastDirectory = "";

  // last convert whitespaces dialog location
  private static Point        convertWhiteDialogLocation = new Point(SWT.DEFAULT,SWT.DEFAULT);

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create repository tab
   * @param onzen Onzen instance
   * @param tabFolder tab folder
   * @param leftRepositoryTab repository tab to left or null
   * @param repository repository
   */
  RepositoryTab(final Onzen onzen, TabFolder tabFolder, RepositoryTab leftRepositoryTab, Repository repository)
  {
    Menu       menu,subMenu;
    MenuItem   menuItem;
    Composite  composite;
    Label      label;
    Button     button;
    TreeColumn treeColumn;
    TreeItem   treeItem;
    Text       text;

    // initialize variables
    this.onzen           = onzen;
    this.repository      = repository;
    this.widgetTabFolder = tabFolder;

    // get shell, display, clipboard
    shell     = tabFolder.getShell();
    display   = shell.getDisplay();
    clipboard = new Clipboard(display);

    // init colors
    COLOR_OK            = new Color(display,Settings.colorStatusOK.background          );
    COLOR_UNKNOWN       = new Color(display,Settings.colorStatusUnknown.background     );
    COLOR_MODIFIED      = new Color(display,Settings.colorStatusModified.background    );
    COLOR_CHECKOUT      = new Color(display,Settings.colorStatusCheckout.background    );
    COLOR_UPDATE        = new Color(display,Settings.colorStatusUpdate.background      );
    COLOR_MERGE         = new Color(display,Settings.colorStatusMerge.background       );
    COLOR_CONFLICT      = new Color(display,Settings.colorStatusConflict.background    );
    COLOR_ADDED         = new Color(display,Settings.colorStatusAdded.background       );
    COLOR_REMOVED       = new Color(display,Settings.colorStatusRemoved.background     );
    COLOR_NOT_EXISTS    = new Color(display,Settings.colorStatusNotExists.background   );
    COLOR_WAITING       = new Color(display,Settings.colorStatusWaiting.background     );
    COLOR_ERROR         = new Color(display,Settings.colorStatusError.background       );
    COLOR_UPDATE_STATUS = new Color(display,Settings.colorStatusUpdateStatus.foreground);
    COLOR_IGNORE        = new Color(display,Settings.colorStatusIgnore.foreground      );

    // create tab
    widgetComposite = Widgets.insertTab(tabFolder,
                                        (leftRepositoryTab != null) ? leftRepositoryTab.widgetComposite : null,
                                        repository.title,
                                        this,
                                        SWT.NONE
                                       );
//widgetComposite.setBackground(Onzen.COLOR_YELLOW);
//Dprintf.dprintf("");
//widgetComposite.setLayout(new TableLayout(1.0,1.0));
    Widgets.layout(widgetComposite,0,0,TableLayoutData.NSWE);
    {
      // file tree
      final String FILES_COLUMN_NAMES[] = new String[]{"Name","Status","Working revision","Revision","Branch"};
      widgetFileTree = Widgets.newTree(widgetComposite,SWT.MULTI);
      Widgets.layout(widgetFileTree,0,0,TableLayoutData.NSWE);
      SelectionListener fileTreeColumnSelectionListener = new SelectionListener()
      {
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          TreeColumn         treeColumn         = (TreeColumn)selectionEvent.widget;
          FileDataComparator fileDataComparator = new FileDataComparator(widgetFileTree,treeColumn);

          synchronized(widgetFileTree)
          {
            Widgets.sortTreeColumn(widgetFileTree,treeColumn,fileDataComparator);
          }
        }
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
      };
      Listener fileTreeColumnResizeListener = new Listener()
      {
        public void handleEvent(Event event)
        {
          Settings.geometryMainColumns = new Settings.ColumnSizes(Widgets.getTreeColumnWidth(widgetFileTree));
        }
      };
      treeColumn = Widgets.addTreeColumn(widgetFileTree,"Name",            SWT.LEFT,390,true);
      treeColumn.setToolTipText("Click to sort for name.");
      treeColumn.addSelectionListener(fileTreeColumnSelectionListener);
      treeColumn.addListener(SWT.Resize,fileTreeColumnResizeListener);
      treeColumn = Widgets.addTreeColumn(widgetFileTree,"Status",          SWT.LEFT,160,true);
      treeColumn.setToolTipText("Click to sort for status.");
      treeColumn.addSelectionListener(fileTreeColumnSelectionListener);
      treeColumn.addSelectionListener(fileTreeColumnSelectionListener);
      treeColumn.addListener(SWT.Resize,fileTreeColumnResizeListener);
      treeColumn = Widgets.addTreeColumn(widgetFileTree,"Working revision",SWT.LEFT,100,true);
      treeColumn.setToolTipText("Click to sort for working revision.");
      treeColumn.addSelectionListener(fileTreeColumnSelectionListener);
      treeColumn.addListener(SWT.Resize,fileTreeColumnResizeListener);
      treeColumn = Widgets.addTreeColumn(widgetFileTree,"Revision",        SWT.LEFT,100,true);
      treeColumn.setToolTipText("Click to sort for revision.");
      treeColumn.addSelectionListener(fileTreeColumnSelectionListener);
      treeColumn.addListener(SWT.Resize,fileTreeColumnResizeListener);
      treeColumn = Widgets.addTreeColumn(widgetFileTree,"Branch",          SWT.LEFT,100,true);
      treeColumn.setToolTipText("Click to sort for branch.");
      treeColumn.addSelectionListener(fileTreeColumnSelectionListener);
      treeColumn.addListener(SWT.Resize,fileTreeColumnResizeListener);

      Widgets.setTreeColumnWidth(widgetFileTree,Settings.geometryMainColumns.width);
      widgetFileTree.setColumnOrder(Settings.geometryMainColumnOrder.getMap(FILES_COLUMN_NAMES));
/*
      widgetFileTree.addListener(SWT.PaintItem, new Listener()
      {
        public void handleEvent(Event event)
        {
          final int HEIGHT = Onzen.IMAGE_LOCK.getBounds().height;

          TreeItem item = (TreeItem)event.item;

          if (event.index == 1)
          {
            Object data = item.getData();
            if (data instanceof FileData)
            {
              FileData fileData = (FileData)item.getData();

              if (fileData.locked)
              {
                event.gc.drawImage(Onzen.IMAGE_LOCK, event.x+event.width, event.y+(event.height-HEIGHT)/2);
              }
            }
          }
        }
      });
*/
      Listener filesColumnResizeListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Settings.geometryMainColumns = new Settings.ColumnSizes(Widgets.getTreeColumnWidth(widgetFileTree));
        }
      };
      Listener filesColumnMoveListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          int[]    columnOrder = widgetFileTree.getColumnOrder();
          String[] names       = new String[columnOrder.length];
          for (int i = 0; i < columnOrder.length; i++)
          {
            names[i] = FILES_COLUMN_NAMES[columnOrder[i]];
          }
          Settings.geometryMainColumnOrder = new SettingUtils.SimpleStringArray(names);
        }
      };
      for (TreeColumn orderTableColumn : widgetFileTree.getColumns())
      {
        orderTableColumn.setMoveable(true);
        orderTableColumn.addListener(SWT.Move,filesColumnMoveListener);
        orderTableColumn.addListener(SWT.Resize,filesColumnResizeListener);
      }

      widgetHeaderMenu = Widgets.newPopupMenu(shell);
      {
        menuItem = Widgets.addMenuItemCheckbox(widgetHeaderMenu,"Name");
        menuItem.setSelection(Settings.geometryMainColumns.get(0) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetFileTree,0,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetHeaderMenu,"Status");
        menuItem.setSelection(Settings.geometryMainColumns.get(1) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetFileTree,1,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetHeaderMenu,"Working revision");
        menuItem.setSelection(Settings.geometryMainColumns.get(2) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetFileTree,2,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetHeaderMenu,"Revision");
        menuItem.setSelection(Settings.geometryMainColumns.get(3) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetFileTree,3,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetHeaderMenu,"Branch");
        menuItem.setSelection(Settings.geometryMainColumns.get(4) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTreeColumn(widgetFileTree,4,widget.getSelection());
          }
        });
      }

      widgetBodyMenu = Widgets.newPopupMenu(shell);
      {
        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Open file"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            FileData fileData = getSelectedFileData();
            if (fileData != null)
            {
              openFile(fileData);
            }
          }
        });

        menuOpenFilesWithCommands = Widgets.addMenu(widgetBodyMenu,"Open files with");//,Settings.keyOpenFilesWith);
        menuItem = Widgets.addMenuItem(menuOpenFilesWithCommands,"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            FileData[] fileData = getSelectedFilesData();
            if (fileData != null)
            {
              openFilesWith(fileData);
            }
          }
        });

        menuShellCommands = Widgets.addMenu(widgetBodyMenu,"Shell");
        Widgets.addMenuItemSeparator(menuShellCommands);
        menuItem = Widgets.addMenuItem(menuShellCommands,"Add new command"+"\u2026");
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            onzen.addShellCommand();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Copy to"+"\u2026",Settings.keyCopyFilesTo);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            copyFilesTo();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Open as new repository");
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            openRepositoryTab();
          }
        });

        menuItem = Widgets.addMenuItemSeparator(widgetBodyMenu);

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Update",Settings.keyUpdate);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            update();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Update all",Settings.keyUpdateAll);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            updateAll();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Commit"+"\u2026",Settings.keyCommit);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            commit();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Create patch"+"\u2026",Settings.keyCreatePatch);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            createPatch();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Revert"+"\u2026",Settings.keyRevert);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            revert();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Resolve"+"\u2026",Settings.keyResolve);
// NYI
menuItem.setEnabled(false);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
Dprintf.dprintf("NYI");
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Set resolved",Settings.keySetResolved);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            setResolved();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Lock",Settings.keyLock);
        menuItem.setEnabled(repository.supportLockUnlock());
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            lock();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Unlock",Settings.keyUnlock);
        menuItem.setEnabled(repository.supportLockUnlock());
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            unlock();
          }
        });

        menuItem = Widgets.addMenuItemSeparator(widgetBodyMenu);

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Add"+"\u2026",Settings.keyAdd);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            add();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Remove"+"\u2026",Settings.keyRemove);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            remove();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Copy"+"\u2026",Settings.keyCopy);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            copy();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Rename/move"+"\u2026",Settings.keyRename);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            rename();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Set file mode"+"\u2026",Settings.keySetFileMode);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            setFileMode();
          }
        });

        menuItem = Widgets.addMenuItemSeparator(widgetBodyMenu);

        menuItem = Widgets.addMenuItemSeparator(widgetBodyMenu);

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Revisions"+"\u2026",Settings.keyRevisions);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            revisions();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Revision info"+"\u2026",Settings.keyRevisionInfo);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            revisionInfo();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Diff"+"\u2026",Settings.keyDiff);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            diff();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Annotations"+"\u2026",Settings.keyAnnotations);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            annotations();
          }
        });

        menuItem = Widgets.addMenuItemSeparator(widgetBodyMenu);

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"New file"+"\u2026",Settings.keyNewFile);
menuItem.setEnabled(false);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            newFile();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"New directory"+"\u2026",Settings.keyNewDirectory);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            newDirectory();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Rename/move local file/directory"+"\u2026",Settings.keyRenameLocal);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            renameLocalFile();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Delete local files/directories"+"\u2026",Settings.keyDeleteLocal);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            deleteLocalFiles();
          }
        });

        menuItem = Widgets.addMenuItemSeparator(widgetBodyMenu);

        subMenu = Widgets.addMenu(widgetBodyMenu,"Add to");
        {
          menuItem = Widgets.addMenuItem(subMenu,"skip convert whitespaces list",Settings.keyAddSkipConvertWhitespacesFile);
          menuItem.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              addSkipConvertWhitespacesFiles();
            }
          });

          menuItem = Widgets.addMenuItem(subMenu,"ignore list",Settings.keyAddIgnoreFile);
          menuItem.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              addIgnoreFiles();
            }
          });

          menuItem = Widgets.addMenuItem(subMenu,"hidden list",Settings.keyAddHiddenFile);
          menuItem.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              addHiddenFiles();
            }
          });
        }

        subMenu = Widgets.addMenu(widgetBodyMenu,"Remove from");
        {
          menuItem = Widgets.addMenuItem(subMenu,"skip convert whitespaces list",Settings.keyRemoveSkipConvertWhitespacesFile);
          menuItem.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              removeSkipConvertWhitespacesFiles();
            }
          });

          menuItem = Widgets.addMenuItem(subMenu,"ignore list",Settings.keyRemoveIgnoreFile);
          menuItem.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              removeIgnoreFiles();
            }
          });
        }

        menuItem = Widgets.addMenuItemSeparator(widgetBodyMenu);

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Copy file names to clipbaord");
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            copyFileNamesToClipboard();
          }
        });

        menuItem = Widgets.addMenuItem(widgetBodyMenu,"Convert whitespaces"+"\u2026",Settings.keyConvertWhitespaces);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            convertWhitespaces();
          }
        });
      }
      widgetFileTree.setMenu(widgetBodyMenu);

      widgetFileTree.addListener(SWT.MenuDetect, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Tree widget = (Tree)event.widget;

          Point     p          = display.map(null, widget, new Point(event.x, event.y));
          Rectangle clientArea = widget.getClientArea();
          boolean   isInHeader =    (clientArea.y <= p.y)
                                 && (p.y < (clientArea.y + widget.getHeaderHeight()));
          widget.setMenu(isInHeader ? widgetHeaderMenu : widgetBodyMenu);
        }
      });
      widgetFileTree.addListener(SWT.Dispose, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          widgetHeaderMenu.dispose();
          widgetBodyMenu.dispose();
        }
      });
    }

    // add root directory
    addRootDirectory();
  }

  /** create repository tab
   * @param onzen Onzen instance
   * @param tabFolder parent tab folder
   * @param repository repository
   */
  RepositoryTab(Onzen onzen, TabFolder tabFolder, Repository repository)
  {
    this(onzen,tabFolder,null,repository);
  }

  /** get repository
   * @param rootPath repository root path
   * @return repository or null if not found
   */
  public Repository getRepository(String rootPath)
  {
    return onzen.getRepository(rootPath);
  }

  /** get repository tab
   * @param rootPath repository root path
   * @return repository tab or null if not found
   */
  public RepositoryTab getRepositoryTab(String rootPath)
  {
    return onzen.getRepositoryTab(rootPath);
  }

  /** get title
   */
  public String getTitle()
  {
    return repository.title;
  }

  /** set title
   * @param title title text to set
   */
  public void setTitle(String title)
  {
    repository.title = title;
    Widgets.setTabTitle(widgetTabFolder,widgetComposite,title);
  }

  /** close repository tab
   */
  public void close()
  {
    Widgets.removeTab(widgetTabFolder,widgetComposite);
  }

  /** move repository tab
   * @param index tab index (0..n-1)
   */
  public void move(int index)
  {
    Widgets.moveTab(widgetTabFolder,widgetComposite,index);
  }

  /** show repository tab
   */
  public void show()
  {
    Widgets.showTab(widgetTabFolder,widgetComposite);
  }

  /** open directory in file-tree
   * @param directory directory to open
   * @param setSelection true to set selection on opened tree item
   * @return openened/selected tree item or null
   */
  public TreeItem openDirectory(String directory, String selectName)
  {
    if (!widgetFileTree.isDisposed())
    {
      return openSubDirectory(widgetFileTree,widgetFileTree.getItems(),directory,selectName);
    }
    else
    {
      return null;
    }
  }

  /** open directory in file-tree
   * @param directory directory to open
   * @return openened/selected tree item or null
   */
  public TreeItem openDirectory(String directory)
  {
    return openDirectory(directory,null);
  }

  /** open directory in file-tree
   * @param directory directory to open
   * @param setSelection true to set selection on opened tree item
   * @return openened/selected tree item or null
   */
  public TreeItem openDirectory(File file)
  {
    return openDirectory(repository.getRelativeFileName(file.getParent()),
                         file.isFile()
                           ? repository.getRelativeFileName(file)
                           : null
                        );
  }

  /** set status text
   * @param format format string
   * @param arguments optional arguments
   */
  public void setStatusText(String format, Object... arguments)
  {
    onzen.setStatusText(format,arguments);
  }

  /** clear status text
   */
  public void clearStatusText()
  {
    onzen.clearStatusText();
  }

  /** update open-files-with commands in context menu
   */
  public void updateOpenFilesWithCommands()
  {
    MenuItem menuItem;

    if (!menuShellCommands.isDisposed())
    {
      // remove old entries in open-files-with command menu
      MenuItem[] menuItems = menuOpenFilesWithCommands.getItems();
      for (int i = 0; i < menuItems.length-1; i++)
      {
        menuItems[i].dispose();
      }

      // add new open-file-with commands to menu
      HashMap<String,Settings.Editor> commandMap = new HashMap<String,Settings.Editor>();
      for (Settings.Editor editor : Settings.editors)
      {
        String name = !editor.name.isEmpty() ? editor.name : editor.commandLine;
        commandMap.put(name,editor);
      }

      ArrayList<String> nameList = new ArrayList<String>(commandMap.keySet());
      Collections.sort(nameList);
      for (int i = 0; i < nameList.size(); i++)
      {
        menuItem = Widgets.addMenuItem(menuOpenFilesWithCommands,i,nameList.get(i),SWT.NONE);
        menuItem.setData(commandMap.get(nameList.get(i)));
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem        widget = (MenuItem)selectionEvent.widget;
            Settings.Editor editor = (Settings.Editor)widget.getData();

            openFilesWith(editor);
          }
        });
      }
    }
  }

  /** update shell commands in context menu
   */
  public void updateShellCommands()
  {
    MenuItem menuItem;

    if (!menuShellCommands.isDisposed())
    {
      // remove old entries in shell command menu
      MenuItem[] menuItems = menuShellCommands.getItems();
      for (int i = 0; i < menuItems.length; i++)
      {
        menuItems[i].dispose();
      }

      // add new shell commands to menu
      for (Settings.ShellCommand shellCommand : Settings.shellCommands)
      {
        menuItem = Widgets.addMenuItem(menuShellCommands,shellCommand.name);
        menuItem.setData(shellCommand);
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem              widget       = (MenuItem)selectionEvent.widget;
            Settings.ShellCommand shellCommand = (Settings.ShellCommand)widget.getData();

            executeShellCommand(shellCommand);
          }
        });
      }
      Widgets.addMenuItemSeparator(menuShellCommands);
      menuItem = Widgets.addMenuItem(menuShellCommands,"Add new command"+"\u2026");
      menuItem.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          onzen.addShellCommand();
        }
      });
    }
  }

  //-----------------------------------------------------------------------

  /** update root path
   */
  public void updateRootPath()
  {
    if (widgetFileTree.getItemCount() > 0)
    {
      widgetFileTree.getItem(0).setText(repository.rootPath);
    }
  }

  /** update states of selected entries/current directory
   */
  public void updateStates()
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    TreeItem[] selectedTreeItems = widgetFileTree.getSelection();
    if (selectedTreeItems.length > 0)
    {
      // get file data (selected and all in same directory)
      for (TreeItem selectedTreeItem : selectedTreeItems)
      {
        fileDataSet.add((FileData)selectedTreeItem.getData());

        TreeItem parentTreeItem = selectedTreeItem.getParentItem();
        if (parentTreeItem != null)
        {
          for (TreeItem treeItem : parentTreeItem.getItems())
          {
            fileDataSet.add((FileData)treeItem.getData());
          }
        }
      }
    }
    else
    {
      // get file data of all open directories
      for (TreeItem treeItem : Widgets.getAllTreeItems(widgetFileTree))
      {
        FileData fileData = (FileData)treeItem.getData();
        if (fileData != null) fileDataSet.add(fileData);
      }
    }

    // start update file data
    asyncUpdateFileStates(fileDataSet,true);
  }

  /** copy selected entries to directory
   */
  public void copyFilesTo()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();

    String directory = Dialogs.directory(shell,
                                         "Select destination directory","Copy "+fileDataSet.size()+" file(s)/directory to:",
                                         lastDirectory
                                        );
    if (directory != null)
    {
      final BusyDialog busyDialog = new BusyDialog(shell,
                                                   "Copy files to '"+directory+"'...",
                                                   400,200,
                                                   BusyDialog.LIST|BusyDialog.AUTO_ANIMATE
                                                  );
      try
      {
        for (FileData fileData : fileDataSet)
        {
          try
          {
            setStatusText("Copy file '%s' -> '%s'...",fileData.getName(repository),directory);

            File fromFile = fileData.getFile();
            File toFile   = new File(directory);

            File newFile = new File(toFile,fromFile.getName());
            if (newFile.exists())
            {
              if (!Dialogs.confirm(shell,"Confirmation",String.format("File\n\n'%s'\n\nexists - overwrite?",newFile.getPath())))
              {
                return;
              }
            }

            FileUtils.copyFile(fromFile,
                               toFile,
                               new FileCopyConfirmation()
                               {
                                 public boolean confirm(File file)
                                 {
                                   busyDialog.updateList(file.getPath());
                                   return true;
                                 }
                               }
                              );
            if (busyDialog.isAborted()) break;
          }
          catch (IOException exception)
          {
            Dialogs.error(shell,"Copy file '"+fileData.getName(repository)+"' fail (error: %s)",exception.getMessage());
            return;
          }
          finally
          {
            clearStatusText();
          }
        }
      }
      finally
      {
        busyDialog.close();
      }

      // start update file data
      asyncUpdateSelectedFileData();

      lastDirectory = directory;
    }
  }

  /** open directory as new repository tab
   * @return repository or null
   */
  public Repository openRepositoryTab()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      File file = fileData.getFile();
      String directoryName;
      if (file.isFile())
      {
        directoryName = file.getParent();
      }
      else
      {
        directoryName = file.getPath();
      }
      return onzen.openRepository(directoryName);
    }
    else
    {
      return null;
    }
  }

  /** update all open or selected entries
   */
  public void update()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();

    Dialogs.info(shell,
                 "Information",
                 Dialogs.booleanFieldUpdater(Settings.class,"showUpdateInfo"),
                 "Update some files may result into build errors\nwhen not all required files are up-to-date.\n\nUse 'Update all' instead."
                );

    try
    {
      setStatusText("Update files...");

      repository.update(fileDataSet);
    }
    catch (RepositoryException exception)
    {
      Dialogs.error(shell,exception.getExtendedMessage(),"Update fail (error: %s)",exception.getMessage());
      return;
    }
    finally
    {
      clearStatusText();
    }

    // start update file data
    asyncUpdateFileStates(fileDataSet);
  }

  /** update all or selected entries
   */
  public void updateAll()
  {
    final Set<FileData> fileDataSet = getAllFileDataSet();

    BusyDialog busyDialog = new BusyDialog(shell,
                                           "Update all directories and files...",
                                           400,200,
                                           BusyDialog.LIST|BusyDialog.AUTO_ANIMATE
                                          );

    Background.run(new BackgroundRunnable(repository,fileDataSet,busyDialog)
    {
      public void run(Repository repository, Set<FileData> fileDataSet, final BusyDialog busyDialog)
      {
        try
        {
          setStatusText("Update all directories and files...");

          repository.updateAll(busyDialog);
        }
        catch (RepositoryException exception)
        {
          final String   exceptionMessage         = exception.getMessage();
          final String[] exceptionExtendedMessage = exception.getExtendedMessage();
          if (!display.isDisposed())
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                if (!shell.isDisposed())
                {
                  Dialogs.error(shell,exceptionExtendedMessage,"Update all fail (error: %s)",exceptionMessage);
                }
              }
            });
          }
          return;
        }
        finally
        {
          display.syncExec(new Runnable()
          {
            public void run()
            {
              busyDialog.close();
            }
          });
          clearStatusText();
        }

        updateFileStates(fileDataSet,true);
      }
    });
  }

  /** commit selected entries
   */
  public void commit()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      CommandCommit commandCommit = new CommandCommit(shell,this,fileDataSet);
      commandCommit.run();
    }
  }

  /** create patch for selected entries
   */
  public void createPatch()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      CommandCreatePatch commandCreatePatch = new CommandCreatePatch(shell,this,fileDataSet);
      commandCreatePatch.run();
    }
  }

  /** show/edit patches
   */
  public void patches()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      CommandPatches commandPatches = new CommandPatches(shell,this);
      commandPatches.run();
    }
  }

  /** add selected entries
   */
  public void add()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      CommandAdd commandAdd = new CommandAdd(shell,this,fileDataSet);
      commandAdd.run();
    }
  }

  /** remove selected entries
   */
  public void remove()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      CommandRemove commandRemove = new CommandRemove(shell,this,fileDataSet);
      commandRemove.run();
    }
  }

  /** copy selected entries
   */
  public void copy()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      CommandCopy commandCopy = new CommandCopy(shell,this,fileDataSet);
      commandCopy.run();
    }
  }

  /** rename selected entry
   */
  public void rename()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      CommandRename commandRename = new CommandRename(shell,this,fileData);
      commandRename.run();
    }
  }

  /** resolve conflicts
   */
  public void revert()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      CommandRevert commandRevert = new CommandRevert(shell,this,fileDataSet);
      commandRevert.run();
    }
  }

  /** resolve conflicts
   */
  public void resolve()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
// NYI
Dprintf.dprintf("NYI");
//      CommandRename commandRename = new CommandRename(shell,this,fileData);
//      commandRename.run();
    }
  }

  /** set conflicts resolved
   */
  public void setResolved()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      try
      {
        setStatusText("Set conflicts resolved...");

        repository.resolve(fileDataSet);
      }
      catch (RepositoryException exception)
      {
        Dialogs.error(shell,exception.getExtendedMessage(),"Set conflicts resolved fail: %s",exception.getMessage());
        return;
      }
      finally
      {
        clearStatusText();
      }

/*???
      // get all entries in directories for update (files may be added or disappear after set solved)
      HashSet<FileData> updateFileDataSet = new HashSet<FileData>();
      for (FileData fileData : fileDataSet)
      {
        TreeItem parentTreeItem = fileData.treeItem.getParentItem();

        for (TreeItem treeItem : parentTreeItem.getItems())
        {
          FileData subFileData = (FileData)treeItem.getData();
           if (subFileData != null) updateFileDataSet.add(subFileData);
        }
      }

      // start update file data
      asyncUpdateFileStates(updateFileDataSet);
*/
      asyncUpdateFileStates(fileDataSet);
    }
  }


  /** change current branch
   */
  public void changeBranch()
  {
    try
    {
      setStatusText("Change branch...");

      String toBranchName = Dialogs.list(shell,
                                         "Change to branch",
                                         "Select branch:",
                                         new ListRunnable()
                                         {
                                           @Override
                                           public Collection<String> getValues()
                                           {
                                             try
                                             {
                                               return repository.getBranchTagNames();
                                             }
                                             catch (RepositoryException exception)
                                             {
                                               return null;
                                             }
                                           }
                                           @Override
                                           public String getSelection()
                                           {
                                             return repository.getCurrentBranchName();
                                           }
                                         },
                                         "Change"
                                        );
      if (toBranchName != null)
      {
        repository.changeBranch(toBranchName);

        asyncUpdateAllFileStates();
      }
    }
    catch (RepositoryException exception)
    {
      Dialogs.error(shell,exception.getExtendedMessage(),"Change branch fail: %s",exception.getMessage());
      return;
    }
    finally
    {
      clearStatusText();
    }
  }

  /** merge other branch/commit
   */
  public void merge()
  {
    CommandMerge commandMerge = new CommandMerge(shell,this);
    commandMerge.run();
  }

  /** show incoming changes
   * @param masterRepository master repository or ""
   */
  public void incomingChanges(String masterRepository)
  {
    CommandChanges commandChanges = new CommandChanges(shell,this,masterRepository,CommandChanges.ChangesTypes.INCOMING);
    commandChanges.run();
  }

  /** show incoming changes
   */
  public void incomingChanges()
  {
//TODO
    incomingChanges(repository.getMasterRepositoryURL().path);
  }

  /** show outgoing changes
   * @param masterRepository master repository or ""
   */
  public void outgoingChanges(String masterRepository)
  {
    CommandChanges commandChanges = new CommandChanges(shell,this,masterRepository,CommandChanges.ChangesTypes.OUTGOING);
    commandChanges.run();
  }

  /** show outgoing changes
   */
  public void outgoingChanges()
  {
//TODO
    outgoingChanges(repository.getMasterRepositoryURL().path);
  }

  /** pull changes
   * @param masterRepository master repository or ""
   * @param moduleName module name
   * @param userName user login name or ""
   */
  public void pullChanges(String masterRepository, String moduleName, String userName)
  {
    try
    {
      setStatusText("Pull changes...");

      RepositoryURL repositoryURL = new RepositoryURL(RepositoryURL.Types.MANUAL,
                                                      repository.getType(),
                                                      masterRepository
                                                     );
      repositoryURL.addToHistory();

      repository.pullChanges(masterRepository,moduleName,userName);
    }
    catch (RepositoryException exception)
    {
      Dialogs.error(shell,exception.getExtendedMessage(),"Pulling changes fail: %s",exception.getMessage());
      return;
    }
    finally
    {
      clearStatusText();
    }
  }

  /** pull changes
   */
  public void pullChanges()
  {
    /** dialog data
     */
    class Data
    {
      ArrayList<RepositoryURL> repositoryURLList;
      String                   path;
      String                   moduleName;
      String                   userName;

      Data()
      {
        this.repositoryURLList = new ArrayList<RepositoryURL>();
        this.path              = null;
        this.moduleName        = "";
        this.userName          = "";

        repositoryURLList.add(new RepositoryURL("file://"));
        repositoryURLList.add(new RepositoryURL("ssh://"));
        repositoryURLList.add(new RepositoryURL("http://"));
        repositoryURLList.add(new RepositoryURL("https://"));
        repositoryURLList.add(new RepositoryURL("rsync://"));
        for (RepositoryURL repositoryURL : RepositoryURL.getHistoryURLs())
        {
          repositoryURLList.add(repositoryURL);
        }
        for (RepositoryURL repositoryURL : Settings.repositoryURLs)
        {
          repositoryURLList.add(repositoryURL);
        }
      }
    }

    Composite composite,subComposite;
    Label     label;
    Button    button;

    final Data  data   = new Data();
    final Shell dialog = Dialogs.openModal(shell,"Select repository for pull changes",500,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Combo  widgetPath;
    final Combo  widgetModuleName;
    final Text   widgetUserName;
    final Button widgetPullChanges;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Repository:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,0,1,TableLayoutData.NSWE);
      {
        widgetPath = Widgets.newCombo(subComposite);
        widgetPath.setToolTipText("Respository path URI.");
        Widgets.layout(widgetPath,0,0,TableLayoutData.WE);

        button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
        Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            String path = Dialogs.directory(shell,
                                            "Select repository path",
                                            widgetPath.getText()
                                           );
            if (path != null)
            {
              widgetPath.setText(path);
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Module:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      widgetModuleName = Widgets.newCombo(composite);
      widgetModuleName.setToolTipText("Module name in repository.");
      widgetModuleName.setText(data.moduleName);
      Widgets.layout(widgetModuleName,1,1,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetModuleName,data)
      {
        public void modified(Combo combo)
        {
          combo.setText(data.moduleName);
        }
      });

      label = Widgets.newLabel(composite,"User name:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      widgetUserName = Widgets.newText(composite);
      widgetUserName.setToolTipText("Check-out user name.");
      widgetUserName.setText(data.userName);
      Widgets.layout(widgetUserName,2,1,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetUserName,data)
      {
        public void modified(Text text)
        {
          text.setText(data.userName);
        }
      });
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,0.0,1.0}));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetPullChanges = Widgets.newButton(composite,"Pull changes");
      widgetPullChanges.setEnabled(!widgetPath.getText().trim().isEmpty());
      Widgets.layout(widgetPullChanges,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetPullChanges.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          String path     = widgetPath.getText().trim();
          String userName = widgetUserName.getText();
          data.path       = path;
          data.moduleName = widgetModuleName.getText().trim();
          data.userName   = userName;

          lastRepositoryURL = new RepositoryURL(data.path,data.moduleName,data.userName);

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,4,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetPath.addModifyListener(new ModifyListener()
    {
      public void modifyText(ModifyEvent modifyEvent)
      {
        widgetPullChanges.setEnabled(!widgetPath.getText().trim().isEmpty());
      }
    });
    widgetPath.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        int index = widget.getSelectionIndex();
        if (index >= 0)
        {
          RepositoryURL repositoryURL = (RepositoryURL)Widgets.getComboItem(widget,index);

          synchronized(data)
          {
            widgetPath.setText(repositoryURL.path);
            widgetModuleName.setText(repositoryURL.moduleName);
            widgetUserName.setText(repositoryURL.userName);

            data.path       = repositoryURL.path;
            data.moduleName = repositoryURL.moduleName;
            data.userName   = repositoryURL.userName;
            Widgets.modified(data);
          }
        }
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        int index = widget.getSelectionIndex();
        if (index >= 0)
        {
          RepositoryURL repositoryURL = (RepositoryURL)Widgets.getComboItem(widget,index);

          synchronized(data)
          {
            widgetPath.setText(repositoryURL.path);
            widgetModuleName.setText(repositoryURL.moduleName);
            widgetUserName.setText(repositoryURL.userName);

            data.path       = repositoryURL.path;
            data.moduleName = repositoryURL.moduleName;
            data.userName   = repositoryURL.userName;
            Widgets.modified(data);
          }
        }
      }
    });

    // focus traversal
    Widgets.setNextFocus(widgetPath,
                         widgetModuleName,
                         widgetUserName,
                         widgetPullChanges
                        );

    // add repository URLs
    for (RepositoryURL repositoryURL : data.repositoryURLList)
    {
      Widgets.addComboItem(widgetPath,repositoryURL,repositoryURL.path);
    }
    widgetPath.setText(lastRepositoryURL.path);
    widgetModuleName.setText(lastRepositoryURL.moduleName);
    widgetUserName.setText(lastRepositoryURL.userName);

    // run dialog
    if ((Boolean)Dialogs.run(dialog,false))
    {
      pullChanges(data.path,data.moduleName,data.userName);
    }
  }

  /** push changes
   * @param masterRepository master repository or ""
   * @param moduleName module name
   * @param userName user login name or ""
   */
  public void pushChanges(String masterRepository, String moduleName, String userName)
  {
    try
    {
      setStatusText("Push changes...");

      RepositoryURL repositoryURL = new RepositoryURL(repository.getType(),
                                                      masterRepository,
                                                      moduleName,
                                                      userName
                                                     );
      repositoryURL.addToHistory();

      repository.pushChanges(masterRepository,moduleName,userName);
    }
    catch (RepositoryException exception)
    {
      Dialogs.error(shell,exception.getExtendedMessage(),"Pushing changes fail: %s",exception.getMessage());
      return;
    }
    finally
    {
      clearStatusText();
    }
  }

  /** push changes
   */
  public void pushChanges()
  {
    /** dialog data
     */
    class Data
    {
      ArrayList<RepositoryURL> repositoryURLList;
      String                   path;
      String                   moduleName;
      String                   userName;

      Data()
      {
        this.repositoryURLList = new ArrayList<RepositoryURL>();
        this.path              = null;
        this.moduleName        = "";
        this.userName          = "";

        repositoryURLList.add(new RepositoryURL("file://"));
        repositoryURLList.add(new RepositoryURL("ssh://"));
        repositoryURLList.add(new RepositoryURL("http://"));
        repositoryURLList.add(new RepositoryURL("https://"));
        repositoryURLList.add(new RepositoryURL("rsync://"));
        for (RepositoryURL repositoryURL : RepositoryURL.getHistoryURLs())
        {
          repositoryURLList.add(repositoryURL);
        }
        for (RepositoryURL repositoryURL : Settings.repositoryURLs)
        {
          repositoryURLList.add(repositoryURL);
        }
      }
    }

    Composite composite,subComposite;
    Label     label;
    Button    button;

    final Data  data   = new Data();
    final Shell dialog = Dialogs.openModal(shell,"Select repository for push changes",500,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Combo  widgetPath;
    final Combo  widgetModuleName;
    final Text   widgetUserName;
    final Button widgetPushChanges;

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Repository:");
      Widgets.layout(label,0,0,TableLayoutData.W);
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
      Widgets.layout(subComposite,0,1,TableLayoutData.NSWE);
      {
        widgetPath = Widgets.newCombo(subComposite);
        widgetPath.setToolTipText("Respository path URI.");
        Widgets.layout(widgetPath,0,0,TableLayoutData.WE);

        button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
        Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            String path = Dialogs.directory(shell,
                                            "Select repository path",
                                            widgetPath.getText()
                                           );
            if (path != null)
            {
              widgetPath.setText(path);
            }
          }
        });
      }

      label = Widgets.newLabel(composite,"Module:");
      Widgets.layout(label,1,0,TableLayoutData.W);
      widgetModuleName = Widgets.newCombo(composite);
      widgetModuleName.setToolTipText("Module name in repository.");
      widgetModuleName.setText(data.moduleName);
      Widgets.layout(widgetModuleName,1,1,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetModuleName,data)
      {
        public void modified(Combo combo)
        {
          combo.setText(data.moduleName);
        }
      });

      label = Widgets.newLabel(composite,"User name:");
      Widgets.layout(label,2,0,TableLayoutData.W);
      widgetUserName = Widgets.newText(composite);
      widgetUserName.setToolTipText("Check-out user name.");
      widgetUserName.setText(data.userName);
      Widgets.layout(widgetUserName,2,1,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetUserName,data)
      {
        public void modified(Text text)
        {
          text.setText(data.userName);
        }
      });
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,0.0,1.0}));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetPushChanges = Widgets.newButton(composite,"Push changes");
      widgetPushChanges.setEnabled(!widgetPath.getText().trim().isEmpty());
      Widgets.layout(widgetPushChanges,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetPushChanges.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          String path     = widgetPath.getText().trim();
          String userName = widgetUserName.getText();
          data.path       = path;
          data.moduleName = widgetModuleName.getText().trim();
          data.userName   = userName;

          lastRepositoryURL = new RepositoryURL(data.path,data.moduleName,data.userName);

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,4,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetPath.addModifyListener(new ModifyListener()
    {
      public void modifyText(ModifyEvent modifyEvent)
      {
        widgetPushChanges.setEnabled(!widgetPath.getText().trim().isEmpty());
      }
    });
    widgetPath.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        int index = widget.getSelectionIndex();
        if (index >= 0)
        {
          RepositoryURL repositoryURL = (RepositoryURL)Widgets.getComboItem(widget,index);

          synchronized(data)
          {
            widgetPath.setText(repositoryURL.path);
            widgetModuleName.setText(repositoryURL.moduleName);
            widgetUserName.setText(repositoryURL.userName);

            data.path       = repositoryURL.path;
            data.moduleName = repositoryURL.moduleName;
            data.userName   = repositoryURL.userName;
            Widgets.modified(data);
          }
        }
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        int index = widget.getSelectionIndex();
        if (index >= 0)
        {
          RepositoryURL repositoryURL = (RepositoryURL)Widgets.getComboItem(widget,index);

          synchronized(data)
          {
            widgetPath.setText(repositoryURL.path);
            widgetModuleName.setText(repositoryURL.moduleName);
            widgetUserName.setText(repositoryURL.userName);

            data.path       = repositoryURL.path;
            data.moduleName = repositoryURL.moduleName;
            data.userName   = repositoryURL.userName;
            Widgets.modified(data);
          }
        }
      }
    });

    // focus traversal
    Widgets.setNextFocus(widgetPath,
                         widgetModuleName,
                         widgetUserName,
                         widgetPushChanges
                        );

    // add repository URLs
    for (RepositoryURL repositoryURL : data.repositoryURLList)
    {
      Widgets.addComboItem(widgetPath,repositoryURL,repositoryURL.path);
    }
    widgetPath.setText(lastRepositoryURL.path);
    widgetModuleName.setText(lastRepositoryURL.moduleName);
    widgetUserName.setText(lastRepositoryURL.userName);

    // run dialog
    if ((Boolean)Dialogs.run(dialog,false))
    {
      pushChanges(data.path,data.moduleName,data.userName);
    }
  }

  /** apply patches
   */
  public void applyPatches()
  {
    try
    {
      setStatusText("Apply patches...");

      repository.applyPatches();
    }
    catch (RepositoryException exception)
    {
      Dialogs.error(shell,exception.getExtendedMessage(),"Apply patches fail: %s",exception.getMessage());
      return;
    }
    finally
    {
      clearStatusText();
    }
  }

  /** unapply patches
   */
  public void unapplyPatches()
  {
    try
    {
      setStatusText("Unapply patches...");

      repository.unapplyPatches();
    }
    catch (RepositoryException exception)
    {
      Dialogs.error(shell,exception.getExtendedMessage(),"Unapply patches fail: %s",exception.getMessage());
      return;
    }
    finally
    {
      clearStatusText();
    }
  }

  /** lock file
   */
  public void lock()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      try
      {
        setStatusText("Lock...");

        repository.lock(fileDataSet);
      }
      catch (RepositoryException exception)
      {
        Dialogs.error(shell,exception.getExtendedMessage(),"Lock fail: %s",exception.getMessage());
        return;
      }
      finally
      {
        clearStatusText();
      }

      // start update file data
      asyncUpdateSelectedFileData();
    }
  }

  /** unlock file
   */
  public void unlock()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      try
      {
        setStatusText("Unlock...");

        repository.unlock(fileDataSet);
      }
      catch (RepositoryException exception)
      {
        Dialogs.error(shell,exception.getExtendedMessage(),"Unlock fail: %s",exception.getMessage());
        return;
      }
      finally
      {
        clearStatusText();
      }

      // start update file data
      asyncUpdateSelectedFileData();
    }
  }

  /** set file mode of selected entries
   */
  public void setFileMode()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      CommandSetFileMode commandSetFileMode = new CommandSetFileMode(shell,this,fileDataSet);
      commandSetFileMode.run();
    }
  }

  /** create new branch/tag
   */
  public void newBranchTag()
  {
    CommandCreateBranch commandCreateBranch = new CommandCreateBranch(shell,this);
    commandCreateBranch.run();
  }

  /** view selected entry
   */
  public void view()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      CommandView commandView = new CommandView(shell,this,fileData);
      commandView.run();
    }
  }

  /** diff selected entry
   */
  public void diff()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      CommandDiff commandDiff = new CommandDiff(shell,this,fileData);
      commandDiff.run();
    }
  }

  /** show changes log
   */
  public void changesLog()
  {
    CommandChangesLog commandChangesLog = new CommandChangesLog(shell,this);
    commandChangesLog.run();
  }

  /** show revision tree of selected entry
   */
  public void revisions()
  {
    CommandRevisions commandRevisions = new CommandRevisions(shell,this,getSelectedFileData());
    commandRevisions.run();
  }

  /** show revision info of selected entry
   */
  public void revisionInfo()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      CommandRevisionInfo commandRevisionInfo = new CommandRevisionInfo(shell,this,fileData);
      commandRevisionInfo.run();
    }
  }

  /** show list of all changed files
   */
  public void changedFiles()
  {
    CommandChangedFiles commandChangedFiles = new CommandChangedFiles(shell,this);
    commandChangedFiles.run();
  }

  /** show annotatiosn of file
   */
  public void annotations()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      CommandAnnotations commandAnnotations = new CommandAnnotations(shell,this,fileData);
      commandAnnotations.run();
    }
  }

  /** open file (with external command)
   * @param file file
   * @param mimeType mime type pattern
   * @param lineNumber line number
   */
  public void openFile(File file, String mimeType, int lineNumber)
  {
    // find editor command with file mime-type or file name pattern
    Settings.Editor fileEditor = null;
    if (fileEditor == null)
    {
      if ((mimeType != null) && !mimeType.isEmpty() && !mimeType.equals("application/octet-stream"))
      {
        for (Settings.Editor editor : Settings.editors)
        {
          if (editor.mimeTypePattern.matcher(mimeType).matches())
          {
            fileEditor = editor;
            break;
          }
        }
      }
    }
    if (fileEditor == null)
    {
      String fileName = file.getPath();
      String baseName = file.getName();
      for (Settings.Editor editor : Settings.editors)
      {
        if (   editor.fileNamePattern.matcher(fileName).matches()
            || editor.fileNamePattern.matcher(baseName).matches()
           )
        {
          fileEditor = editor;
          break;
        }
      }
    }

    // if no editor command found -> ask for command
    if (fileEditor == null)
    {
      fileEditor  = getFileEditor(file,mimeType);
    }

    // execute external command
    if (fileEditor != null)
    {
      // expand command
      Macro macro = new Macro(StringUtils.split(new ArrayList<String>(),fileEditor.commandLine,StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS),Macro.PATTERN_PERCENTAGE);
//      if (!macro.contains("files") && !macro.contains("file")) macro.add("file");
      macro.expand("files",     new String[]{file.getPath()});
      macro.expand("file",      file.getPath());
      macro.expand("f",         file.getPath());
      macro.expand("lineNumber",lineNumber);
      macro.expand("n",         lineNumber);
      macro.expand("%%",        "%");
      Command command = new Command(macro);
//Dprintf.dprintf("command=%s",command);

      // run command
      Background.run(new BackgroundRunnable(command)
      {
        public void run(final Command command)
        {
          Process process = null;
          try
          {
            // start process (Note: use Runtime.exec() because it is a background process without i/o here)
            process = Runtime.getRuntime().exec(command.getCommandArray());

            // collect stderr output
            BufferedReader stderr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
            final ArrayList<String> stderrList = new ArrayList<String>();
            String line;
            while ((line = stderr.readLine()) != null)
            {
//Dprintf.dprintf("line=%s",line);
              stderrList.add(line);
            }

            // wait for exit
            final int exitCode = process.waitFor();
            if (exitCode != 0)
            {
              if (!display.isDisposed())
              {
                display.syncExec(new Runnable()
                {
                  public void run()
                  {
                    if (!shell.isDisposed())
                    {
                      Dialogs.error(shell,
                                    Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                                    stderrList,
                                    "Execute external command fail: \n\n'%s'\n\n (exitcode: %d)",
                                    command,
                                    exitCode
                                   );
                    }
                  }
                });
              }
              return;
            }

            process.destroy(); process = null;
          }
          catch (InterruptedException exception)
          {
            final String message = exception.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
//TODO: remove?
//                              Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                              "Execute external command fail: \n\n'%s'\n\n (error: %s)",
                              command,
                              message
                             );
              }
            });
            return;
          }
          catch (IOException exception)
          {
            final String message = Onzen.reniceIOException(exception).getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
//TODO: remove?
//                              Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                              "Execute external command fail: \n\n'%s'\n\n (error: %s)",
                              command,
                              message
                             );
              }
            });
            return;
          }
          catch (Exception exception)
          {
            final String message = exception.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
                              Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                              "Execute external command fail: \n\n'%s'\n\n (error: %s)",
                              command,
                              message
                             );
              }
            });
            return;
          }
          finally
          {
            if (process != null) process.destroy();
          }
        }
      });
    }
  }

  /** open file (with external command)
   * @param file file to open
   * @param mimeType mime type pattern
   */
  public void openFile(File file, String mimeType)
  {
    openFile(file,mimeType,0);
  }

  /** open file (with external command)
   * @param file file to open
   */
  public void openFile(File file)
  {
    openFile(file,Onzen.getMimeType(file));
  }

  /** open file (with external command)
   * @param fileName file to open
   * @param mimeType mime type pattern
   * @param lineNumber line number
   */
  public void openFile(String fileName, String mimeType, int lineNumber)
  {
    openFile(new File(fileName),mimeType,lineNumber);
  }

  /** open file (with external command)
   * @param fileName file name
   * @param mimeType mime type pattern
   */
  public void openFile(String fileName, String mimeType)
  {
    openFile(fileName,mimeType,0);
  }

  /** open file (with external command)
   * @param fileData file to open
   */
  public void openFile(FileData fileData)
  {
    openFile(fileData.getName(),
             fileData.getMimeType()
            );
  }

  /** open file (with external command)
   */
  public void openFile()
  {
    final FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      openFileWith(fileData);
    }
  }

  /** open files with external program
   * @param editor editor to open files with
   * @param fileNames file names
   * @param lineNumber line number
   */
  public void openFilesWith(final Settings.Editor editor, String[] fileNames, int lineNumber)
  {
    if (editor != null)
    {
      // get directory names
      String                directory    = null;
      LinkedHashSet<String> directorySet = new LinkedHashSet<String>();
      for (String fileName : fileNames)
      {
        File file = new File(fileName);
        if (file.isDirectory())
        {
          if (directory == null) directory = fileName;
          directorySet.add(fileName);
        }
        else
        {
          if (directory == null) directory = file.getParent();
          directorySet.add(file.getParent());
        }
      }

      // expand command
      Macro macro = new Macro(StringUtils.split(new ArrayList<String>(),editor.commandLine,StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS),Macro.PATTERN_PERCENTAGE);
/*      if (   !macro.contains("files")
          && !macro.contains("file")
          && !macro.contains("f")
          && !macro.contains("directories")
          && !macro.contains("directory")
          && !macro.contains("dir")
          && !macro.contains("d")
         )
      {
        macro.add("file");
      }*/
      macro.expand("files",      fileNames);
      macro.expand("file",       (fileNames.length > 0) ? fileNames[0] : "");
      macro.expand("f",          (fileNames.length > 0) ? fileNames[0] : "");
      macro.expand("directories",directorySet);
      macro.expand("directory",  (directory != null) ? directory : "");
      macro.expand("dir",        (directory != null) ? directory : "");
      macro.expand("d",          (directory != null) ? directory : "");
      macro.expand("lineNumber", lineNumber);
      macro.expand("n",          lineNumber);
      macro.expand("%%",        "%");
      Command command = new Command(macro);
//Dprintf.dprintf("command=%s",command);

      // run command
      Background.run(new BackgroundRunnable(command)
      {
        public void run(final Command command)
        {
          Process process = null;
          try
          {
            // start process (Note: use Runtime.exec() because it is a background process without i/o here)
            process = Runtime.getRuntime().exec(command.getCommandArray());

            // collect stderr output
            BufferedReader stderr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
            final ArrayList<String> stderrList = new ArrayList<String>();
            String line;
            while ((line = stderr.readLine()) != null)
            {
//Dprintf.dprintf("line=%s",line);
              stderrList.add(line);
            }

            // wait for exit
            final int exitCode = process.waitFor();
            if (exitCode > editor.validExitcode)
            {
              if (!display.isDisposed())
              {
                display.syncExec(new Runnable()
                {
                  public void run()
                  {
                    if (!shell.isDisposed())
                    {
                      Dialogs.error(shell,
                                    Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                                    stderrList,
                                    "Execute external command fail: \n\n'%s'\n\n (exitcode: %d)",
                                    command,
                                    exitCode
                                   );
                    }
                  }
                });
              }
              return;
            }

            process.destroy(); process = null;
          }
          catch (InterruptedException exception)
          {
            final String message = exception.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
                              Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                              "Execute external command fail: \n\n'%s'\n\n (error: %s)",
                              command,
                              message
                             );
              }
            });
            return;
          }
          catch (IOException exception)
          {
            final String message = exception.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
                              Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                              "Execute external command fail: \n\n'%s'\n\n (error: %s)",
                              command,
                              message
                             );
              }
            });
            return;
          }
          catch (Throwable throwable)
          {
            Onzen.printInternalError(throwable);
            return;
          }
          finally
          {
            if (process != null) process.destroy();
          }
        }
      });
    }
  }

  /** open file with external program
   * @param fileName file names
   * @param lineNumber line number
   */
  public void openFilesWith(String[] fileNames, int lineNumber)
  {
    String fileName = fileNames[0];
    Settings.Editor fileEditor = getFileEditor(fileName,
                                               Onzen.getMimeType(fileName)
                                              );
    openFilesWith(fileEditor,fileNames,lineNumber);
  }

  /** open files with external program
   * @param editor editor
   * @param fileName file names
   */
  public void openFilesWith(Settings.Editor editor, String[] fileNames)
  {
    openFilesWith(editor,fileNames,0);
  }

  /** open files with external program
   * @param editor editor
   * @param files files
   * @param lineNumber line number
   */
  public void openFilesWith(Settings.Editor editor, File[] files, int lineNumber)
  {
    String[] fileNames = new String[files.length];
    for (int i = 0; i < files.length; i++)
    {
      fileNames[i] = files[i].getPath();
    }
    openFilesWith(editor,fileNames,lineNumber);
  }

  /** open files with external program
   * @param file file
   * @param lineNumber line number
   */
  public void openFilesWith(File[] files, int lineNumber)
  {
    String fileName = files[0].getPath();
    Settings.Editor fileEditor = getFileEditor(fileName,
                                               Onzen.getMimeType(fileName)
                                              );
    openFilesWith(fileEditor,files,lineNumber);
  }

  /** open files with external program
   * @param editor editor
   * @param fileData file data
   */
  public void openFilesWith(Settings.Editor editor, FileData[] fileData)
  {
    String[] fileNames = new String[fileData.length];
    for (int i = 0; i < fileData.length; i++)
    {
      fileNames[i] = fileData[i].getName();
    }
    openFilesWith(editor,fileNames);
  }

  /** open files with external program
   * @param fileData file data
   */
  public void openFilesWith(FileData[] fileData)
  {
    Settings.Editor fileEditor = getFileEditor(fileData[0].getName(),
                                               fileData[0].getMimeType()
                                              );
    openFilesWith(fileEditor,fileData);
  }

  /** open files with external program
   * @param editor editor
   * @param file file
   */
  public void openFilesWith(Settings.Editor editor, File[] files)
  {
    openFilesWith(editor,files,0);
  }

  /** open files with external program
   * @param file file
   */
  public void openFilesWith(File[] files)
  {
    openFilesWith(files,0);
  }

  /** open files with external program
   * @param editor editor
   */
  public void openFilesWith(Settings.Editor editor)
  {
    FileData[] fileData = getSelectedFilesData();
    if (fileData != null)
    {
      openFilesWith(editor,fileData);
    }
  }

  /** open files with external program
   */
  public void openFilesWith()
  {
    FileData[] fileData = getSelectedFilesData();
    if (fileData != null)
    {
      openFilesWith(fileData);
    }
  }

  /** open file with external program
   * @param fileName file name
   * @param lineNumber line number
   */
  public void openFileWith(String fileName, int lineNumber)
  {
    Settings.Editor fileEditor = getFileEditor(fileName,
                                               Onzen.getMimeType(fileName)
                                              );
    openFilesWith(fileEditor,new String[]{fileName},lineNumber);
  }

  /** open files with external program
   * @param editor editor
   * @param fileName file name
   */
  public void openFileWith(Settings.Editor editor, String fileName)
  {
    openFilesWith(editor,new String[]{fileName},0);
  }

  /** open files with external program
   * @param editor editor
   * @param file file
   * @param lineNumber line number
   */
  public void openFileWith(Settings.Editor editor, File file, int lineNumber)
  {
    openFilesWith(editor,new String[]{file.getPath()},lineNumber);
  }

  /** open files with external program
   * @param file file
   * @param lineNumber line number
   */
  public void openFileWith(File file, int lineNumber)
  {
    openFileWith(file.getPath(),lineNumber);
  }

  /** open files with external program
   * @param editor editor
   * @param fileData file data
   */
  public void openFileWith(Settings.Editor editor, FileData fileData)
  {
    openFileWith(editor,fileData.getName());
  }

  /** open files with external program
   * @param fileData file data
   */
  public void openFileWith(FileData fileData)
  {
    Settings.Editor fileEditor = getFileEditor(fileData.getName(),
                                               fileData.getMimeType()
                                              );
    openFileWith(fileEditor,fileData.getName());
  }

  /** open files with external program
   * @param editor editor
   * @param file file
   */
  public void openFileWith(Settings.Editor editor, File file)
  {
    openFileWith(editor,file,0);
  }

  /** open files with external program
   * @param file file
   */
  public void openFileWith(File file)
  {
    openFileWith(file,0);
  }

  /** create new file
   * @param fileData file data
   */
  public void newFile(FileData fileData)
  {
    /** dialog data
     */
    class Data
    {
      String fileName;

      Data()
      {
        this.fileName = null;
      }
    };

    final Data  data = new Data();
    final Shell dialog;
    Composite   composite;
    Label       label;
    Button      button;

    if (fileData != null)
    {
      data.fileName = fileData.getName(repository);
    }

    // new directory dialog
    dialog = Dialogs.openModal(shell,"New file",300,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Text   widgetFileName;
    final Button widgetCreate;
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      label = Widgets.newLabel(composite,"File name:");
      Widgets.layout(label,0,0,TableLayoutData.W);

      widgetFileName = Widgets.newText(composite);
      if (data.fileName != null) widgetFileName.setText(data.fileName);
      widgetFileName.setSelection(data.fileName.length(),data.fileName.length());
      Widgets.layout(widgetFileName,1,1,TableLayoutData.WE);
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetCreate = Widgets.newButton(composite,"Create");
      widgetCreate.setToolTipText("Create new file.");
      Widgets.layout(widgetCreate,0,0,TableLayoutData.W,0,0,0,0,70,SWT.DEFAULT);
      widgetCreate.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          File file = new File(widgetFileName.getText());
          if      (file.exists())
          {
            Dialogs.error(shell,"File or directory '%s' already exists!",file.getPath());
            return;
          }
          else if ((file.getParentFile() != null) && !file.getParentFile().canWrite())
          {
            Dialogs.error(shell,"Permission denied for parent directory '%s'!",file.getParentFile().getPath());
            return;
          }
          else
          {
            data.fileName = widgetFileName.getText();
            Dialogs.close(dialog,true);
          }
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,3,TableLayoutData.E,0,0,0,0,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetFileName.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        widgetCreate.setFocus();
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });

    // show dialog
    Dialogs.show(dialog,Settings.setWindowLocation);

    // run
    widgetFileName.setFocus();
    if ((Boolean)Dialogs.run(dialog,false))
    {
      // create file
      File file = new File(data.fileName);
Dprintf.dprintf("");
/*
      if (!file.mkdirs())
      {
        Dialogs.error(shell,"Cannot create new file '%s'",data.fileName);
        return;
      }
*/

      // update file list
      asyncUpdateSelectedFileData();
    }
  }

  /** create new file
   */
  public void newFile()
  {
    newFile(null);
  }

  /** create new directory
   * @param fileData file data for pre-set directory name
   */
  public void newDirectory(FileData fileData)
  {
    /** dialog data
     */
    class Data
    {
      String path;

      Data()
      {
        this.path = null;
      }
    };

    final Data  data = new Data();
    final Shell dialog;
    Composite   composite,subComposite;
    Label       label;
    Button      button;

    if (fileData != null)
    {
      if (fileData.type == FileData.Types.DIRECTORY)
      {
        data.path = fileData.getName(repository);
      }
      else
      {
        data.path = fileData.getDirectoryName(repository);
      }
    }

    // new directory dialog
    dialog = Dialogs.openModal(shell,"New directory in: "+repository.rootPath,300,SWT.DEFAULT,new double[]{1.0,0.0},1.0);

    final Text   widgetPath;
    final List   widgetHistory;
    final Text   widgetMessage;
    final Button widgetCreate;
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(null,1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0,0.0}));
      Widgets.layout(subComposite,0,0,TableLayoutData.WE);
      {
        label = Widgets.newLabel(subComposite,"Sub-directory:");
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetPath = Widgets.newText(subComposite);
        if (!data.path.isEmpty())
        {
          widgetPath.setText(data.path+File.separator);
          widgetPath.setSelection(data.path.length()+File.separator.length(),data.path.length()+File.separator.length());
        }
        Widgets.layout(widgetPath,0,1,TableLayoutData.WE);

        button = Widgets.newButton(subComposite,Onzen.IMAGE_DIRECTORY);
        Widgets.layout(button,0,2,TableLayoutData.DEFAULT);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            String path = Dialogs.directory(shell,
                                           "Select path",
                                            widgetPath.getText()
                                           );
            if (path != null)
            {
              widgetPath.setText(path);
            }
          }
        });
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,2,0,TableLayoutData.WE,0,0,4);
    {
      widgetCreate = Widgets.newButton(composite,"Create");
      widgetCreate.setToolTipText("Create new directory.");
      Widgets.layout(widgetCreate,0,0,TableLayoutData.W,0,0,0,0,70,SWT.DEFAULT);
      widgetCreate.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          File file = new File(repository.rootPath,widgetPath.getText());
          if      (file.exists())
          {
            Dialogs.error(shell,"File or directory '%s' already exists!",file.getPath());
            return;
          }
          else if ((file.getParentFile() != null) && !file.getParentFile().canWrite())
          {
            Dialogs.error(shell,"Permission denied for parent directory '%s'!",file.getParentFile().getPath());
            return;
          }
          else
          {
            data.path = widgetPath.getText();
            Dialogs.close(dialog,true);
          }
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,3,TableLayoutData.E,0,0,0,0,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetPath.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        widgetCreate.setFocus();
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });

    // show dialog
    Dialogs.show(dialog,Settings.setWindowLocation);

    // run
    widgetPath.setFocus();
    if ((Boolean)Dialogs.run(dialog,false))
    {
      // create directory
      File file = new File(repository.rootPath,data.path);
      if (!file.mkdirs())
      {
        Dialogs.error(shell,"Cannot create new directory '%s'",data.path);
        return;
      }

      // add directory tree item
      addTreeItem(repository.getFileDataInstance(data.path,FileData.Types.DIRECTORY));
    }
  }

  /** create new directory
   */
  public void newDirectory()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      newDirectory(fileData);
      asyncUpdateSelectedFileData();
    }
  }

  /** rename local file/directory
   * @param fileData file to rename
   */
  public void renameLocalFile(FileData fileData)
  {
    /** dialog data
     */
    class Data
    {
      String newName;

      Data()
      {
        this.newName = null;
      }
    };

    final Data  data = new Data();
    final Shell dialog;
    Composite   composite;
    Label       label;
    Text        text;
    Button      button;

    final String oldName = fileData.getName(repository);
    data.newName = oldName;

    // rename file/directory dialog
    dialog = Dialogs.openModal(shell,"Rename/move file/directory",new double[]{1.0,0.0},1.0);

    final Text   widgetNewName;
    final Button widgetRename;
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0},4));
    Widgets.layout(composite,0,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"Old name:");
      Widgets.layout(label,0,0,TableLayoutData.W);

      text = Widgets.newStringView(composite);
      text.setText(oldName);
      Widgets.layout(text,0,1,TableLayoutData.WE);

      label = Widgets.newLabel(composite,"New name:");
      Widgets.layout(label,1,0,TableLayoutData.W);

      widgetNewName = Widgets.newText(composite);
      widgetNewName.setText(data.newName);
      Widgets.layout(widgetNewName,1,1,TableLayoutData.WE);
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetRename = Widgets.newButton(composite,"Rename/Move");
      widgetRename.setToolTipText("Rename/move file.");
      widgetRename.setEnabled(false);
      Widgets.layout(widgetRename,0,0,TableLayoutData.W,0,0,0,0,120,SWT.DEFAULT);
      widgetRename.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.newName                     = widgetNewName.getText();
          Settings.geometryRenameLocalFile = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,80,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetNewName.addModifyListener(new ModifyListener()
    {
      public void modifyText(ModifyEvent modifyEvent)
      {
        Text widget = (Text)modifyEvent.widget;

        widgetRename.setEnabled(!widget.getText().equals(oldName));
      }
    });
    widgetNewName.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        widgetRename.setFocus();
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });

    // show dialog
    Dialogs.show(dialog,Settings.geometryRenameLocalFile.x,SWT.DEFAULT,Settings.setWindowLocation);

    // run
    Widgets.setFocus(widgetNewName);
    if ((Boolean)Dialogs.run(dialog,false))
    {
      // rename local file/directory
      File oldFile = new File(repository.rootPath,oldName);
      File newFile = new File(repository.rootPath,data.newName);

      if (!oldFile.renameTo(newFile))
      {
        Dialogs.error(shell,"Cannot rename file/directory\n\n'%s'\n\nto\n\n'%s'",oldFile.getName(),newFile.getName());
        return;
      }

      // update tree
      fileData.setFileName(data.newName);
      updateTreeItem(fileData);
    }
  }

  /** rename local file/directory
   */
  public void renameLocalFile()
  {
    FileData fileData = getSelectedFileData();
    if (fileData != null)
    {
      renameLocalFile(fileData);
    }
  }

  /** delete local files/directories
   * @param fileNameSet files to delete
   */
  public void deleteLocalFiles(Set<FileData> fileDataSet)
  {
    if (!fileDataSet.isEmpty())
    {
      final Shell dialog;
      Composite   composite;
      Button      button;

      // delete dialog
      dialog = Dialogs.openModal(shell,"Delete local files/directories",new double[]{1.0,0.0},1.0);

      final List   widgetFiles;
      final Button widgetDelete;
      final Button widgetCancel;
      composite = Widgets.newComposite(dialog);
      composite.setLayout(new TableLayout(1.0,1.0,4));
      Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
      {
        widgetFiles = Widgets.newList(composite);
        widgetFiles.setToolTipText("Local files/directories to delete.");
        widgetFiles.setBackground(Onzen.COLOR_GRAY);
        Widgets.layout(widgetFiles,0,0,TableLayoutData.NSWE);
      }

      // buttons
      composite = Widgets.newComposite(dialog);
      composite.setLayout(new TableLayout(0.0,1.0));
      Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
      {
        widgetDelete = Widgets.newButton(composite,"Delete");
        widgetDelete.setToolTipText("Delete local files/directories.");
        Widgets.layout(widgetDelete,0,0,TableLayoutData.W,0,0,0,0,70,SWT.DEFAULT);
        widgetDelete.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Settings.geometryDeleteLocalFiles = dialog.getSize();

            Dialogs.close(dialog,true);
          }
        });

        widgetCancel = Widgets.newButton(composite,"Cancel");
        Widgets.layout(widgetCancel,0,3,TableLayoutData.E,0,0,0,0,70,SWT.DEFAULT);
        widgetCancel.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Dialogs.close(dialog,false);
          }
        });
      }

      // listeners

      // show dialog
      Dialogs.show(dialog,Settings.geometryDeleteLocalFiles,Settings.setWindowLocation);

      // add files
      for (FileData fileData : fileDataSet)
      {
        widgetFiles.add(fileData.getName(repository));
      }

      // run
      widgetCancel.setFocus();
      if ((Boolean)Dialogs.run(dialog,false))
      {
        HashSet<FileData> updateFileDataSet = new HashSet<FileData>();

        boolean deleteAll  = false;
        boolean skipErrors = false;
        for (FileData fileData : fileDataSet)
        {
          File file = fileData.getFile();

          // delete local file/directory
          try
          {
            if (file.isDirectory())
            {
              // confirm deleting directory
              if (!deleteAll)
              {
                switch (Dialogs.select(shell,
                                       "Error",
                                       String.format("'%s' is a directory.\n\nDelete directory?",file.getPath()),
                                       new String[]{"Yes","Yes, always","No"},
                                       2
                                      )
                       )
                {
                  case 0:
                    break;
                  case 1:
                    deleteAll = true;
                    break;
                  case 2:
                    continue;
                }
              }

              // delete local directory tree
              FileUtils.deleteDirectoryTree(file);
            }
            else
            {
              // delete local file
              FileUtils.deleteFile(file);
            }

            // store entry and parent for refresh
            updateFileDataSet.add(fileData);
            FileData parentFileData = fileData.getParent();
            if (parentFileData != null) updateFileDataSet.add(parentFileData);
          }
          catch (IOException exception)
          {
            if (!skipErrors)
            {
              switch (Dialogs.select(shell,
                                     "Error",
                                     String.format("Cannot delete local file/directory '%s'.\n\nContinue?",file.getPath()),
                                     new String[]{"Yes","Yes, always","No"},
                                     2
                                    )
                     )
              {
                case 0:
                  break;
                case 1:
                  skipErrors = true;
                  break;
                case 2:
                  return;
              }
            }
          }
        }

        // start update file data
        asyncUpdateFileStates(updateFileDataSet);
      }
    }
  }

  /** delete local file/directory
   */
  public void deleteLocalFile(FileData fileData)
  {
    deleteLocalFiles(fileData.toSet());
  }

  /** delete selected local files/directories
   */
  public void deleteLocalFiles()
  {
    deleteLocalFiles(getSelectedFileDataSet());
  }

  /** add to skip convert whitespaces file list
   * @param fileNameSet files to add to skip convert whitespaces list
   */
  public void addSkipConvertWhitespacesFiles(Set<FileData> fileDataSet)
  {
    // add files
    for (FileData fileData : fileDataSet)
    {
      repository.addSkipConvertWhitespacesFilePattern(fileData.getName(repository));
    }

    // refresh
    asyncUpdateFileStates(fileDataSet);
  }

  /** add selected files to skip convert whitespaces file list
   * @param fileNameSet files to add to skip convert whitespaces list
   */
  public void addSkipConvertWhitespacesFiles()
  {
    addSkipConvertWhitespacesFiles(getSelectedFileDataSet());
    asyncUpdateSelectedFileData();
  }

  /** remove from skip convert whitespaces file list
   * @param fileNameSet files to remove from skip convert whitespaces list
   */
  public void removeSkipConvertWhitespacesFiles(Set<FileData> fileDataSet)
  {
    // add files
    for (FileData fileData : fileDataSet)
    {
      repository.removeSkipConvertWhitespacesFilePattern(fileData.getName(repository));
    }

    // refresh
    asyncUpdateFileStates(fileDataSet);
  }

  /** remove selected files from skip convert whitespaces file list
   * @param fileNameSet files to remove from skip convert whitespaces list
   */
  public void removeSkipConvertWhitespacesFiles()
  {
    addSkipConvertWhitespacesFiles(getSelectedFileDataSet());
    asyncUpdateSelectedFileData();
  }

  /** add to ignore file list
   * @param fileNameSet files to add to ignore list
   */
  public void addIgnoreFiles(Set<FileData> fileDataSet)
  {
    // add files
    for (FileData fileData : fileDataSet)
    {
      repository.addIgnoreFilePattern(fileData.getName(repository));
    }

    // refresh
    asyncUpdateFileStates(fileDataSet);
  }

  /** add selected files to ignore file list
   * @param fileNameSet files to add to ignore list
   */
  public void addIgnoreFiles()
  {
    addIgnoreFiles(getSelectedFileDataSet());
    asyncUpdateSelectedFileData();
  }

  /** remove from ignore file list
   * @param fileNameSet files to remove from ignore list
   */
  public void removeIgnoreFiles(Set<FileData> fileDataSet)
  {
    // add files
    for (FileData fileData : fileDataSet)
    {
      repository.removeIgnoreFilePattern(fileData.getName(repository));
    }

    // refresh
    asyncUpdateFileStates(fileDataSet);
  }

  /** remove selected files from ignore file list
   * @param fileNameSet files to remove from ignore list
   */
  public void removeIgnoreFiles()
  {
    addIgnoreFiles(getSelectedFileDataSet());
    asyncUpdateSelectedFileData();
  }

  /** add to hidden list
   * @param fileNameSet files to add to ignore list
   */
  public void addHiddenFiles(Set<FileData> fileDataSet)
  {
    // get pattern maps
    LinkedHashMap<String,Settings.FilePattern> hiddenFilePatternsMap = new LinkedHashMap<String,Settings.FilePattern>();
    for (Settings.FilePattern hiddenFilePattern : Settings.hiddenFilePatterns)
    {
      hiddenFilePatternsMap.put(hiddenFilePattern.string,hiddenFilePattern);
    }
    LinkedHashMap<String,Settings.FilePattern> hiddenDirectoryPatternsMap = new LinkedHashMap<String,Settings.FilePattern>();
    for (Settings.FilePattern hiddenDirectoryPattern : Settings.hiddenDirectoryPatterns)
    {
      hiddenDirectoryPatternsMap.put(hiddenDirectoryPattern.string,hiddenDirectoryPattern);
    }

    // add patterns
    for (FileData fileData : fileDataSet)
    {
      if (fileData.type == FileData.Types.DIRECTORY)
      {
        hiddenDirectoryPatternsMap.put(fileData.getName(),new Settings.FilePattern(fileData.getName()));
      }
      else
      {
        hiddenFilePatternsMap.put(fileData.getName(),new Settings.FilePattern(fileData.getName()));
      }
    }

    // store new pattern maps
    Settings.hiddenFilePatterns      = hiddenFilePatternsMap.values().toArray(new Settings.FilePattern[hiddenFilePatternsMap.size()]);
    Settings.hiddenDirectoryPatterns = hiddenDirectoryPatternsMap.values().toArray(new Settings.FilePattern[hiddenDirectoryPatternsMap.size()]);

    // refresh
//TODO refresh parent directories
    asyncUpdateFileStates(fileDataSet);
  }

  /** add selected files to hidden list
   * @param fileNameSet files to add to hidden list
   */
  public void addHiddenFiles()
  {
    addHiddenFiles(getSelectedFileDataSet());
    asyncUpdateSelectedFileData();
  }

  /** find files by name
   */
  public void findFilesByName()
  {
    CommandFindFiles commandFindFiles = new CommandFindFiles(shell,this,CommandFindFiles.Types.NAME);
    commandFindFiles.run();
  }

  /** find files by content
   */
  public void findFilesByContent()
  {
    CommandFindFiles commandFindFiles = new CommandFindFiles(shell,this,CommandFindFiles.Types.CONTENT);
    commandFindFiles.run();
  }

  /** copy file names to clipboard
   */
  public void copyFileNamesToClipboard()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      StringBuilder buffer = new StringBuilder();
      for (FileData fileData : fileDataSet)
      {
        buffer.append(fileData.getName()+"\n");
      }
      Widgets.setClipboard(clipboard,buffer.toString());
    }
  }

  /** convert whitespaces dialog in selected files
   */
  public void convertWhitespaces()
  {
    Set<FileData> fileDataSet = getSelectedFileDataSet();
    if (fileDataSet != null)
    {
      /** dialog data
       */
      class Data
      {
        EnumSet<ConvertWhitespacesModes> convertWhitespacesModes;
        int                              spacesPerTAB;

        Data()
        {
          this.convertWhitespacesModes = EnumSet.noneOf(ConvertWhitespacesModes.class);
          this.spacesPerTAB            = Settings.convertSpacesPerTAB;
        }
      };

      final Data data = new Data();

      Composite composite,subComposite;
      Control   control;
      Label     label;
      Button    button;

      // convert dialog
      final Shell dialog = Dialogs.openModal(shell,"Convert whitespaces",new double[]{1.0,0.0},1.0);

      final List    widgetFiles;
      final Spinner widgetSpacesPerTAB;
      final Button  widgetConvertTABs;
      final Button  widgetRemoveTrailingWhitespaces;
      final Button  widgetConvert;
      final Button  widgetCancel;
      composite = Widgets.newComposite(dialog);
      composite.setLayout(new TableLayout(new double[]{0.0,1.0,0.0,0.0},1.0,4));
      Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
      {
        label = Widgets.newLabel(composite,"The following files contain TABs/trailing whitespaces:");
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetFiles = Widgets.newList(composite,SWT.BORDER);
        widgetFiles.setToolTipText("Files to convert TABs/whitespaces.");
        widgetFiles.setBackground(Onzen.COLOR_GRAY);
        Widgets.layout(widgetFiles,1,0,TableLayoutData.NSWE);

        subComposite = Widgets.newComposite(composite);
        subComposite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0}));
        Widgets.layout(subComposite,2,0,TableLayoutData.W);
        {
          widgetConvertTABs = Widgets.newCheckbox(subComposite,"Convert TABs to spaces:");
          widgetConvertTABs.setSelection(data.spacesPerTAB > 0);
          Widgets.layout(widgetConvertTABs,0,0,TableLayoutData.W);

          widgetSpacesPerTAB = Widgets.newSpinner(subComposite);
          widgetSpacesPerTAB.setMinimum(1);
          widgetSpacesPerTAB.setMaximum(255);
          widgetSpacesPerTAB.setEnabled(data.spacesPerTAB > 0);
          widgetSpacesPerTAB.setSelection(data.spacesPerTAB);
          Widgets.layout(widgetSpacesPerTAB,0,1,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);
        }

        widgetRemoveTrailingWhitespaces = Widgets.newCheckbox(composite,"Remove trailing whitespaces");
        widgetRemoveTrailingWhitespaces.setSelection(data.convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING));
        Widgets.layout(widgetRemoveTrailingWhitespaces,3,0,TableLayoutData.W);
      }

      // buttons
      composite = Widgets.newComposite(dialog);
      composite.setLayout(new TableLayout(0.0,1.0));
      Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
      {
        widgetConvert = Widgets.newButton(composite,"Convert");
        widgetConvert.setToolTipText("Convert whitespaces in files.");
        Widgets.layout(widgetConvert,0,0,TableLayoutData.W,0,0,0,0,70,SWT.DEFAULT);
        widgetConvert.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            data.convertWhitespacesModes.clear();
            if (widgetConvertTABs.getSelection()              ) data.convertWhitespacesModes.add(ConvertWhitespacesModes.TABS    );
            if (widgetRemoveTrailingWhitespaces.getSelection()) data.convertWhitespacesModes.add(ConvertWhitespacesModes.TRAILING);
            data.spacesPerTAB = widgetSpacesPerTAB.getSelection();

            convertWhiteDialogLocation          = dialog.getLocation();
            Settings.geometryConvertWhitespaces = dialog.getSize();

            Dialogs.close(dialog,true);
          }
        });

        widgetCancel = Widgets.newButton(composite,"Cancel");
        Widgets.layout(widgetCancel,0,3,TableLayoutData.E,0,0,0,0,70,SWT.DEFAULT);
        widgetCancel.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Dialogs.close(dialog,false);
          }
        });
      }

      // listeners

      // show dialog
      Dialogs.show(dialog,convertWhiteDialogLocation,Settings.geometryConvertWhitespaces,Settings.setWindowLocation);

      // add files
      LinkedList<String> fileNameList = new LinkedList<String>();
      for (FileData fileData : fileDataSet)
      {
        String name = fileData.getName(repository);
        if (!repository.isSkipConvertWhitespacesFile(name))
        {
          widgetFiles.add(name);
          fileNameList.add(name);
        }
      }

      // run
      widgetCancel.setFocus();
      if ((Boolean)Dialogs.run(dialog,false))
      {
        // convert TABs/whitespaces in files
        boolean convertAll = false;
        while (!fileNameList.isEmpty())
        {
          String fileName = fileNameList.removeFirst();
          File   file     = new File(fileName);

          // convert  TABs/whitespaces in file
          boolean converted  = false;
          boolean skipErrors = false;
          if (file.isDirectory())
          {
            // confirm convert all files in directory
            if (!convertAll)
            {
              switch (Dialogs.select(shell,
                                     "Error",
                                     String.format("'%s' is a directory.\n\nConvert all files in directory?",fileName),
                                     new String[]{"Yes","Yes, always","No"},
                                     2
                                    )
                     )
              {
                case 0:
                  break;
                case 1:
                  convertAll = true;
                  break;
                case 2:
                  continue;
              }
            }

            // add files in directory for checking for whitespaces
            for (String directoryFileName : file.list())
            {
              fileNameList.add(directoryFileName);
            }
          }
          else
          {
            // convert whitespaces in file
            try
            {
              convertWhitespaces(fileName,
                                 data.convertWhitespacesModes,
                                 data.spacesPerTAB
                                );
            }
            catch (IOException exception)
            {
              if (!skipErrors)
              {
                switch (Dialogs.select(shell,
                                       "Error",
                                       String.format("Cannot convert file\n\n'%s'\n\n(error: %s).\n\nContinue?",file.getPath(),exception.getMessage()),
                                       new String[]{"Yes","Yes, always","No"},
                                       2
                                      )
                       )
                {
                  case 0:
                    break;
                  case 1:
                    skipErrors = true;
                    break;
                  case 2:
                    return;
                }
              }
            }
          }
        }

        // start update file data
        asyncUpdateFileStates(fileDataSet);
      }
    }
  }

  /** convert whitespaces dialog
   * @param fileName file name
   * @param fileNameList all file names
   * @param convertWhitespacesModes convert whitespaces modes
   * @return true iff OK, false for abort
   */
  public ConvertWhitespacesResult convertWhitespaces(String                           fileName,
                                                     java.util.List<String>           fileNameList,
                                                     String                           message,
                                                     EnumSet<ConvertWhitespacesModes> convertWhitespacesModes
                                                    )
  {
    /** dialog data
     */
    class Data
    {
      EnumSet<ConvertWhitespacesModes> convertWhitespacesModes;
      int                              spacesPerTAB;
      boolean                          convertAll;

      Data()
      {
        this.convertWhitespacesModes = EnumSet.noneOf(ConvertWhitespacesModes.class);
        this.spacesPerTAB            = Settings.convertSpacesPerTAB;
        this.convertAll              = false;
      }
    };

    final Data data = new Data();

    Composite composite,subComposite;
    Control   control;
    Label     label;
    Button    button;

    // convert dialog
    final Shell dialog = Dialogs.openModal(shell,"Convert whitespaces",new double[]{1.0,0.0},1.0);

    final List    widgetFiles;
    final Spinner widgetSpacesPerTAB;
    final Button  widgetConvertTABs;
    final Button  widgetRemoveTrailingWhitespaces;
    final Button  widgetConvert;
    final Button  widgetKeep;
    final Button  widgetConvertAll;
    final Button  widgetKeepAll;
    final Button  widgetAbort;
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{0.0,1.0,0.0,0.0},1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      label = Widgets.newLabel(composite,message);
      Widgets.layout(label,0,0,TableLayoutData.W);

      widgetFiles = Widgets.newList(composite,SWT.BORDER);
      widgetFiles.setToolTipText("Files to convert TABs/whitespaces.");
      widgetFiles.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetFiles,1,0,TableLayoutData.NSWE);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(1.0,0.0));
      Widgets.layout(subComposite,2,0,TableLayoutData.W);
      {
        widgetConvertTABs = Widgets.newCheckbox(subComposite,"Convert TABs to spaces:");
        if (convertWhitespacesModes.contains(ConvertWhitespacesModes.TABS))
        {
          widgetConvertTABs.setSelection(data.spacesPerTAB > 0);
        }
        widgetConvertTABs.setEnabled(convertWhitespacesModes.contains(ConvertWhitespacesModes.TABS));
        Widgets.layout(widgetConvertTABs,0,0,TableLayoutData.W);

        widgetSpacesPerTAB = Widgets.newSpinner(subComposite);
        widgetSpacesPerTAB.setMinimum(1);
        widgetSpacesPerTAB.setMaximum(255);
        widgetSpacesPerTAB.setSelection(data.spacesPerTAB);
        widgetSpacesPerTAB.setEnabled(convertWhitespacesModes.contains(ConvertWhitespacesModes.TABS) && (data.spacesPerTAB > 0));
        Widgets.layout(widgetSpacesPerTAB,0,1,TableLayoutData.W,0,0,0,0,80,SWT.DEFAULT);
      }

      widgetRemoveTrailingWhitespaces = Widgets.newCheckbox(composite,"Remove trailing whitespaces");
      if (convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING))
      {
        widgetRemoveTrailingWhitespaces.setSelection(convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING));
      }
      widgetRemoveTrailingWhitespaces.setEnabled(convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING));
      Widgets.layout(widgetRemoveTrailingWhitespaces,3,0,TableLayoutData.W);
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,0.0,0.0,1.0}));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetConvert = Widgets.newButton(composite,"Convert");
      widgetConvert.setToolTipText("Convert whitespaces in file.");
      Widgets.layout(widgetConvert,0,0,TableLayoutData.W,0,0,0,0,140,SWT.DEFAULT);
      widgetConvert.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.convertWhitespacesModes.clear();
          if (widgetConvertTABs.getSelection()              ) data.convertWhitespacesModes.add(ConvertWhitespacesModes.TABS    );
          if (widgetRemoveTrailingWhitespaces.getSelection()) data.convertWhitespacesModes.add(ConvertWhitespacesModes.TRAILING);
          data.spacesPerTAB = widgetSpacesPerTAB.getSelection();
          data.convertAll   = false;

          convertWhiteDialogLocation          = dialog.getLocation();
          Settings.geometryConvertWhitespaces = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      widgetKeep = Widgets.newButton(composite,"Keep");
      widgetKeep.setToolTipText("Keep whitespaces in file.");
      Widgets.layout(widgetKeep,0,1,TableLayoutData.W,0,0,0,0,140,SWT.DEFAULT);
      widgetKeep.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.convertWhitespacesModes.clear();
          data.convertAll = false;

          convertWhiteDialogLocation          = dialog.getLocation();
          Settings.geometryConvertWhitespaces = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      widgetConvertAll = Widgets.newButton(composite,"Convert in all files");
      widgetConvertAll.setToolTipText("Convert whitespaces in all files.");
      Widgets.layout(widgetConvertAll,0,2,TableLayoutData.W,0,0,0,0,140,SWT.DEFAULT);
      widgetConvertAll.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.convertWhitespacesModes.clear();
          if (widgetConvertTABs.getSelection()              ) data.convertWhitespacesModes.add(ConvertWhitespacesModes.TABS    );
          if (widgetRemoveTrailingWhitespaces.getSelection()) data.convertWhitespacesModes.add(ConvertWhitespacesModes.TRAILING);
          data.spacesPerTAB = widgetSpacesPerTAB.getSelection();
          data.convertAll   = true;

          convertWhiteDialogLocation          = dialog.getLocation();
          Settings.geometryConvertWhitespaces = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      widgetKeepAll = Widgets.newButton(composite,"Keep in all files");
      widgetKeepAll.setToolTipText("Keep whitespaces in all files.");
      Widgets.layout(widgetKeepAll,0,3,TableLayoutData.W,0,0,0,0,140,SWT.DEFAULT);
      widgetKeepAll.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.convertWhitespacesModes.clear();
          data.convertAll = true;

          convertWhiteDialogLocation          = dialog.getLocation();
          Settings.geometryConvertWhitespaces = dialog.getSize();

          Dialogs.close(dialog,true);
        }
      });

      widgetAbort = Widgets.newButton(composite,"Abort");
      Widgets.layout(widgetAbort,0,4,TableLayoutData.E,0,0,0,0,70,SWT.DEFAULT);
      widgetAbort.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners

    // show dialog
    Dialogs.show(dialog,convertWhiteDialogLocation,Settings.geometryConvertWhitespaces,Settings.setWindowLocation);

    // add file
    widgetFiles.add(fileName);

    // run
    widgetAbort.setFocus();
    if ((Boolean)Dialogs.run(dialog,false))
    {
      // convert whitespaces in files
      if (data.convertAll)
      {
        try
        {
          convertWhitespaces(fileNameList,
                             data.convertWhitespacesModes,
                             data.spacesPerTAB
                            );
        }
        catch (IOException exception)
        {
          Dialogs.error(shell,
                        String.format("Cannot convert whitespaces in files\n(error: %s).",exception.getMessage())
                       );
          return ConvertWhitespacesResult.FAIL;
        }

        if (data.convertWhitespacesModes.contains(ConvertWhitespacesModes.TABS) && data.convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING))
        {
          return ConvertWhitespacesResult.ALL;
        }
        else if (data.convertWhitespacesModes.contains(ConvertWhitespacesModes.TABS))
        {
          return ConvertWhitespacesResult.ALL;
        }
        else if (data.convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING))
        {
          return ConvertWhitespacesResult.ALL;
        }
        else
        {
          return ConvertWhitespacesResult.OK;
        }
      }
      else
      {
        try
        {
          convertWhitespaces(fileName,
                             data.convertWhitespacesModes,
                             data.spacesPerTAB
                            );
        }
        catch (IOException exception)
        {
          if (!Dialogs.confirm(shell,
                               "Error",
                               String.format("Cannot convert whitespaces in file\n\n'%s'\n\n(error: %s).\n\nContinue?",fileName,exception.getMessage())
                              )
             )
          {
            return ConvertWhitespacesResult.FAIL;
          }
        }
        return ConvertWhitespacesResult.OK;
      }
    }
    else
    {
      return ConvertWhitespacesResult.ABORTED;
    }
  }

  /** trim trailing whitespaces in file
   * @param fileName file name
   */
  public void trimTrailingWhitespaces(String fileName)
    throws IOException
  {
    ArrayList<String> fileNameList = new ArrayList<String>();
    fileNameList.add(fileName);
    convertWhitespaces(fileNameList,
                       EnumSet.of(ConvertWhitespacesModes.TRAILING),
                       0  // spacesPerTAB
                      );
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "RepositoryTab {"+repository.title+"}";
  }

  /** execute shell command
   * @param fileDataSet files for shell command
   */
  public void executeShellCommand(final Settings.ShellCommand shellCommand, Set<FileData> fileDataSet)
  {
    if (fileDataSet != null)
    {
      FileData[] fileDataArray = fileDataSet.toArray(new FileData[fileDataSet.size()]);

      // expand command
      Macro macro = new Macro(StringUtils.split(new ArrayList<String>(),shellCommand.commandLine,StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS),Macro.PATTERN_PERCENTAGE);
/*      if (   !macro.contains("files")
          && !macro.contains("file1") && !macro.contains("directory1")
          && !macro.contains("file2") && !macro.contains("directory2")
          && !macro.contains("file3") && !macro.contains("directory3")
          && !macro.contains("file4") && !macro.contains("directory4")
          && !macro.contains("file5") && !macro.contains("directory5")
          && !macro.contains("file6") && !macro.contains("directory6")
          && !macro.contains("file7") && !macro.contains("directory7")
          && !macro.contains("file8") && !macro.contains("directory8")
          && !macro.contains("file9") && !macro.contains("directory9")
          && !macro.contains("file") && !macro.contains("directory")
         )
      {
        macro.add("file");
      } */
      macro.expand("files",     fileDataSet,'"'," ");
      macro.expand("file1",     (fileDataArray.length >= 1) ? fileDataArray[0].getName()      : "");
      macro.expand("directory1",(fileDataArray.length >= 1) ? fileDataArray[0].getDirectoryName() : "");
      macro.expand("file2",     (fileDataArray.length >= 2) ? fileDataArray[1].getName()      : "");
      macro.expand("directory2",(fileDataArray.length >= 2) ? fileDataArray[1].getDirectoryName() : "");
      macro.expand("file3",     (fileDataArray.length >= 3) ? fileDataArray[2].getName()      : "");
      macro.expand("directory3",(fileDataArray.length >= 3) ? fileDataArray[2].getDirectoryName() : "");
      macro.expand("file4",     (fileDataArray.length >= 4) ? fileDataArray[3].getName()      : "");
      macro.expand("directory4",(fileDataArray.length >= 4) ? fileDataArray[3].getDirectoryName() : "");
      macro.expand("file5",     (fileDataArray.length >= 5) ? fileDataArray[4].getName()      : "");
      macro.expand("directory5",(fileDataArray.length >= 5) ? fileDataArray[4].getDirectoryName() : "");
      macro.expand("file6",     (fileDataArray.length >= 6) ? fileDataArray[5].getName()      : "");
      macro.expand("directory6",(fileDataArray.length >= 6) ? fileDataArray[5].getDirectoryName() : "");
      macro.expand("file7",     (fileDataArray.length >= 7) ? fileDataArray[6].getName()      : "");
      macro.expand("directory7",(fileDataArray.length >= 7) ? fileDataArray[6].getDirectoryName() : "");
      macro.expand("file8",     (fileDataArray.length >= 8) ? fileDataArray[7].getName()      : "");
      macro.expand("directory8",(fileDataArray.length >= 8) ? fileDataArray[7].getDirectoryName() : "");
      macro.expand("file9",     (fileDataArray.length >= 9) ? fileDataArray[8].getName()      : "");
      macro.expand("directory9",(fileDataArray.length >= 9) ? fileDataArray[8].getDirectoryName() : "");
      macro.expand("file",      (fileDataArray.length >= 1) ? fileDataArray[0].getName()      : "");
      macro.expand("f",         (fileDataArray.length >= 1) ? fileDataArray[0].getName()      : "");
      macro.expand("directory", (fileDataArray.length >= 1) ? fileDataArray[0].getDirectoryName() : "");
      macro.expand("d",         (fileDataArray.length >= 1) ? fileDataArray[0].getDirectoryName() : "");
      macro.expand("rootPath",  repository.rootPath);
      macro.expand("r",         repository.rootPath);
      macro.expand("",          "%");
      Command command = new Command(macro);

      // run command
      Background.run(new BackgroundRunnable(command)
      {
        public void run(final Command command)
        {
          Process process = null;
          try
          {
            // start process (Note: use Runtime.exec() because it is a background process without i/o here)
            process = Runtime.getRuntime().exec(command.getCommandArray());

            // collect stderr output
            BufferedReader stderr = new BufferedReader(new InputStreamReader(process.getErrorStream()));
            final ArrayList<String> stderrList = new ArrayList<String>();
            String line;
            while ((line = stderr.readLine()) != null)
            {
              stderrList.add(line);
            }

            // wait for exit
            final int exitCode = process.waitFor();
            if (exitCode > shellCommand.validExitcode)
            {
              if (!display.isDisposed())
              {
                display.syncExec(new Runnable()
                {
                  public void run()
                  {
                   if (!shell.isDisposed())
                    {
                      Dialogs.error(shell,
                                    Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                                    stderrList,
                                    "Execute external shell command fail: \n\n'%s'\n\n (exitcode: %d)",
                                    command,
                                    exitCode
                                   );
                    }
                  }
                });
              }
              return;
            }

            process.destroy(); process = null;
          }
          catch (InterruptedException exception)
          {
            final String message = exception.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
                              Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                              "Execute external shell command fail: \n\n'%s'\n\n (error: %s)",
                              command,
                              message
                             );
              }
            });
            return;
          }
          catch (IOException exception)
          {
            final String message = exception.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
                              Dialogs.booleanFieldUpdater(Settings.class,"showExecuteErrors"),
                              "Execute external shell command fail: \n\n'%s'\n\n (error: %s)",
                              command,
                              message
                             );
              }
            });
            return;
          }
          finally
          {
            if (process != null) process.destroy();
          }
        }
      });
    }
  }

  /** run shell command
   */
  public void executeShellCommand(Settings.ShellCommand shellCommand)
  {
    executeShellCommand(shellCommand,getSelectedFileDataSet());
    asyncUpdateSelectedFileData();
  }

  //-----------------------------------------------------------------------

  /** add root directory
   */
  private void addRootDirectory()
  {
    FileData rootFileData = repository.getFileDataInstance("",
                                                           FileData.Types.DIRECTORY
                                                          );

    // create root tree item
    TreeItem rootTreeItem = Widgets.addTreeItem(widgetFileTree,rootFileData,true);
    rootTreeItem.setText(repository.rootPath);
    rootTreeItem.setImage(Onzen.IMAGE_DIRECTORY);

    // add tree item reference
    rootFileData.addTreeItem(rootTreeItem);

    // add listeners
    widgetFileTree.addTreeListener(new TreeListener()
    {
      public void treeCollapsed(TreeEvent treeEvent)
      {
        TreeItem treeItem = (TreeItem)treeEvent.item;
        closeFileTreeItem(treeItem);
      }
      public void treeExpanded(TreeEvent treeEvent)
      {
        TreeItem treeItem = (TreeItem)treeEvent.item;
        FileData fileData = (FileData)treeItem.getData();
        if (fileData.type == FileData.Types.DIRECTORY)
        {
          openFileTreeItem(treeItem);
        }
      }
    });
    widgetFileTree.addMouseListener(new MouseListener()
    {
      public void mouseDoubleClick(MouseEvent mouseEvent)
      {
        TreeItem treeItem = widgetFileTree.getItem(new Point(mouseEvent.x,mouseEvent.y));
        if (treeItem != null)
        {
          FileData fileData = (FileData)treeItem.getData();
          switch (fileData.type)
          {
            case DIRECTORY:
              /* On Linux a double-click does not open directory by default. Send a
                 event to initiate this behavior on Linux.
              */
              if (Settings.hostSystem == Settings.HostSystems.LINUX)
              {
                Event treeEvent = new Event();
                treeEvent.item = treeItem;
                if (treeItem.getExpanded())
                {
                  widgetFileTree.notifyListeners(SWT.Collapse,treeEvent);
                }
                else
                {
                  widgetFileTree.notifyListeners(SWT.Expand,treeEvent);
                }
              }
              break;
            case FILE:
            case LINK:
              openFile(fileData);
              break;
            default:
              break;
          }
        }
      }

      public void mouseDown(final MouseEvent mouseEvent)
      {
      }

      public void mouseUp(final MouseEvent mouseEvent)
      {
      }
    });
    /* Note: do not use SelectionListener, because after a mouse-click
       the selection listener is executed, too! (this cause open and
       closing the sub-tree immediately)
    */
    widgetFileTree.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
//Dprintf.dprintf("keyEvent=%s",keyEvent);
        if ((keyEvent.stateMask == 0) && (keyEvent.keyCode == SWT.CR))
        {
// NYI: handling more than one item?
          TreeItem[] treeItems = widgetFileTree.getSelection();
          if ((treeItems != null) && (treeItems.length > 0) && (treeItems[0] != null))
          {
            FileData fileData = (FileData)treeItems[0].getData();
            switch (fileData.type)
            {
              case DIRECTORY:
                Event treeEvent = new Event();
                treeEvent.item = treeItems[0];
                if (treeItems[0].getExpanded())
                {
                  widgetFileTree.notifyListeners(SWT.Collapse,treeEvent);
                }
                else
                {
                  widgetFileTree.notifyListeners(SWT.Expand,treeEvent);
                }
                break;
              case FILE:
              case LINK:
                openFile(fileData);
                break;
              default:
                break;
            }
          }
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'c'))
        {
          TreeItem[] treeItems = widgetFileTree.getSelection();
          if (treeItems != null)
          {
            Widgets.setClipboard(clipboard,treeItems,0);
          }
        }
      }

      public void keyReleased(KeyEvent keyEvent)
      {
      }
    });
  }

  /** clear file tree, close all sub-directories
   */
  private void clearFileTree()
  {
    // close all directories
    synchronized(widgetFileTree)
    {
      for (TreeItem treeItem : widgetFileTree.getItems())
      {
        treeItem.removeAll();
        new TreeItem(treeItem,SWT.NONE);
      }
    }
  }

  /** find tree item
   * @param name name of tree item
   * @return tree item or null if not found
   */
  private TreeItem findTreeItem(TreeItem treeItems[], String name)
  {
    for (TreeItem treeItem : treeItems)
    {
      FileData fileData = (FileData)treeItem.getData();

      if ((fileData != null) && fileData.getName(repository).equals(name))
      {
        return treeItem;
      }
    }

    return null;
  }

  /** find index for insert of tree item in sorted list of tree items
   * @param treeItem tree item
   * @param fileData data of tree item
   * @return index in tree item
   */
  private int findFilesTreeIndex(TreeItem treeItem, FileData fileData)
  {
    int index = 0;

    synchronized(widgetFileTree)
    {
      TreeItem           subTreeItems[]     = treeItem.getItems();
      FileDataComparator fileDataComparator = new FileDataComparator(widgetFileTree);

      while (   (index < subTreeItems.length)
             && (fileDataComparator.compare(fileData,(FileData)subTreeItems[index].getData()) > 0)
            )
      {
        index++;
      }
    }

    return index;
  }

  /** open file sub-tree item
   * @param treeItem tree item to open
   */
  private void openFileTreeItem(TreeItem treeItem)
  {
    FileData directoryData = (FileData)treeItem.getData();

    if (!treeItem.getExpanded())
    {
      // remove existing sub-items
      treeItem.removeAll();

      // open directory and get file list data
      Set<FileData> fileDataSet = repository.openDirectory(directoryData.getName(repository));

      // add sub-tree
      for (FileData fileData : fileDataSet)
      {
        // create tree item
        TreeItem subTreeItem = Widgets.insertTreeItem(treeItem,
                                                      findFilesTreeIndex(treeItem,fileData),
                                                      fileData,
                                                      false
                                                     );
        if (fileData.type == FileData.Types.DIRECTORY) new TreeItem(subTreeItem,SWT.NONE);

        // add tree item reference
        fileData.addTreeItem(subTreeItem);

        // update
        updateTreeItem(subTreeItem,fileData);
      }
      treeItem.setExpanded(true);

      // start update states
      asyncUpdateFileStates(fileDataSet,true,directoryData.getName(repository));
    }
  }

  /** close file sub-tree
   * @param treeItem tree item to close
   */
  private void closeFileTreeItem(TreeItem treeItem)
  {
    FileData fileData = (FileData)treeItem.getData();

    // close directory
    repository.closeDirectory(fileData.getName(repository));

    // remove sub-tree
    treeItem.setExpanded(false);
    treeItem.removeAll();
    new TreeItem(treeItem,SWT.NONE);
  }

  /** open sub-directory and select item
   * @param tree tree
   * @param treeItems tree items
   * @param directory directory to open
   * @param selectName name if time to select or null
   * @return openend/selected tree item, null otherwise
   */
  private TreeItem openSubDirectory(Tree tree, TreeItem[] treeItems, String directory, String selectName)
  {
    TreeItem openedTreeItem = null;

    for (TreeItem treeItem : treeItems)
    {
      if (!treeItem.isDisposed())
      {
        FileData fileData = (FileData)treeItem.getData();
        String   name     = fileData.getName(repository);
//Dprintf.dprintf("name=#%s# directory=#%s#",name,directory);

        if (   name.isEmpty()
            || directory.equals(name)
            || directory.startsWith(name+File.separator)
           )
        {
          // open this tree item
          openFileTreeItem(treeItem);

//TODO: heere compare with absolute name?
          // go down to sub-items
          if (!directory.equals(fileData.getName()))
          {
            openedTreeItem = openSubDirectory(tree,treeItem.getItems(),directory,selectName);
          }
        }
        else
        {
          // select entry
          if (selectName != null)
          {
            if (selectName.equals(name))
            {
              tree.deselectAll();
              tree.select(treeItem);
              tree.setTopItem(treeItem);

              openedTreeItem = treeItem;
              break;
            }
          }
        }
      }
    }

    return openedTreeItem;
  }

  /** get file data image
   * @param fileData file data
   * @return image
   */
  private Image getFileDataImage(FileData fileData)
  {
    Image image = null;
    switch (fileData.type)
    {
      case DIRECTORY: image = Onzen.IMAGE_DIRECTORY; break;
      case FILE:      image = Onzen.IMAGE_FILE;      break;
      case LINK:      image = Onzen.IMAGE_LINK;      break;
      default:        image = Onzen.IMAGE_FILE;      break;
    }

    return image;
  }

  /** get file data state string
   * @param fileData file data
   * @return string
   */
  private String getFileDataStateString(FileData fileData)
  {
    String string = null;
    if (!repository.isIgnoreFile(fileData.getName(repository)))
    {
      switch (fileData.type)
      {
        case DIRECTORY:
          string = "";
          break;
        case FILE:
        case LINK:
          string = fileData.state.toString();
          break;
        default:
          string = "";
          break;
      }
    }
    else
    {
      string = "";
    }

    return string;
  }

  /** get file data foreground color
   * @param fileData file data
   * @return color
   */
  private Color getFileDataForeground(FileData fileData)
  {
    Color color = null;
    if (!repository.isIgnoreFile(fileData.getName(repository)))
    {
      color = Onzen.COLOR_BLACK;
    }
    else
    {
      color = COLOR_IGNORE;
    }

    return color;
  }

  /** get file data background color
   * @param fileData file data
   * @return color
   */
  private Color getFileDataBackground(FileData fileData)
  {
    Color color = null;
    if (!repository.isIgnoreFile(fileData.getName(repository)))
    {
      switch (fileData.type)
      {
        case DIRECTORY:
        case FILE:
        case LINK:
          switch (fileData.state)
          {
            case OK        : color = COLOR_OK;         break;
            case UNKNOWN   : color = COLOR_UNKNOWN;    break;
            case MODIFIED  : color = COLOR_MODIFIED;   break;
            case CHECKOUT  : color = COLOR_CHECKOUT;   break;
            case UPDATE    : color = COLOR_UPDATE;     break;
            case MERGE     : color = COLOR_MERGE;      break;
            case CONFLICT  : color = COLOR_CONFLICT;   break;
            case ADDED     : color = COLOR_ADDED;      break;
            case REMOVED   : color = COLOR_REMOVED;    break;
            case NOT_EXISTS: color = COLOR_NOT_EXISTS; break;
            case WAITING   : color = COLOR_WAITING;    break;
            case ERROR     : color = COLOR_ERROR;      break;
            default        : color = COLOR_UNKNOWN;    break;
          }
          break;
        default:
          color = COLOR_OK;
          break;
      }
    }
    else
    {
      color = COLOR_OK;
    }

    return color;
  }

  /** get selected file data
   * @return selected file data (first)
   */
  private FileData getSelectedFileData()
  {
    FileData fileData;

    TreeItem[] selectedTreeItems = widgetFileTree.getSelection();
    if (selectedTreeItems.length > 0)
    {
      fileData = (FileData)selectedTreeItems[0].getData();
    }
    else
    {
      fileData = null;
    }

    return fileData;
  }

  /** get selected files data
   * @return selected file data array
   */
  private FileData[] getSelectedFilesData()
  {
    FileData[] fileData;

    TreeItem[] selectedTreeItems = widgetFileTree.getSelection();
    if (selectedTreeItems.length > 0)
    {
      fileData = new FileData[selectedTreeItems.length];
      for (int i = 0; i < selectedTreeItems.length; i++)
      {
        fileData[i] = (FileData)selectedTreeItems[i].getData();
      }
    }
    else
    {
      fileData = null;
    }

    return fileData;
  }

  /** get selected file data set
   * @return selected file data hash set
   */
  private Set<FileData> getSelectedFileDataSet()
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    TreeItem[] selectedTreeItems = widgetFileTree.getSelection();
    for (TreeItem selectedTreeItem : selectedTreeItems)
    {
      fileDataSet.add((FileData)selectedTreeItem.getData());
    }

    return fileDataSet;
  }

  /** asynchronous update of selected file data set
   */
  private void asyncUpdateSelectedFileData()
  {
    ArrayList<TreeItem> updateTreeItemList = new ArrayList<TreeItem>();

    for (TreeItem selectedTreeItem : widgetFileTree.getSelection())
    {
      updateTreeItemList.add(selectedTreeItem);
    }
    asyncUpdateFileStates(updateTreeItemList.toArray(new TreeItem[updateTreeItemList.size()]));
  }

  /** get all file data set
   * @return all file data hash set
   */
  private Set<FileData> getAllFileDataSet()
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    LinkedList<TreeItem> treeItems = new LinkedList<TreeItem>();
    for (TreeItem treeItem : widgetFileTree.getItems())
    {
      treeItems.add(treeItem);
    }
    while (!treeItems.isEmpty())
    {
      TreeItem treeItem = treeItems.removeFirst();
      if (treeItem != null)
      {
        FileData fileData = (FileData)treeItem.getData();
        if (fileData != null) fileDataSet.add(fileData);

        for (TreeItem subTreeItem : treeItem.getItems())
        {
          treeItems.add(subTreeItem);
        }
      }
    }

    return fileDataSet;
  }

  //-----------------------------------------------------------------------

  /** update file tree item
   * @param treeItem tree item to update
   */
  private void updateTreeItem(final TreeItem treeItem, final FileData fileData)
  {
    if (!treeItem.isDisposed())
    {
      String name = fileData.getBaseName();
      if (name.isEmpty()) name = repository.rootPath;

      // set text, colors
      treeItem.setImage(getFileDataImage(fileData));
      treeItem.setForeground(getFileDataForeground(fileData));
      treeItem.setBackground(getFileDataBackground(fileData));
      treeItem.setText(0,name);
//treeItem.setImage(1,(fileData.state == FileData.States.OK) ? Onzen.IMAGE_LOCK : Onzen.IMAGE_EMPTY);
      treeItem.setText(1,getFileDataStateString(fileData));
      if (fileData.type != FileData.Types.DIRECTORY)
      {
        treeItem.setText(2,fileData.workingRevision);
        treeItem.setText(3,fileData.repositoryRevision);
        treeItem.setText(4,fileData.branch);
      }
      else
      {
        if (treeItem.getExpanded() && (treeItem.getItemCount() <= 0))
        {
          closeFileTreeItem(treeItem);
        }
      }
    }
  }

  /** update file tree item
   * @param treeItem tree item to update
   */
  private void updateTreeItem(final FileData fileData)
  {
    display.syncExec(new Runnable()
    {
      public void run()
      {
        for (TreeItem treeItem : fileData.getTreeItems())
        {
          updateTreeItem(treeItem,fileData);
        }
      }
    });
  }

  /** add file tree item
   * @param fileData file data
   */
  private void addTreeItem(final FileData fileData)
  {
    display.syncExec(new Runnable()
    {
      public void run()
      {
        synchronized(widgetFileTree)
        {
          if (!widgetFileTree.isDisposed())
          {
            // find parent tree item to insert new tree item
            TreeItem parentTreeItem = null;
            for (TreeItem treeItem : Widgets.getAllTreeItems(widgetFileTree))
            {
              FileData parentFileData = (FileData)treeItem.getData();
//if (fileData != null) Dprintf.dprintf("  fileData.getFileName=#%s#",fileData.getFileName());
              if ((parentFileData != null) && parentFileData.getName().equals(fileData.getDirectoryName()))
              {
//Dprintf.dprintf("    found treeItem=%s fileData=%s",treeItem,fileData);
                parentTreeItem = treeItem;
                break;
              }
            }

            if ((parentTreeItem != null) && !parentTreeItem.isDisposed())
            {
              // create tree item
              TreeItem treeItem = Widgets.insertTreeItem(parentTreeItem,
                                                         findFilesTreeIndex(parentTreeItem,fileData),
                                                         fileData,
                                                         false);

              // add tree item reference
              fileData.addTreeItem(treeItem);

              // update
              updateTreeItem(treeItem,fileData);
            }
          }
        }
      }
    });
  }

  /** remove file tree item from all repository tabs
   * @param treeItem tree item to update
   */
  private void removeTreeItem(TreeItem treeItem)
  {
    if (!treeItem.isDisposed())
    {
      TreeItem parentTreeItem = treeItem.getParentItem();

      if ((parentTreeItem != null) && (parentTreeItem.getItemCount() <= 1))
      {
        parentTreeItem.setExpanded(false);
      }
      treeItem.dispose();
      if ((parentTreeItem != null) && (parentTreeItem.getItemCount() <= 0))
      {
        closeFileTreeItem(parentTreeItem);
      }
    }
  }

  /** remove file tree item from all repository tabs
   * @param fileData file data
   */
  protected void removeTreeItems(final FileData fileData)
  {
    display.syncExec(new Runnable()
    {
      public void run()
      {
        for (TreeItem treeItem : fileData.getTreeItems())
        {
          removeTreeItem(treeItem);
        }
      }
    });
  }

  /** set file tree item foreground color in all repository tabs
   * @param fileData file data
   * @param color color
   */
  private void setTreeItemsForeground(final FileData fileData, final Color color)
  {
    display.syncExec(new Runnable()
    {
      public void run()
      {
        for (TreeItem treeItem : fileData.getTreeItems())
        {
          if (!treeItem.isDisposed())
          {
            treeItem.setForeground(color);
          }
        }
      }
    });
  }

  /** update file states in all tabs
   * @param fileDataSet file data set to update states
   * @param addNew true to add new file data
   * @param title title to show in status line or null
   */
  protected void updateFileStates(Set<FileData> fileDataSet, boolean addNew, String title)
  {
//synchronized(this) { Dprintf.dprintf("-------------------- %s",Thread.currentThread()); Dprintf.dprintf("addNew=%s",addNew); for (FileData f : fileDataSet) Dprintf.dprintf("  fileDataSet=%s",f); }
    try
    {
      setStatusText("Update status of '%s'...",(title != null) ? title : repository.title);

      // update tree items: set status "update in progress" color
      for (final FileData fileData : fileDataSet)
      {
        setTreeItemsForeground(fileData,COLOR_UPDATE_STATUS);
      }

      // update states, get new files
      HashSet<FileData> newFileDataSet = addNew ? new HashSet<FileData>() : null;
      try
      {
//synchronized(this) { Dprintf.dprintf("--- fileDataSet"); for (FileData f : fileDataSet) Dprintf.dprintf("  %s",f); }
        repository.updateStates(fileDataSet,newFileDataSet);
//synchronized(this) { Dprintf.dprintf("--- fileDataSet updated"); for (FileData f : fileDataSet) Dprintf.dprintf("  %s",f); }
//if (addNew) synchronized(this) { Dprintf.dprintf("--- newFileDataSet"); for (FileData f : newFileDataSet) Dprintf.dprintf("  %s",f); }
      }
      catch (final RepositoryException exception)
      {
        synchronized(Settings.showUpdateStatusErrors)
        {
          if (Settings.showUpdateStatusErrors)
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
                              Dialogs.booleanFieldUpdater(Settings.class,"showUpdateStatusErrors"),
                              exception.getExtendedMessage(),
                              "Getting file states fail:\n\n%s",
                              exception.getMessage()
                             );
              }
            });
          }
        }
      }

//Dprintf.dprintf("update ----------------- %s",Thread.currentThread());
      // update tree items: set status color, remove not existing entries
      FileData[] fileDataArray = fileDataSet.toArray(new FileData[fileDataSet.size()]);
      for (final FileData fileData : fileDataArray)
      {
        if (fileData.exists())
        {
          // file exists -> update state
          updateTreeItem(fileData);
        }
        else
        {
          // file disappeared -> remove tree entry
          fileDataSet.remove(fileData);
//Dprintf.dprintf("removed %s %s",fileData,fileData.getName());
          removeTreeItems(fileData);
        }
      }
//Dprintf.dprintf("update ----------------- done %s",Thread.currentThread());

      // add new tree items
      if (addNew)
      {
        for (final FileData newFileData : newFileDataSet)
        {
//???          TreeItem treeItem = newFileData.treeItem;
//          if ((treeItem == null) || treeItem.isDisposed())
          {
//Dprintf.dprintf("newFileData=%s newFileData.getDirectoryName=#%s#",newFileData,newFileData.getDirectoryName());
            // add new tree item
            addTreeItem(newFileData);
          }
        }
      }
    }
    finally
    {
      clearStatusText();
    }
//Dprintf.dprintf("-------------------- %s done",Thread.currentThread());
  }

  /** update file states
   * @param fileDataSet file data set to update states
   * @param addNew true to add new file data
   */
  protected void updateFileStates(Set<FileData> fileDataSet, boolean addNew)
  {
    if (fileDataSet != null)
    {
      updateFileStates(fileDataSet,addNew,(String)null);
    }
  }

  /** update file states
   * @param fileDataSet file data set to update states
   */
  protected void updateFileStates(Set<FileData> fileDataSet)
  {
    if (fileDataSet != null)
    {
      updateFileStates(fileDataSet,false);
    }
  }

  /** update file state
   * @param treeItems tree items
   * @param addNew true to add new file data
   * @param title title to show in status line or null
   */
  protected void updateFileStates(final TreeItem[] treeItems, boolean addNew, String title)
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    for (TreeItem treeItem : treeItems)
    {
      fileDataSet.add((FileData)treeItem.getData());
    }
    updateFileStates(fileDataSet,addNew,title);
  }

  /** update file state
   * @param treeItems tree items
   * @param addNew true to add new file data
   */
  protected void updateFileStates(final TreeItem[] treeItems, boolean addNew)
  {
    updateFileStates(treeItems,addNew,(String)null);
  }

  /** update file state
   * @param treeItems tree items
   */
  protected void updateFileStates(final TreeItem[] treeItems)
  {
    updateFileStates(treeItems,false);
  }

  /** update file state
   * @param fileData file data
   * @param title title to show in status line or null
   */
  protected void updateFileState(FileData fileData, String title)
  {
//synchronized(this) { Dprintf.dprintf("-------------------- %s",Thread.currentThread()); for (FileData f : fileDataSet) Dprintf.dprintf("  fileDataSet=%s",f); }
    try
    {
      setStatusText("Update status of '%s'...",(title != null) ? title : repository.title);

      // update tree items: set status "update in progress" color
      setTreeItemsForeground(fileData,COLOR_UPDATE_STATUS);

      // update states, get new files
      try
      {
//synchronized(this) { Dprintf.dprintf("-------------------- %s: %s",Thread.currentThread(),title); Dprintf.dprintf("  fileData=%s",fileData); }
        repository.updateStates(fileData.toSet(),null);
      }
      catch (final RepositoryException exception)
      {
        synchronized(Settings.showUpdateStatusErrors)
        {
          if (Settings.showUpdateStatusErrors)
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(shell,
                              Dialogs.booleanFieldUpdater(Settings.class,"showUpdateStatusErrors"),
                              exception.getExtendedMessage(),
                              "Getting file states fail:\n\n%s",
                              exception.getMessage()
                             );
              }
            });
          }
        }
      }

//Dprintf.dprintf("update ----------------- %s",Thread.currentThread());
      // update tree items: set status color, remove not existing entries
      if (fileData.exists())
      {
        // file exists -> update state
        updateTreeItem(fileData);
      }
      else
      {
        // file disappeared -> remove tree entry
//Dprintf.dprintf("removed %s %s",fileData,fileData.getName());
        removeTreeItems(fileData);
      }
    }
    finally
    {
      clearStatusText();
    }
//Dprintf.dprintf("-------------------- %s done",Thread.currentThread());
  }

  /** update file state
   * @param fileData file data
   */
  protected void updateFileState(FileData fileData)
  {
    updateFileState(fileData,null);
  }

  /** update file state
   * @param treeItem tree item
   */
  protected void updateFileState(final TreeItem treeItem)
  {
    updateFileState((FileData)treeItem.getData());
  }

  /** update file state
   * @param treeItem tree item
   */
  protected void updateAllFileStates()
  {
    updateFileStates(Widgets.getAllTreeItems(widgetFileTree),true);
  }

  /** asyncronous update file states
   * @param fileDataSet file data set to update states
   * @param addNew true to add new file data
   * @param title title to show in status line or null
   */
  protected void asyncUpdateFileStates(Set<FileData> fileDataSet, final boolean addNew, String title)
  {
    if (fileDataSet != null)
    {
      Background.run(new BackgroundRunnable(repository,fileDataSet,title)
      {
        public void run(Repository repository, Set<FileData> fileDataSet, String title)
        {
          try
          {
            updateFileStates(fileDataSet,addNew,(title != null) ? title : repository.title);
          }
          catch (Exception exception)
          {
            Onzen.printStacktrace(exception);
          }
        }
      });
    }
  }

  /** asyncronous update file states
   * @param fileDataSet file data set to update states
   * @param addNew true to add new file data
   */
  protected void asyncUpdateFileStates(Set<FileData> fileDataSet, boolean addNew)
  {
    asyncUpdateFileStates(fileDataSet,addNew,null);
  }

  /** asyncronous update file states
   * @param fileDataSet file data set to update states
   */
  protected void asyncUpdateFileStates(Set<FileData> fileDataSet)
  {
    asyncUpdateFileStates(fileDataSet,false);
  }

  /** asyncronous update file state
   * @param fileData file data
   * @param addNew true to add new file data
   * @param title title to show in status line or null
   */
  protected void asyncUpdateFileStates(FileData fileData, boolean addNew, String title)
  {
    Set<FileData> fileDataSet;
    if (fileData != null)
    {
      fileDataSet = fileData.toSet();
    }
    else
    {
      fileDataSet = new HashSet<FileData>();
      for (TreeItem treeItem : Widgets.getAllTreeItems(widgetFileTree))
      {
        fileData = (FileData)treeItem.getData();
        if (fileData != null) fileDataSet.add(fileData);
      }
    }
    asyncUpdateFileStates(fileDataSet,addNew,title);
  }

  /** asyncronous update file state
   * @param fileData file data
   * @param addNew true to add new file data
   */
  protected void asyncUpdateFileStates(FileData fileData, boolean addNew)
  {
    asyncUpdateFileStates(fileData,addNew,null);
  }

  /** asyncronous update file state
   * @param fileData file data
   */
  protected void asyncUpdateFileStates(FileData fileData)
  {
    asyncUpdateFileStates(fileData,false);
  }

  /** update file state
   * @param treeItems tree items
   * @param addNew true to add new file data
   * @param title title to show in status line or null
   */
  protected void asyncUpdateFileStates(final TreeItem[] treeItems, boolean addNew, String title)
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    for (TreeItem treeItem : treeItems)
    {
      FileData fileData = (FileData)treeItem.getData();
      if (fileData != null)
      {
        fileDataSet.add(fileData);
      }
    }
    asyncUpdateFileStates(fileDataSet,addNew,title);
  }

  /** update file state
   * @param treeItems tree items
   * @param title title to show in status line or null
   */
  protected void asyncUpdateFileStates(final TreeItem[] treeItems, String title)
  {
    asyncUpdateFileStates(treeItems,false,title);
  }

  /** update file state
   * @param treeItems tree items
   */
  protected void asyncUpdateFileStates(final TreeItem[] treeItems)
  {
    asyncUpdateFileStates(treeItems,(String)null);
  }

  /** asyncronous update file state
   * @param fileData file data
   */
  protected void asyncUpdateFileState(FileData fileData, String title)
  {
    Background.run(new BackgroundRunnable(repository,fileData,title)
    {
      public void run(Repository repository, FileData fileData, String title)
      {
        updateFileState(fileData,(title != null) ? title : repository.title);
      }
    });
  }

  /** update file state
   * @param treeItem tree item
   */
  protected void asyncUpdateFileStates(final TreeItem treeItem)
  {
    FileData fileData = (FileData)treeItem.getData();

    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(fileData);
    for (TreeItem subTreeItem : treeItem.getItems())
    {
      fileDataSet.add((FileData)subTreeItem.getData());
    }

    asyncUpdateFileStates(fileData);
  }

  /** update file state
   * @param treeItem tree item
   */
  protected void asyncUpdateAllFileStates()
  {
    display.syncExec(new Runnable()
    {
      public void run()
      {
        asyncUpdateFileStates(Widgets.getAllTreeItems(widgetFileTree));
      }
    });
  }

  //-----------------------------------------------------------------------

  /** get file editor
   * @param file file
   * @param mimeType mime type pattern
   * @return editor or null
   */
  private Settings.Editor getFileEditor(File file, String mimeType)
  {
    /** dialog data
     */
    class Data
    {
      String  name;
      String  mimeType;
      String  fileName;
      String  commandLine;
      int     validExitcode;
      boolean addNewCommand;

      Data()
      {
        this.name          = "";
        this.mimeType      = null;
        this.fileName      = null;
        this.commandLine   = null;
        this.validExitcode = 0;
        this.addNewCommand = false;
      }
    };

    Settings.Editor fileEditor = null;
    final Data      data       = new Data();

    Composite composite,subComposite,subSubComposite;
    Label     label;
    Button    button;

    data.fileName = file.getName();
    data.mimeType = mimeType;

    // get file name pattern and command from file associations if possible
    String suffix = (data.fileName.lastIndexOf('.') >= 0) ? data.fileName.substring(data.fileName.lastIndexOf('.')).toLowerCase() : null;
    if (suffix != null)
    {
      data.fileName    = "*"+suffix;
      data.commandLine = Onzen.getFileAssociation(suffix);
    }

    // command selection dialog
    final Shell dialog = Dialogs.openModal(shell,"Select command to open file '"+file.getPath()+"'",300,200,new double[]{1.0,0.0},1.0);

    // create widgets
    final Table   widgetEditors;
    final Text    widgetName;
    final Text    widgetMimeType;
    final Text    widgetFileName;
    final Text    widgetCommandLine;
    final Spinner widgetValidExitcode;
    final Button  widgetAddNewCommand;
    final Button  widgetOpen;
    composite = Widgets.newComposite(dialog,SWT.NONE,4);
    composite.setLayout(new TableLayout(new double[]{1.0,0.0,0.0},1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      widgetEditors = Widgets.newTable(composite);
      widgetEditors.setToolTipText("Edit commands.");
      Widgets.layout(widgetEditors,0,0,TableLayoutData.NSWE);
      Widgets.addTableColumn(widgetEditors,0,"Name",     SWT.LEFT,100,false);
      Widgets.addTableColumn(widgetEditors,1,"Mime type",SWT.LEFT,100,false);
      Widgets.addTableColumn(widgetEditors,2,"File name",SWT.LEFT,100,false);
      Widgets.addTableColumn(widgetEditors,3,"Command",  SWT.LEFT,400,true );
      Widgets.setTableColumnWidth(widgetEditors,Settings.geometryOpenFileColumns.width);

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0}));
      Widgets.layout(subComposite,1,0,TableLayoutData.WE);
      {
        label = Widgets.newLabel(subComposite,"Name:");
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetName = Widgets.newText(subComposite);
        widgetName.setToolTipText("Name of editor or empty.\n");
        Widgets.layout(widgetName,0,1,TableLayoutData.WE);

        label = Widgets.newLabel(subComposite,"Mime type:");
        Widgets.layout(label,1,0,TableLayoutData.W);

        widgetMimeType = Widgets.newText(subComposite);
        widgetMimeType.setToolTipText("Mime type pattern. Format: <type>/<sub-type>\n");
        widgetMimeType.setText(mimeType);
        Widgets.layout(widgetMimeType,1,1,TableLayoutData.WE);

        label = Widgets.newLabel(subComposite,"File name:");
        Widgets.layout(label,2,0,TableLayoutData.W);

        widgetFileName = Widgets.newText(subComposite);
        widgetFileName.setToolTipText("Simple file file name pattern, e. g. *.pdf.\n");
        if (data.fileName != null) widgetFileName.setText(data.fileName);
        Widgets.layout(widgetFileName,2,1,TableLayoutData.WE);

        label = Widgets.newLabel(subComposite,"Command:");
        Widgets.layout(label,3,0,TableLayoutData.W);

        subSubComposite = Widgets.newComposite(subComposite);
        subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0}));
        Widgets.layout(subSubComposite,3,1,TableLayoutData.WE);
        {
          widgetCommandLine = Widgets.newText(subSubComposite);
          widgetCommandLine.setToolTipText("Command to open file with.\nMacros:\n  %file% - file name\n  %files% - file names\n  %directory% - directory name\n  %directories% - directory names\n  %n% - line number\n  %% - %");
          if (data.commandLine != null) widgetCommandLine.setText(data.commandLine);
          Widgets.layout(widgetCommandLine,0,0,TableLayoutData.WE);

          button = Widgets.newButton(subSubComposite,Onzen.IMAGE_DIRECTORY);
          Widgets.layout(button,0,1,TableLayoutData.DEFAULT);
          button.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              String fileName = Dialogs.fileOpen(shell,
                                                 "Select program",
                                                 widgetCommandLine.getText(),
                                                 new String[]{"All files",  Onzen.ALL_FILE_EXTENSION,
                                                              "Scripts",    "*.sh",
                                                              "Batch files","*.cmd",
                                                              "Executables","*.exe"
                                                             },
                                                 Onzen.EXECUTABLE_FILE_EXTENSION
                                                );
              if (fileName != null)
              {
                widgetCommandLine.setText(FileUtils.escape(fileName));
              }
            }
          });
        }

        label = Widgets.newLabel(subComposite,"Valid exitcode:");
        Widgets.layout(label,4,0,TableLayoutData.W);

        widgetValidExitcode = Widgets.newSpinner(subComposite);
        widgetValidExitcode.setSelection(data.validExitcode);
        widgetValidExitcode.setMinimum(0);
        widgetValidExitcode.setMaximum(255);
        Widgets.layout(widgetValidExitcode,4,1,TableLayoutData.W);

        widgetAddNewCommand = Widgets.newCheckbox(subComposite,"add as new command");
        Widgets.layout(widgetAddNewCommand,5,1,TableLayoutData.W);
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog,SWT.NONE,4);
    composite.setLayout(new TableLayout(0.0,1.0));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetOpen = Widgets.newButton(composite,"Open");
      widgetOpen.setEnabled(!widgetCommandLine.getText().trim().isEmpty());
      Widgets.layout(widgetOpen,0,0,TableLayoutData.W,0,0,0,0,60,SWT.DEFAULT);
      widgetOpen.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          data.name          = widgetName.getText().trim();
          data.mimeType      = widgetMimeType.getText().trim();
          data.fileName      = widgetFileName.getText().trim();
          data.commandLine   = widgetCommandLine.getText();
          data.validExitcode = widgetValidExitcode.getSelection();
          data.addNewCommand = widgetAddNewCommand.getSelection();

          Settings.geometryOpenFile        = dialog.getSize();
          Settings.geometryOpenFileColumns = new Settings.ColumnSizes(Widgets.getTableColumnWidth(widgetEditors));

          Dialogs.close(dialog,true);
        }
      });

      button = Widgets.newButton(composite,"Cancel");
      Widgets.layout(button,0,1,TableLayoutData.E,0,0,0,0,60,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog,false);
        }
      });
    }

    // listeners
    widgetEditors.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        int index = widgetEditors.getSelectionIndex();
        if (index >= 0)
        {
          TableItem       tableItem = widgetEditors.getItem(index);
          Settings.Editor editor    = (Settings.Editor)tableItem.getData();

          widgetName.setText(editor.name);
          widgetCommandLine.setText(editor.commandLine);
        }
      }
    });
    widgetEditors.addListener(SWT.MouseDoubleClick,new Listener()
    {
      public void handleEvent(final Event event)
      {
        int index = widgetEditors.getSelectionIndex();
        if (index >= 0)
        {
          TableItem       tableItem = widgetEditors.getItem(index);
          Settings.Editor editor    = (Settings.Editor)tableItem.getData();

          data.name          = editor.name;
          data.mimeType      = "";
          data.fileName      = "";
          data.commandLine   = editor.commandLine;
          data.validExitcode = editor.validExitcode;
          data.addNewCommand = widgetAddNewCommand.getSelection();

          Settings.geometryOpenFile        = dialog.getSize();
          Settings.geometryOpenFileColumns = new Settings.ColumnSizes(Widgets.getTableColumnWidth(widgetEditors));

          Dialogs.close(dialog,true);
        }
      }
    });

    widgetCommandLine.addModifyListener(new ModifyListener()
    {
      public void modifyText(ModifyEvent modifyEvent)
      {
        Text widget = (Text)modifyEvent.widget;

        widgetOpen.setEnabled(!widget.getText().trim().isEmpty());
      }
    });

    // add existing editors
    for (Settings.Editor editor : Settings.editors)
    {
      TableItem tableItem = new TableItem(widgetEditors,SWT.NONE);
      tableItem.setData(editor);
      tableItem.setText(0,editor.name);
      tableItem.setText(1,editor.mimeType);
      tableItem.setText(2,editor.fileName);
      tableItem.setText(3,editor.commandLine);
    }

    // show dialog
    Dialogs.show(dialog,Settings.geometryOpenFile,Settings.setWindowLocation);

    // run
    widgetCommandLine.setFocus();
    if ((Boolean)Dialogs.run(dialog,false))
    {
      if (!data.commandLine.trim().isEmpty())
      {
        fileEditor = new Settings.Editor(data.name,
                                         data.mimeType,
                                         data.fileName,
                                         data.commandLine,
                                         data.validExitcode
                                        );
        if (data.addNewCommand)
        {
          // add editor
          Settings.editors = Arrays.copyOf(Settings.editors,Settings.editors.length+1);
          Settings.editors[Settings.editors.length-1] = fileEditor;
        }
      }
    }

    return fileEditor;
  }

  /** get file editor
   * @param file file
   * @param mimeType mime type pattern
   * @return editor or null
   */
  private Settings.Editor getFileEditor(String fileName, String mimeType)
  {
    return getFileEditor(new File(fileName),mimeType);
  }

  /** check if file contain TABs/trailing whitespaces
   * @param fileName file name
   * @param checkTABs TRUE to check for TABs
   * @param checkTrailingWhitespace TRUE to check for trailing whitespaces
   * @return TRUE iff file contain TABs or trailing whitespaces
   */
  private boolean containWhitespaces(String fileName, boolean checkTABs, boolean checkTrailingWhitespace)
  {
    boolean        trailingWhitespaces = false;
    boolean        trailingEmptyLines  = false;

    BufferedReader input = null;
    try
    {
      // open
      input = new BufferedReader(new FileReader(fileName));

      // check
      String        line;
      StringBuilder buffer = new StringBuilder();
      while ((line = input.readLine()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        int n = line.length();

        if (checkTrailingWhitespace)
        {
          // check if trailing whitespaces in line
          if  ((n > 0) && Character.isWhitespace(line.charAt(n-1)))
          {
            trailingWhitespaces = true;
          }
        }

        if (checkTABs)
        {
          // check for TABs
          for (int z = 0; z < n; z++)
          {
            char ch = line.charAt(z);
            if (ch == '\t')
            {
              trailingWhitespaces = true;
            }
          }
        }

        if (checkTrailingWhitespace)
        {
          // check for empty lines at end of file
          trailingEmptyLines = line.isEmpty();
        }
      }

      // close
      input.close(); input = null;
    }
    catch (IOException exception)
    {
      // ignored
    }
    finally
    {
      if (input != null) try { input.close(); } catch (IOException exception) { /* ignored */ }
    }

    return trailingWhitespaces || trailingEmptyLines;
  }

  /** check if file contain TABs
   * @param fileName file name
   * @return TRUE iff file contain TABs
   */
  protected boolean containTABs(String fileName)
  {
    return containWhitespaces(fileName,true,false);
  }

  /** check if file contain trailing whitespaces
   * @param fileName file name
   * @return TRUE iff file contain trailing whitespaces
   */
  protected boolean containTrailingWhitespaces(String fileName)
  {
    return containWhitespaces(fileName,false,true);
  }

  /** convert whitespaces in files
   * @param fileNames file name array
   * @param convertWhitespacesModes convert whitespaces modes
   * @param spacesPerTAB number of spaces per TAB
   */
  private void convertWhitespaces(java.util.List<String>           fileNameList,
                                  EnumSet<ConvertWhitespacesModes> convertWhitespacesModes,
                                  int                              spacesPerTAB
                                 )
    throws IOException
  {
    if (   convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING)
        || (spacesPerTAB > 0)
       )
    {
      final String EOL = Settings.eolType.get();

      for (String fileName : fileNameList)
      {
        if (!repository.isSkipConvertWhitespacesFile(fileName))
        {
          File           tmpFile = null;
          BufferedReader input   = null;
          PrintWriter    output  = null;
          try
          {
            // open
            File file = new File(fileName);
            tmpFile = File.createTempFile("convert-",".tmp",file.getParentFile());
            input  = new BufferedReader(new FileReader(file));
            output = new PrintWriter(new FileWriter(tmpFile));

            // convert
            int           trailingEmptyLineCount = 0;
            String        line;
            StringBuilder buffer = new StringBuilder();
            while ((line = input.readLine()) != null)
            {
  //Dprintf.dprintf("line=#%s#",line);
              // get length of line
              int n = line.length();

              if (convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING))
              {
                // get length of line without trailing whitespaces
                while ((n > 0) && Character.isWhitespace(line.charAt(n-1)))
                {
                  n--;
                }
              }

              // convert TABs
              buffer.setLength(0);
              for (int z = 0; z < n; z++)
              {
                char ch = line.charAt(z);
                if (convertWhitespacesModes.contains(ConvertWhitespacesModes.TABS) && (spacesPerTAB > 0) && (ch == '\t'))
                {
                  for (int i = 0; i < spacesPerTAB; i++)
                  {
                    buffer.append(' ');
                  }
                }
                else
                {
                  buffer.append(ch);
                }
              }

              // check if empty line
              if (convertWhitespacesModes.contains(ConvertWhitespacesModes.TRAILING) && (buffer.length() == 0))
              {
                trailingEmptyLineCount++;
              }
              else
              {
                // output previous collected empty lines
                while (trailingEmptyLineCount > 0)
                {
                  output.write(EOL);
                  trailingEmptyLineCount--;
                }

                // output line
                output.write(buffer.toString());
                output.write(EOL);
              }
            }

            // close
            output.close(); output = null;
            input.close(); input = null;

            // rename (keep executable flag, last modified)
            boolean isExecutable = file.canExecute();
            long lastModified = file.lastModified();
            File backupFile = new File(fileName+Settings.backupFileSuffix);
            backupFile.delete();
            if (!file.renameTo(backupFile))
            {
              throw new IOException("create backup file fail");
            }
            if (!tmpFile.renameTo(file))
            {
              throw new IOException("rename temporary file fail");
            }
            tmpFile = null;
            file.setExecutable(isExecutable);
            file.setLastModified(lastModified);
          }
          finally
          {
            if (output != null) output.close();
            if (input != null) try { input.close(); } catch (IOException exception) { /* ignored */ }
            if (tmpFile != null) tmpFile.delete();
          }
        }
      }
    }
  }

  /** convert whitespaces in file
   * @param fileName file name
   * @param convertWhitespacesModes convert whitespaces modes
   * @param spacesPerTAB number of spaces per TAB; 0 for not converting TABs
   */
  private void convertWhitespaces(String                           fileName,
                                  EnumSet<ConvertWhitespacesModes> convertWhitespacesModes,
                                  int                              spacesPerTAB
                                 )
    throws IOException
  {
    ArrayList<String> fileNameList = new ArrayList<String>();
    fileNameList.add(fileName);
    convertWhitespaces(fileNameList,convertWhitespacesModes,spacesPerTAB);
  }
}

/* end of file */
