/***********************************************************************\
*
* $Revision: 239 $
* $Date: 2015-09-04 17:26:34 +0200 (Fri, 04 Sep 2015) $
* $Author: torsten $
* Contents: Tree combo
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.util.ArrayList;

// graphics
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.TypedEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.graphics.TextLayout;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeColumn;
import org.eclipse.swt.widgets.TreeItem;
import org.eclipse.swt.widgets.TypedListener;
import org.eclipse.swt.widgets.Widget;

import java.util.EventListener;

/****************************** Classes ********************************/

/** tree combo widget
 */
public class TreeCombo extends Composite
{
  /** tree combo item
   */
  class Item<T>
  {
    TreeItem treeItem;
    Item<T>  parentItem;
    ItemList subItemList;
    ItemList filteredSubItemList;
    T        data;
    Object   values[];

    /** create item
     * @param data data
     * @param parentItem parent item
     * @param column values
     */
    public Item(T data, Item<T> parentItem, Object values[])
    {
      this.treeItem            = null;
      this.parentItem          = parentItem;
      this.subItemList         = new ItemList();
      this.filteredSubItemList = new ItemList();
      this.data                = data;
      this.values              = values;
    }

    /** create item
     * @param data data
     * @param column values
     */
    public Item(T data, Object values[])
    {
      this(data,(Item<T>)null,values);
    }

    /** create item
     */
    public Item(TreeItem treeItem)
    {
      this((T)null,(Object[])null);
      this.treeItem = treeItem;
    }

    /** check if item is visible
     * @param text text
     * @return true if text match to item data or values
     */
    boolean isVisible(String filters[])
    {
      if (filters.length > 0)
      {
        boolean visibleFlag;

        // check if data matches
        visibleFlag = true;
        String text = (data != null) ? getFilterText(data.toString()) : "";
        for (String filter : filters)
        {
          if (text.indexOf(filter) < 0)
          {
            visibleFlag = false;
            break;
          }
        }
        if (visibleFlag) return true;

        // check if text matches
        visibleFlag = true;
        for (String filter : filters)
        {
          visibleFlag = false;
          for (int i = 0; i < values.length; i++)
          {
            if (values[i] != null)
            {
              if      (values[i] instanceof String)
              {
                if (getFilterText((String)values[i]).indexOf(filter) >= 0)
                {
                  visibleFlag = true;
                  break;
                }
              }
              else if (values[i] instanceof Image)
              {
                // nothing to do
              }
              else
              {
                if (getFilterText(values[i].toString()).indexOf(filter) >= 0)
                {
                  visibleFlag = true;
                  break;
                }
              }
            }
          }
          if (!visibleFlag) break;
        }
        if (visibleFlag) return true;

        // check if some sub-item match text
        for (Item subItem : subItemList)
        {
          if (subItem.isVisible(filters))
          {
            visibleFlag = true;
            break;
          }
        }
        if (visibleFlag) return true;

        return false;
      }
      else
      {
        return true;
      }
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      StringBuilder buffer = new StringBuilder();
      for (int i = 0; i < values.length; i++)
      {
        if (values[i] != null)
        {
          buffer.append(' ');
          if      (values[i] instanceof String)
          {
            buffer.append((String)values[i]);
          }
          else if (values[i] instanceof Image)
          {
          }
          else
          {
            buffer.append(values[i].toString());
          }
        }
      }

      return "Item { data="+data.toString()+", parent="+((parentItem != null) ? parentItem.data : "<none>")+":"+buffer.toString()+", subItemCount="+subItemList.size()+" }";
    }
  };

  /** tree combo item list
   */
  class ItemList extends ArrayList<Item>
  {
  };

  public final static int DEFAULT_HEIGHT = 200;

  private final int additionalWidth;
  private final int height;
  private ItemList  itemList         = new ItemList();
  private ItemList  filteredItemList = new ItemList();
  private Button    widgetText;
  private Shell     dialog;
  private Tree      widgetTree;
  private Text      widgetFilter;
  private Object    data = null;         // selected data item

  /** create tree combo
   * @param composite parent composite
   * @param rightComposite additional (optional) widget on the right
   * @param additionalWidth additional width
   * @param height height
   * @param style style
   */
  public TreeCombo(Composite composite, Composite rightComposite, final int additionalWidth, final int height, int style)
  {
    super(composite,SWT.NONE);

    Composite subComposite,subSubComposite;
    Label     label;

    this.additionalWidth = additionalWidth;
    this.height          = height;

    // layout
    FillLayout fillLayout = new FillLayout();
    fillLayout.type = SWT.HORIZONTAL;
    this.setLayout(fillLayout);

    // combo text
    widgetText = new Button(this,SWT.PUSH|SWT.LEFT);

    // combo dialog
    dialog = new Shell(getShell(),SWT.TOOL);
    dialog.setLayout(new TableLayout(1.0,new double[]{1.0,0.0}));
    dialog.setVisible(false);

    subComposite = Widgets.newComposite(dialog);
    subComposite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
    Widgets.layout(subComposite,0,0,TableLayoutData.NSWE,0,0,2,2);
    {
      widgetTree = Widgets.newTree(subComposite,SWT.MULTI|SWT.VIRTUAL);
      widgetTree.setHeaderVisible(false);
      widgetTree.setLayout(new TableLayout(1.0,new double[]{0.0,1.0}));
      Widgets.layout(widgetTree,0,0,TableLayoutData.NSWE);

      subSubComposite = Widgets.newComposite(subComposite);
      subSubComposite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0}));
      Widgets.layout(subSubComposite,1,0,TableLayoutData.NSWE,0,0,2,2);
      {
        label = Widgets.newLabel(subSubComposite,"Filter:");
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetFilter = Widgets.newText(subSubComposite,SWT.SEARCH|SWT.ICON_CANCEL);
        Widgets.layout(widgetFilter,0,1,TableLayoutData.NSWE);
      }
    }
    if (rightComposite != null)
    {
      rightComposite.setParent(dialog);
      Widgets.layout(rightComposite,0,1,TableLayoutData.NS,0,0,2,2);
    }

    // add listeners
    getShell().addListener(SWT.Move,new Listener()
    {
      public void handleEvent(Event event)
      {
        if (!dialog.isDisposed())
        {
          if (dialog.isVisible())
          {
            Rectangle bounds = widgetText.getBounds();

            Point p = toDisplay(bounds.x,bounds.y);
            dialog.setLocation(p.x,p.y+bounds.height);
          }
        }
      }
    });
    getShell().addListener(SWT.Resize,new Listener()
    {
      public void handleEvent(Event event)
      {
        if (!dialog.isDisposed())
        {
          if (dialog.isVisible())
          {
            Rectangle bounds = widgetText.getBounds();

            dialog.setSize(bounds.width+additionalWidth,height);

            Point p = toDisplay(bounds.x,bounds.y);
            dialog.setLocation(p.x,p.y+bounds.height);
          }
        }
      }
    });

    dialog.addListener(SWT.Close,new Listener()
    {
      public void handleEvent(Event event)
      {
        close();
        event.doit = false;
      }
    });

    widgetText.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        if (!dialog.isDisposed())
        {
          if (dialog.isVisible())
          {
            close();
          }
          else
          {
            open();
          }
        }
      }
    });

    widgetTree.addListener(SWT.SetData,new Listener()
    {
      @Override
      public void handleEvent(final Event event)
      {
//TODO: remove if
        if (dialog.isVisible())
        {
          Tree     widget         = (Tree)event.widget;
          TreeItem treeItem       = (TreeItem)event.item;
          TreeItem parentTreeItem = treeItem.getParentItem();

          // get item
          int  index;
          Item item;
          if (parentTreeItem != null)
          {
            Item parentItem = (Item)parentTreeItem.getData();
            index = parentTreeItem.indexOf(treeItem);
            item  = parentItem.filteredSubItemList.get(index);
          }
          else
          {
            index = widget.indexOf(treeItem);
            item  = filteredItemList.get(index);
          }
//Dprintf.dprintf("set data index=%d item.subItemCount=%d item=%s parentTreeItem=%s",index,item.subItemList.size(),item,parentTreeItem);

          // store data
          treeItem.setData(item);
          item.treeItem = treeItem;

          // init tree item
          treeItem.setItemCount(item.filteredSubItemList.size());
          if (!item.filteredSubItemList.isEmpty()) treeItem.setExpanded(true);
//if (parentTreeItem != null) parentTreeItem.setExpanded(true);

          // format tree item columns
          for (int i = 0; i < item.values.length; i++)
          {
            if (item.values[i] != null)
            {
              if      (item.values[i] instanceof String)
              {
                treeItem.setText(i,(String)item.values[i]);
              }
              else if (item.values[i] instanceof Image)
              {
                treeItem.setImage(i,(Image)item.values[i]);
              }
              else
              {
                treeItem.setText(i,item.values[i].toString());
              }
            }
          }
        }
      }
    });
    widgetTree.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Tree widget = (Tree)selectionEvent.widget;

        TreeItem selectedTreeItem = getSelectedTreeItem();
        if (selectedTreeItem != null)
        {
          Event event;

          // get selected item data
          data = ((Item)selectedTreeItem.getData()).data;

          // trigger selection listeners
          event = new Event();
          event.widget = widgetText;
          event.item = selectedTreeItem;
          notifyListeners(SWT.Selection,event);

          // trigger modify listeners
          event = new Event();
          event.widget = widgetText;
          notifyListeners(SWT.Modify,event);
        }

        close();
      }
    });

    widgetFilter.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (Widgets.isAccelerator(keyEvent,SWT.ARROW_UP))
        {
          TreeItem selectedTreeItem = getSelectedTreeItem();

          // find next selectable item
          TreeItem newSelectedTreeItem = null;
          if (selectedTreeItem != null)
          {
            TreeItem parentItem = selectedTreeItem.getParentItem();

            // select parent item by default as new selected item
            newSelectedTreeItem = parentItem;

            // check if previous sub-item
            TreeItem treeItems[] = (parentItem != null) ? parentItem.getItems() : widgetTree.getItems();
            for (int i = 1; i < treeItems.length; i++)
            {
              if (treeItems[i] == selectedTreeItem)
              {
                newSelectedTreeItem = getLastTreeItem(treeItems[i-1]);
                break;
              }
            }
          }

          // select new item
          if (newSelectedTreeItem != null)
          {
            widgetTree.setSelection(newSelectedTreeItem);

            // trigger modify listeners
            Event event = new Event();
            event.widget = widgetText;
            notifyListeners(SWT.Modify,event);
          }
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.ARROW_DOWN))
        {
          TreeItem selectedTreeItem = getSelectedTreeItem();

          // find previous selectable item
          TreeItem newSelectedTreeItem = null;
          if (selectedTreeItem != null)
          {
            TreeItem subTreeItems[] = selectedTreeItem.getItems();
            if (subTreeItems.length > 0)
            {
              // select sub-item
              newSelectedTreeItem = subTreeItems[0];
            }
            else
            {
              do
              {
                TreeItem parentItem = selectedTreeItem.getParentItem();

                // check if next item
                TreeItem treeItems[] = (parentItem != null) ? parentItem.getItems() : widgetTree.getItems();
                for (int i = 0; i < treeItems.length-1; i++)
                {
                  if (treeItems[i] == selectedTreeItem)
                  {
                    newSelectedTreeItem = treeItems[i+1];
                    break;
                  }
                }

                // go up one level
                selectedTreeItem = parentItem;
              }
              while ((newSelectedTreeItem == null) && (selectedTreeItem != null));
            }
          }

          // select new item
          if (newSelectedTreeItem != null)
          {
            widgetTree.setSelection(newSelectedTreeItem);

            // trigger modify listeners
            Event event = new Event();
            event.widget = widgetText;
            notifyListeners(SWT.Modify,event);
          }
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'x'))
        {
          widgetFilter.setText("");
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    });
    widgetFilter.addModifyListener(new ModifyListener()
    {
      public void modifyText(ModifyEvent modifyEvent)
      {
        filterItems();
      }
    });
    widgetFilter.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        TreeItem selectedTreeItem = getSelectedTreeItem();
        if (selectedTreeItem != null)
        {
          Event event;

          // get selected item data
          data = ((Item)selectedTreeItem.getData()).data;

          // trigger selection listeners
          event = new Event();
          event.widget = widgetText;
          event.item = selectedTreeItem;
          notifyListeners(SWT.Selection,event);

          // trigger modify listeners
          event = new Event();
          event.widget = widgetText;
          notifyListeners(SWT.Modify,event);
        }

        close();
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });
  }

  /** create tree combo
   * @param composite parent composite
   * @param additionalWidth additional width
   * @param height height
   * @param style style
   */
  public TreeCombo(Composite composite, int additionalWidth, int height, int style)
  {
    this(composite,(Composite)null,additionalWidth,height,style);
  }

  /** create tree combo
   * @param composite parent composite
   * @param rightComposite additional (optional) widget on the right
   * @param style style
   */
  public TreeCombo(Composite composite, Composite rightComposite, int style)
  {
    this(composite,0,DEFAULT_HEIGHT,style);
  }

  /** create tree combo
   * @param composite parent composite
   * @param style style
   */
  public TreeCombo(Composite composite, int style)
  {
    this(composite,(Composite)null,style);
  }

  /** add column to tree widget
   * @param title column title
   * @param style style
   * @param width width of column
   * @param resizable TRUE iff resizable column
   * @return new tree column
   */
  public void addTreeColumn(String title, int style, int width, boolean resizable)
  {
    widgetTree.setHeaderVisible(true);
    Widgets.addTreeColumn(widgetTree,title,style,width,resizable);
  }

  /** set tree combo layout
   * @param columnWidth column width
   */
  public void setLayout(double columnWeight)
  {
    widgetTree.setLayout(new TableLayout(1.0,columnWeight));
  }

  /** set tree combo layout
   * @param columnWidths column widths
   */
  public void setLayout(double columnWeights[])
  {
    widgetTree.setLayout(new TableLayout(1.0,columnWeights));
  }

  /** set enabled
   */
  public void setEnabled(boolean enabled)
  {
    widgetText.setEnabled(enabled);
    if (!enabled && !dialog.isDisposed() && dialog.isVisible())
    {
      close();
    }
  }

  /** add tree item
   * @param data data
   * @param parentTreeItem parent tree item
   * @param values tree column values
   * @return tree item
   */
  public <T> Item addTreeItem(T data, Item parentItem, Object... values)
  {
    Item<T> item = new Item<T>(data,parentItem,values);
    if (parentItem != null)
    {
      parentItem.subItemList.add(item);
    }
    else
    {
      itemList.add(item);
    }

    return item;
  }

  /** add tree item
   * @param data data
   * @param values tree column values
   * @return tree item
   */
  public <T> Item addTreeItem(T data, Object... values)
  {
    return addTreeItem(data,(Item)null,values);
  }

  /** remove all tree items
   */
  public void removeAllTreeItems()
  {
    widgetTree.removeAll();
    itemList.clear();
    filteredItemList.clear();
  }

  /** filter visible items
   */
  public void filterItems()
  {
    // save selected item
    Item selectedItem = getSelectedItem();

    // clear
    widgetTree.removeAll();

    // filter items
    String filters[] = getFilterText(widgetFilter.getText()).split("\\s+");
    filterItems(itemList,filteredItemList,filters);
    for (final Item item : itemList)
    {
      filterItems(item.subItemList,item.filteredSubItemList,filters);
    }

    // refresh
//Dprintf.dprintf("---------- filterItems size %d",filteredItemList.size());
    widgetTree.setItemCount(filteredItemList.size());

    // get selected tree item
    TreeItem treeItems[] = widgetTree.getItems();
    TreeItem selectedTreeItem = null;
    for (final TreeItem treeItem : treeItems)
    {
      Item item = (Item)treeItem.getData();
      if (item == selectedItem)
      {
        selectedTreeItem = treeItem;
        break;
      }
    }

    // set selected item
    if (selectedTreeItem != null)
    {
      // restore selected item
      widgetTree.setSelection(selectedTreeItem);
    }
    else
    {
      // select first item
      if (treeItems.length > 0)
      {
        widgetTree.setSelection(treeItems[0]);
      }
    }
  }

  /** add modify listener
   * @param modifyListener modify listener to add
   */
  public void addModifyListener(ModifyListener modifyListener)
  {
    TypedListener typedListener = new TypedListener(modifyListener);
    addListener(SWT.Modify,typedListener);
  }

  /** add selection listener
   * @param selectionListener selection listener to add
   */
  public void addSelectionListener(SelectionListener selectionListener)
  {
    TypedListener typedListener = new TypedListener(selectionListener);
    addListener(SWT.Selection,typedListener);
    addListener(SWT.DefaultSelection,typedListener);
  }

  /** set focus
   */
  public boolean setFocus()
  {
    return widgetText.setFocus();
  }

  /** get selection
   * @return selected value text
   */
  public String getText()
  {
    return widgetText.getText();
  }

  /** set selection text
   * @param text value text
   */
  public void setText(String value)
  {
    widgetText.setText(value);
  }

  /** get selection
   * @param defaultValue default value
   * @return selected value or default value
   */
  public <T> T getSelection(T defaultValue)
  {
    return (data != null) ? (T)data : defaultValue;
  }

  /** get selection
   * @return selected value
   */
  public <T> T getSelection()
  {
    return getSelection((T)null);
  }

  /** select tree combo item
   * @param data data
   * @param text text format string
   * @param arguments optional arguments for text
   */
  public <T> void setSelection(T data, String text, Object... arguments)
  {
    this.widgetText.setText(String.format(text,arguments));
    this.data = data;

    if (data != null)
    {
      widgetTree.deselectAll();
      for (TreeItem treeItem : widgetTree.getItems())
      {
        if (select(treeItem,data))
        {
          break;
        }
      }
      widgetTree.showSelection();
    }
  }

  /** select tree combo item
   * @param text text format string
   * @param arguments optional arguments for text
   */
  public <T> void setSelection(String text, Object... arguments)
  {
    setSelection((T)null,text,arguments);
  }

  /** convert to string
   * @return string
   */
  public String toString()
  {
    return "TreeCombo";
  }

  // -----------------------------------------------------------------

  /** open tree combo
   */
  private void open()
  {
    Rectangle bounds = widgetText.getBounds();

    // set size
    dialog.setSize(bounds.width+additionalWidth,height);

    // show
    Point p = toDisplay(bounds.x,bounds.y);
    dialog.setLocation(p.x,p.y+bounds.height);
    dialog.setVisible(true);

    // set focus
    widgetFilter.setFocus();
    widgetFilter.setText("");

    // refresh
    widgetTree.setItemCount(filteredItemList.size());
    widgetTree.clearAll(true);
    widgetTree.setRedraw(true);
  }

  /** close tree combo
   */
  private void close()
  {
    if (!dialog.isDisposed())
    {
      if (dialog.isVisible())
      {
        dialog.setVisible(false);
      }
    }
  }

  /** get selected item
   * @return selected item or null
   */
  private Item getSelectedItem()
  {
    TreeItem selectedTreeItems[] = widgetTree.getSelection();
    return (selectedTreeItems.length > 0) ? (Item)selectedTreeItems[0].getData() : null;
  }

  /** set selected item in tree sub-item
   * @param treeItem tree item
   * @param item item
   * @return true if selected
   */
  private boolean setSelectedItem(TreeItem treeItem, Item item)
  {
    for (TreeItem subTreeItem : treeItem.getItems())
    {
      if      (subTreeItem.getData() == item)
      {
        widgetTree.setSelection(subTreeItem);
        return true;
      }
      else if (setSelectedItem(subTreeItem,item))
      {
        return true;
      }
    }

    return false;
  }

  /** set selected item
   * @param item item
   */
  private void setSelectedItem(Item item)
  {
    for (TreeItem treeItem : widgetTree.getItems())
    {
      if      (treeItem.getData() == item)
      {
        widgetTree.setSelection(treeItem);
        break;
      }
      else if (setSelectedItem(treeItem,item))
      {
        break;
      }
    }
  }

  /** get last tree item in sub-tree
   * @param treeItem tree item
   * @return last item in sub-tree of tree item or tree item
   */
  private TreeItem getSelectedTreeItem()
  {
    TreeItem selectedTreeItems[] = widgetTree.getSelection();
    return (selectedTreeItems.length > 0) ? selectedTreeItems[0] : null;
  }

  /** get selected tree item index
   * @return selected tree item index or -1
   */
  private int getSelectedTreeItemIndex()
  {
    TreeItem selectedTreeItems[] = widgetTree.getSelection();
    if (selectedTreeItems.length > 0)
    {
      TreeItem parentTreeItem = selectedTreeItems[0].getParentItem();
      return (parentTreeItem != null) ? parentTreeItem.indexOf(selectedTreeItems[0]) : widgetTree.indexOf(selectedTreeItems[0]);
    }
    else
    {
      return -1;
    }
  }

  /** get last tree item in sub-tree
   * @param treeItem tree item
   * @return last item in sub-tree of tree item or tree item
   */
  private TreeItem getLastTreeItem(TreeItem treeItem)
  {
    TreeItem subTreeItems[] = treeItem.getItems();
    if (subTreeItems.length > 0)
    {
      return getLastTreeItem(subTreeItems[subTreeItems.length-1]);
    }
    else
    {
      return treeItem;
    }
  }

  /** add item to tree combo
   * @param item item to add
   * @param filters filters
   */
//TODO: remove
/*
  private void XXXaddItem(Item item, String filters[])
  {
if (false){
    if (item.parentItem != null)
    {
      assert item.parentItem.treeItem != null;

      // add sub-item
      item.treeItem = new TreeItem(item.parentItem.treeItem,SWT.NONE);
      item.parentItem.treeItem.setExpanded(true);
    }
    else
    {
      // add item
      item.treeItem = new TreeItem(widgetTree,SWT.NONE);
    }

    // store data
    item.treeItem.setData(item);

    // format item columns
    for (int i = 0; i < item.values.length; i++)
    {
      if (item.values[i] != null)
      {
        if      (item.values[i] instanceof String)
        {
          item.treeItem.setText(i,(String)item.values[i]);
        }
        else if (item.values[i] instanceof Image)
        {
          item.treeItem.setImage(i,(Image)item.values[i]);
        }
        else
        {
          item.treeItem.setText(i,item.values[i].toString());
        }
      }
    }

    // add sub-items
    for (final Item subItem : filteredItemList)
    {
      if (   (subItem.parentItem == item)
          && isVisibleItem(subItem,filters)
         )
      {
        XXXaddItem(subItem,filters);
      }
    }
}
  }
*/

  /** get filter text
   * @param text text
   * @return filtered text
   */
  private String getFilterText(String text)
  {
    return text.replaceAll("[^A-Za-z0-9 ]","").toLowerCase().trim();
  }

  /** filter visible items
   * @param itemList item list
   * @param filteredItemList filtered item list
   * @param filters filters
   */
  private void filterItems(ItemList itemList, ItemList filteredItemList, String[] filters)
  {
    filteredItemList.clear();
    for (final Item item : itemList)
    {
      if (item.isVisible(filters))
      {
        filteredItemList.add(item);
        filterItems(item.subItemList,item.filteredSubItemList,filters);
      }
    }
  }

  /** select sub-tree combo item
   * @param data data
   * @param text text format string
   * @param arguments optional arguments for text
   */
  public <T> boolean select(TreeItem treeItem, T data)
  {
    Item item = (Item)treeItem.getData();
    if      ((item != null) && item.data.equals(data))
    {
      widgetTree.setSelection(treeItem);
      return true;
    }

    for (TreeItem subTreeItem : treeItem.getItems())
    {
      if (select(subTreeItem,data))
      {
        return true;
      }
    }

    return false;
  }
}

/* end of file */
